/*  Program to do a basic analysis of IBD inference with no LD */
/*  Up to 4 genomes, or genotypic pairs  */

/* Also includes options to simulate the ibd status of chromosomes    */
/*  Initially these will just be segregants  from first cousins */
/*  We will make the chromosomes 2 Morgans, so most have an IBD segment */

#define DEFINE_DATA
#define PNAM "phoenix"

#include "parsie.h"
#include "nghds.h"
#include "rans.h"

typedef struct _Chromseg
{
   int   fgl;
   double endpos;
} Chromseg;

typedef struct _Chromseglist
{
   struct _Chromseg *current;
   struct _Chromseglist *next;
} Chromseglist;

int   skipmark;                         /* number of markers skipped in analysis */

double **pprob, *penmat;                /* nph pheno, ns states */
double **transmat;                      /* ns x ns transition matrix */
                                  /* penmat and transmat now reset each locus */
                                  /*  and penmat only for given data */
static double tmat[15][15];
static int tmatset = 0;

static double hh, gg, delta;

#include  "haplo_process.c"
#include "make_ibd.c"
#include "hap_fwd_bk_compute.c"

int main (int argc, char *argv[])
{

   int   i, j, k;
   int   ns, nph;                       /* number of states, and phenotypes */
   int   nhd;                           /* number of actual data haplotypes (2 or 4 ) */
   int   nprs, **prs;                   /* if additional pairwise-analysis required   */
   int   nset;                          /* number of sets of data haplotypes  */
   int   totset;                        /*  nset + number of extracted pairs */
   int   nhap, nbg;                     /* number of haplos, and start and end of SNP block */
   int   nblk, ngap;                    /* spacing of chosen SNPs */
   int   nsnps1, nsnps;                 /* number of SNPS in analysis */
   int   nsnps2;                        /* count of snps extracted from real chromosome */
   double qq;                           /* required minor allele freq for inclusion */
   double fkin;                         /* population kinship */
   int   tot_marker;                    /* total number of markers in input */
   double tot_length;                   /*  total length to be analyzed      */

   int **haplo, **haplo1, *dex1, *dex;
   double *pif, *allefreq;              /* realized allele frequencies at each marker */
   /* and copied values used in analysis  */

   int **haplo2, **haplodat, *name;
   /* constructed haplos, and data haplotypes  */

   Chromseglist **chrom_dat, *true_ibd;
   int   chromotemp[4][200];
   double segs_ends[200];
   int   st15[200], st9[200];           /* 15-state and 9-state true ibd for 200
                                           chunks, as read from file   */
   int   nloc, *hapdat;                 /* number of loci and haplo data for analysis */
   double *snpdist, *snppos;            /* marker distances and positions  */

   double **qq_ibd;
   double ffibd[15];                    /* overall inbr,  marginal ibd prob */
   double ffrate;                       /* change of inbr, per obs mrk, given skipmark */

   int   igdx;                          /* indicator of genotypic data   */

   char  compu_fname[60], hapdat_fname[60], mrks_fname[60];
   /* file names in XTRA */
   FILE *fptr, *fptr2;

   initMorgProg (argv);

   /* parse input parameter statements  */
   if (!proc_pars (argc, argv, PNAM))
      exit (EXIT_FAILURE);

   prntStatusPP ();

   printf
      (" === PHOENIX PROGRAM: NO CHECKING OF PARAMETER STATEMENTS === \n \n");

   if (!(fptr2 = open_fil (XTRA)))      /* we will now use this for filenames */
      give_up (ERR_qit);

   fscanf (fptr2, "%s", compu_fname);
   fscanf (fptr2, "%s", hapdat_fname);
   fscanf (fptr2, "%s", mrks_fname);
   fclose (fptr2);

   printf ("%s  ", compu_fname);
   printf ("%s  ", hapdat_fname);
   printf ("%s  ", mrks_fname);

/* we will read nsnps SNPS taking best from blocks size nblk at gap ngap */
/*   at min freq qq */

   fptr2 = fopen (compu_fname, "r");
   fscanf (fptr2, "%d %d %d", &ns, &nph, &igdx);
   fscanf (fptr2, "%d %d %d", &nset, &nhd, &nprs);
   if (nprs)
   {
      prs = alloc_int2d (nprs, 2);
      for (i = 0; i < nprs; ++i)
         fscanf (fptr2, "%d %d", &prs[i][0], &prs[i][1]);
   }
   fscanf (fptr2, "%d %d %lf %lf", &tot_marker, &nhap, &tot_length, &qq);
   fscanf (fptr2, "%d %d %lf %lf", &nsnps, &skipmark, &fkin, &ffrate);
   fscanf (fptr2, "%d %d %d", &nblk, &ngap, &nbg);
   fclose (fptr2);

   printf ("\n Data check: from file %s", compu_fname);
   printf (" \n ns = %d, nph=%d, igdx=%d", ns, nph, igdx);
   printf ("  nset= %d, nhd= %d, npairs= %d", nset, nhd, nprs);
   printf ("\n tot-mark %d, nhap= %d, chro= %f, qq= %f",
           tot_marker, nhap, tot_length, qq);
   printf ("  nsnps=%d, skpmk= %d, fkin= %f, ffrate=%f", nsnps, skipmark, fkin,
           ffrate);
   printf ("\n nblk= %d, ngap= %d, nbg= %d", nblk, ngap, nbg);
   fflush (stdout);

   nloc = nsnps / skipmark;             /* number loci used in analysis  */
   nsnps1 = nsnps * nblk;               /* number SNPs actually read in  */

   if (!tot_marker)                     /* real data on sets of chromosomes */
   {
      nhap = nhd * nset;
      if (igdx)                         /* genotypes of individuals */
         name = malloc (nhap / 2 * sizeof (int));
      else                              /* list of hap-"names"   */
         name = malloc (nhap * sizeof (int));

      snppos = malloc (nloc * sizeof (double));
      haplodat = alloc_int2d (nhap, nsnps);
   }

   else                                 /* data to be constructed */
   {
      dex1 = malloc (nsnps1 * sizeof (int)); /* SNPs in input  */
      dex = malloc (nsnps * sizeof (int)); /* SNPs available for analysis */
      pif = malloc (nsnps * sizeof (double));

      haplo1 = alloc_int2d (nhap, nsnps1);
      haplo = alloc_int2d (nhap, nsnps);
      haplo2 = alloc_int2d (8, nsnps);
      haplodat = alloc_int2d (nhd, nsnps);
   }

   hapdat = malloc (nloc * sizeof (int)); /* All cases: analysis data */
   allefreq = malloc (nloc * sizeof (double));
   snpdist = malloc (nloc * sizeof (double));
   penmat = malloc (ns * sizeof (double));
   transmat = alloc_dbl2d (ns, ns);
   qq_ibd = alloc_dbl2d (nsnps, ns);
   pprob = alloc_dbl2d (nsnps, ns);

   if (!tot_marker)
   {                                    /* read the haplodata from a fixed file name */

      printf ("\n now to read the real genetic data ");
      fflush (stdout);
      if (!(fptr = fopen (hapdat_fname, "r")))
         give_up (ERR_qit);

      if (igdx)                         /* real data are genotypic */
      {
         for (i = 0; i < nhap; i = i + 2)
         {
            fscanf (fptr, "%d", &(name[i / 2]));
            for (j = 0; j < nsnps; ++j)
               fscanf (fptr, "%d %d", &(haplodat[i][j]), &(haplodat[i + 1][j]));
         }
      }
      else                              /* data are haplotypic  */
      {
         for (i = 0; i < nhap; ++i)
         {
            fscanf (fptr, "%d", &(name[i]));
            for (j = 0; j < nsnps; ++j)
               fscanf (fptr, "%d", &(haplodat[i][j]));
         }
         for (i = 0; i < nhap; ++i)
         {
            printf ("\n %d", name[i]);
            for (j = 0; j < 10; ++j)
               printf (" %d", haplodat[i][j]);
         }
      }
      printf ("\n now have read the  real data ");
      fflush (stdout);

/* set input SNP info data file  */
      if (!(fptr2 = fopen (mrks_fname, "r")))
         give_up (ERR_qit);

      for (j = 0; j < nsnps; ++j)
         fscanf (fptr2, "%lf", &(snppos[j]));
      for (j = 0; j < nsnps - 1; ++j)
         snpdist[j] = snppos[j + 1] - snppos[j];
      for (j = 0; j < nsnps; ++j)
         fscanf (fptr2, "%*d %lf %*f", allefreq + j);

   }
   else
   {
/* set input haplos data file  and input haplos */
      if (!(fptr2 = fopen (hapdat_fname, "r")))
         give_up (ERR_qit);

/* starting at nbg, we will read a block of
    nblk snps at a gap of ngap:  we will do this for nsnps block.
   we will then subselect the best SNP in each block   */

      input_haps (fptr2, nhap, nbg, ngap, nblk, nsnps, haplo1, dex1,
                  tot_marker);

      nsnps2 =
         select_snps (haplo1, haplo, nhap, nblk, nsnps, pif, dex1, dex, qq);

/* score frequencies */

      actual_freq (haplo, nhap, nsnps2, pif);

/* now we make the ibd structures as in the sim data */

/* allocate space and make fgl segments for chromosomes */
      chrom_dat = (Chromseglist **) malloc (nhd * sizeof (Chromseglist *));
      alloc_chk (WHERE, chrom_dat);
      for (i = 0; i < nhd; ++i)
      {
         chrom_dat[i] = (Chromseglist *) malloc (sizeof (Chromseglist));
         alloc_chk (WHERE, chrom_dat[i]);
      }

      true_ibd = (Chromseglist *) malloc (sizeof (Chromseglist));
      alloc_chk (WHERE, true_ibd);

/* read in the chomosomes from fixed file names */
/*
{
   fptr = fopen("chrom_ma.dat","r");
   read_chrom (fptr, chrom_dat[0], tot_length);
   fclose (fptr);

   fptr = fopen("chrom_pa.dat","r");
   read_chrom (fptr, chrom_dat[1], tot_length);
   fclose (fptr);
}
*/

/*  new format read-in of 4 chromosomes at fixed chunks */
      printf ("\n now to read the chromochunks");
      fflush (stdout);

      if (!(fptr = fopen (mrks_fname, "r")))
         give_up (ERR_qit);

      read_chunks (fptr, nhd, chromotemp, segs_ends, st15, st9);
      for (k = 0; k < nhd; ++k)
         make_chunks (chrom_dat[k], chromotemp[k], segs_ends, tot_length);
      fclose (fptr);

/*  make the true IBD status between the chromosomes    */
/*
make_true_ibd(true_ibd, chrom_dat[0], chrom_dat[1], tot_length);
print_chrom(true_ibd);
printf("\n \n");
*/

/* print out the chrom_dat info */
      for (i = 0; i < nhd; ++i)
      {
         print_chrom (chrom_dat[i]);
         printf ("\n \n");
      }

/* make the haplotypes from the selected SNPS:  2000x5 = 10000 */

      for (i = 0; i < 8; ++i)
      {
         printf ("\n now have make haplo i=%d ", i);
         fflush (stdout);
         make_haplo (haplo2[i], haplo, nsnps, 5 * i, 3, nsnps2);
      }

      make_freq (nsnps, 3, nsnps2, allefreq, pif);

      for (i = 0; i < nhd; ++i)
      {
         printf ("\n now have make haplodat i=%d ", i);
         fflush (stdout);
         make_haplo_segs (haplodat[i], nsnps, chrom_dat[i], haplo2, 0.01);
      }

      printf ("\n now have make haplo_segs ");
      fflush (stdout);

   }

/* set output file */
   if (!(fptr = open_fil (OXTR)))
      give_up (ERR_qit);

/* make the data vector of states from the haplo-data */
/* first haplodat has switched from 0,1 to 1,2 (with 0 missing  */
/* second we can now do several groups of 4 haplotypes: for 1 set kk=0 */

   if (tot_marker)
   {
      ffrate = -skipmark * log (0.995); /* a fiction -- to be fixed */
      for (j = 0; j < nloc - 1; ++j)
         snpdist[j] = 1.0;
   }

   delta = 0.2;
   hh = 1 - fkin;
   gg = fkin;

   set_marg_ibd (ns, ffibd, fkin);      /* fkin is overall kinship */
   printf ("\n equilib state probs at kinship %f\n", fkin);
   for (i = 0; i < ns; ++i)
      printf (" %5.3f", ffibd[i]);
   fflush (stdout);

/* reset data for test */
   totset = nset + nprs;

   for (k = 0; k < totset; ++k)         /* k th pair of genos */
   {
      printf ("\n data vector for %d markers in set %d to be set", nloc, k + 1);
      fflush (stdout);
      if ((!nprs) || (k == 0))
         make_data (hapdat, haplodat, nhd, k, igdx, nloc, skipmark, 0);
      else
      {
         ns = 2;
         nph = 4;
         make_pair_data (hapdat, haplodat, nloc, skipmark, 0, prs[k - 1]);
      }

      printf ("\n data vector for %d markers in set %d is set", nloc, k + 1);
      fflush (stdout);

/* printf(" for individuals %d and %d \n", name[2*k],name[2*k+1]); */
/*
for (i=0; i<nloc; ++i)
printf("\n marker %d, data %d %d %d %d, hapdat %d  allefrq %f dist %f",
       i, haplodat[4*k][i], haplodat[4*k+1][i], haplodat[4*k+2][i], 
       haplodat[4*k+3][i], hapdat[i], allefreq[i], snpdist[i]);
fflush(stdout);
*/

/*  check transmat */
      set_transmat (ns, transmat, ffibd, ffrate, 1.0);
      for (i = 0; i < ns; ++i)
      {
         printf ("\n");
         for (j = 0; j < ns; ++j)
            printf (" %7.5f", transmat[i][j]);
      }

      printf ("\n markers %d, phen %d , states %d", nloc, nph, ns);

      hap_fwd_compute (hapdat, nloc, nph, ns, penmat, transmat, allefreq,
                       ffibd, ffrate, snpdist);
      hap_bck_compute (hapdat, nloc, nph, ns, qq_ibd, penmat, transmat,
                       allefreq, ffibd, ffrate, snpdist);

      printf ("\n \n markers = %d, skip = %d \n", nloc, skipmark);
      fflush (stdout);
      if (!tot_marker)
         printf (" names %d %d \n", name[2 * k], name[2 * k + 1]);
      fprintf (fptr, "markers = %d, skip = %d \n", nloc, skipmark);
      fflush (fptr);
      if (!tot_marker)
         fprintf (fptr, " names %d %d \n", name[2 * k], name[2 * k + 1]);
      else if ((nprs) && (k > 0))
         fprintf (fptr, " haplotype pair %d %d \n", prs[k - 1][0],
                  prs[k - 1][1]);

      if ((skipmark == 1) && (tot_marker))
         for (i = 0; i < nloc; i = i + 10)
         {
            fprintf (fptr, " %d %f ", i, (i * skipmark) / 100.0);
            for (j = 0; j < ns; ++j)
               fprintf (fptr, " %f", qq_ibd[i][j]);
            fprintf (fptr, "\n");
         }
      else
         for (i = 0; i < nloc; ++i)
         {
            fprintf (fptr, " %d %f ", i, snppos[i]);
            for (j = 0; j < ns; ++j)
               fprintf (fptr, " %7.4f", qq_ibd[i][j]);
            fprintf (fptr, "\n");
         }
   }
   return EXIT_SUCCESS;
}
