/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
#include "gldisp_lm.h"

static int Num_seg_indicators = 3;
static char **Seg_indic_labels = 0;

static void clean_glut_lm (void)
{
   if (Seg_indic_labels)
   {
      int   i = 0;
      for (i = 0; i < Num_seg_indicators; ++i)
      {
         if (Seg_indic_labels[i])
         {
            free (Seg_indic_labels[i]);
         }
      }
   }
   free (Seg_indic_labels);
}

void gfmUserDefd_DefineWindows (void)
{
   extern int NGDispLoc;
   extern int *NGDispNams;

   /* name the views files */
   gfm_set_output_file_prefix ("lm_auto");

   gfm_setup_window (0, "Info Window");
   gfm_set_draw_func (lmaConsoleWindow);

   gfm_setup_window (1, "Segregation Indicators");
   gfm_set_draw_func (lmaDrawSegIndVectors);
   gfm_set_initial_clip (-(NGDispLoc) * 0.30, (NGDispLoc) * 1.15,
                         -NGDispNams[0] * 4 * 0.03, NGDispNams[0] * 4 * 1.03,
                         -1.0, 1.0);
   gsCOLOR_SCHEME (DEEP_BLUE);

   if (!Seg_indic_labels)
   {
      int   i = 0;
      Seg_indic_labels = malloc (sizeof *Seg_indic_labels * Num_seg_indicators);
      for (i = 0; i < Num_seg_indicators; ++i)
      {
         Seg_indic_labels[i] = malloc (sizeof *Seg_indic_labels[i] * 30);
      }
      strncpy (Seg_indic_labels[0], "0 (GrandMat)", 30);
      strncpy (Seg_indic_labels[1], "1 (GrandFat)", 30);
      strncpy (Seg_indic_labels[2], "-1 (Founder)", 30);

      atexit (clean_glut_lm);
   }

   gWinDefs[gNumWinDefs]->NumColorKeys = &Num_seg_indicators;
   gWinDefs[gNumWinDefs]->ColorKey = Seg_indic_labels;

   gfm_setup_window (2, "Founder Genome Labels");
   gfm_set_draw_func (lmaDrawGGVectors);
   gfm_set_initial_clip (-(NGDispLoc) * 0.30, (NGDispLoc) * 1.15,
                         -NGDispNams[0] * 4 * 0.03, NGDispNams[0] * 4 * 1.03,
                         -1.0, 1.0);
   gsCOLOR_SCHEME (FISHER_PRICE);
}

void gfmUserDefd_ResetAllAverages (void)
{
   return;
}

void gfmUserDefd_InitializeChain (void)
{
   return;
}

void gfmUserDefd_DefineMenus (void)
{
   return;
}

void gfmUserDefd_LastWords (void)
{
   return;
}

void lmaConsoleWindow (void)
{
   char  temp[250];
   int   CW = glutGetWindow ();
   int   i;
   GLfloat hl;

   /* global variables */
   extern int gNumSweeps;
   extern int NGDispLoc;

   ColorScheme3f *CS = gWindowsSettings[CW]->ColorScheme;

   glColor3fv (CS->Text);
   sprintf (temp, "Program:  %s", ProgNam);
   gfmRenderBitmapString (-.1f, 1.0f, GLUT_BITMAP_HELVETICA_12, temp);
   gfmRenderBitmapString (-.1f, .9f, GLUT_BITMAP_HELVETICA_12,
                          "Author:  EAT and others");

   sprintf (temp, "Number of MCMC iterations:  %d", gNumSweeps);
   gfmRenderBitmapString (-.1f, .8f, GLUT_BITMAP_HELVETICA_12, temp);

   sprintf (temp, "Number of loci:  %d", NGDispLoc);
   gfmRenderBitmapString (-.1f, .7f, GLUT_BITMAP_HELVETICA_12, temp);
   i = 0;
   hl = .7f;
   while (i < NGDispLoc)
   {
      hl -= .1f;
      fmt_locnamlin (temp, &i, NGDispLoc, GDispLoc);
      gfmRenderBitmapString (-.1f, hl, GLUT_BITMAP_HELVETICA_12, temp);
   }
}

void lmaDrawSegIndVectors (void)
{
   int   i, j;
   /* height of the bottom of the individual's box */
   GLfloat hh, h = 0.f, w = .8f;
   GLfloat hc;
   char  tempStr[100];
   ColorScheme3f *CS = gWindowsSettings[glutGetWindow ()]->ColorScheme;
   GLfloat WoH = gfmWoH ();

   int   idx = 0;
   int   locus_idx = 0;

   /* globals */
   extern int *NGDispNams;
   extern int NGDispLoc;
   extern char ***GDispNam;
   extern int *GDispLoc;
   extern int **xx0;
   extern int **xx1;

   /* cycle over the individuals in reverse */
   for (i = NGDispNams[0] - 1; i >= 0; i--, h += 4.f)
   {
      /* draw the wire box around the individual's two rows in the axis
       * color */
      glColor3fv (CS->Axes);
      gfduWireRectf (-0.1, h, (double) NGDispLoc, h + 4.);

      idx = GDispIx[0][i];

      /* put the individual's name to the left side of that box */
      sprintf (tempStr, "%s ", nghd[idx].self);
      /* the character size is scaled for number of chars in name */
      /* this has not been tested on windows */
      hc = GDispNamLen < 4 ? 2.f : 2.f - (GDispNamLen - 1) / 9.f;
      gfmStrokeString (tempStr, hc, .0f, h + .8f, 4, 0.f, WoH);

      hh = h + .75f;
      /* cycle over the loci */
      for (j = 0; j < NGDispLoc; j++)
      {
         locus_idx = GDispLoc[j];

         /* draw first series of rectangles for the individual */
         /*  if it's a founder the xx's will be -1.  So use Series
            color 2 for that */
         if (xx0[locus_idx][idx] < 0)
            glColor3fv (CS->Series[2]);
         else
            glColor3fv (CS->Series[xx0[locus_idx][idx]]);

         glRectf (j, hh, j + w, hh + 1.f);

         /* draw second series of rectangles for the individual, 1.5 units
            above the bottom of the first series  */
         if (xx1[locus_idx][idx] < 0)
            glColor3fv (CS->Series[2]);
         else
            glColor3fv (CS->Series[xx1[locus_idx][idx]]);
         glRectf (j, hh + 1.5, j + w, hh + 2.5f);
      }
   }
}

void lmaDrawGGVectors (void)
{
   int   i, j;
   /* height of the bottom of the individual's box */
   GLfloat hh, h = 0.f, w = .8f;
   GLfloat hc;
   char  tempStr[100];
   ColorScheme3f *CS = gWindowsSettings[glutGetWindow ()]->ColorScheme;
   GLfloat WoH = gfmWoH ();

   int   idx = 0;
   int   locus_idx = 0;

   extern int *NGDispNams;
   extern int NGDispLoc;
   extern char ***GDispNam;
   extern int *GDispLoc;
   extern int **gg0;
   extern int **gg1;

   /* cycle over the individuals in reverse */
   for (i = NGDispNams[0] - 1; i >= 0; i--, h += 4.f)
   {

      /* draw the wire box around the individual's two rows in the axis
       * color */
      glColor3fv (CS->Axes);
      gfduWireRectf (-0.1, h, (double) NGDispLoc, h + 4.);

      idx = GDispIx[0][i];

      /* put the individuals number to the left side of that box */
      sprintf (tempStr, "%s ", nghd[idx].self);
      /* the character size is adjusted by number of chars in name */
      /* this has not been tested on windows */
      hc = GDispNamLen < 4 ? 2.f : 2.f - (GDispNamLen - 1) / 9.f;
      gfmStrokeString (tempStr, hc, .0f, h + .8f, 4, 0.f, WoH);

      hh = h + .75f;
      /* cycle over the loci */
      for (j = 0; j < NGDispLoc; j++)
      {
         locus_idx = GDispLoc[j];

         /* draw first series of rectangles for the individual */
         /*  if it's a founder the xx's will be -1.  So use Series
            color 3 for that */
         if (gg0[locus_idx][idx] < 0)
            glColor3fv (CS->Text);
         else
            glColor3fv (CS->Series[gg0[locus_idx][idx] % 25]);

         glRectf (j, hh, j + w, hh + 1.f);

         /* draw second series of rectangles for the individual, 1.5 units
            above the bottom of the first series  */
         if (gg1[locus_idx][idx] < 0)
            glColor3fv (CS->Text);
         else
            glColor3fv (CS->Series[gg1[locus_idx][idx] % 25]);
         glRectf (j, hh + 1.5, j + w, hh + 2.5f);
      }
   }
}
