/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
#ifndef PARSEPROC_H
#define PARSEPROC_H

/**
 * @addtogroup parse
 * @{
 */

/* parsedef.h code  count variable           data arrays
   ...............  ..............           ...........
        "i"         ni_par                   j_par
        "k"         nk_par                     "
        "h"         nh_par                   u_par
        "r"         nr_par                   d_par
        "x"         nx_par                     "
        "n"         nm_par                   m_par
        "g"         nm_par, nk_par           m_par, j_par
        "a"         ni_par, nm_par, nk_par   j_par, m_par
        "s"         ns_par                   s_par
 */

static int      max_npar, max_tchar;
static int      ni_par, nk_par, nh_par, nr_par, nx_par, nm_par, ns_par;
static int    * j_par;                  /* integers              */
static uint32 * u_par;                  /* hexadecimal no.s      */
static double * d_par;                  /* doubles               */
static char  ** m_par;                  /* names                 */
static char     s_par[MAXLINLEN+2];     /* string                */
static int      prntd_flg;


/* *** function prototypes for proc_pars.c *** */

char  get_data_list (void); char *get_line (void);
int   close_par_fil (void);
int   get_aggr_key (void);
int   get_cmd_fnams (int, char **);
int   get_tokens (void);
int   open_par_fil (void);
int   proc_pars (int, char **, char *);
void  chk_stor (int, int);
void  free_arrs (void);
void  realloc_arrs (void);
void  realloc_toks (int);
void  set_defaults (void);

/* *** function prototypes for proc_parf.c and proc_parf_ss.c *** */

static FREQ_PTR alloc_dflt_frq (void);
static FREQ_PTR alloc_frq (void);
static FREQ_PTR alloc_frqx (int);
static INCO_PENE_PTR alloc_inco_pene_node (int, int);
static INCO_PENE_PTR chk_inco_pene_node (int, int);
static GENO_MEANS_PTR alloc_geno_means_node (int, int);
static GENO_MEANS_PTR chk_geno_means_node (int, int);
static GENO_MEANS_PTR chk_joint_geno_means_node (int, int);
static INHERTLOC_PTR alloc_inh_tl_node (int);
static INHERTLOC_PTR chk_inh_tl_node (int);
static IMAP_PTR add_vint (VMAP_PTR, int);
static IMAP_PTR chk_vint (VMAP_PTR, int, int, int);
static INDNAM *alloc_indnam_node (int, int);
static INDNAM *chk_indnam_node (int, int);
/* static JMAP *add_jmap_node (int, int, int); */
/* static JMAP *chk_jmap_node (int, int, int); */
/* static JNTMAP_PTR alloc_jnt (void); */
static KINS *alloc_kinshp_node (int, int);
static KINS *chk_kinshp_node (int, int);
static LOCUS *alloc_locus_nodex (int, int, int);
static LOCUS *chk_locus_node (int, int, int);
static LOCUS_PTR alloc_dflt_locus_node (int, int, int);
static LOCUS_PTR alloc_locus_node (int, int, int);
static LOCWND *alloc_wind_node (int);
static LOCWND *chk_wind_node (int);
static MAP_PTR add_map_node (int, int, int);
static MAP_PTR chk_map_node (int, int, int);
static MARKDAT *alloc_mrkdat_node (int);
static MARKDAT *chk_mrkdat_node (int);
static MARKNAM *alloc_mrknam_node (int);
static MARKNAM *chk_mrknam_node (int);
static PROBGAM *alloc_prbgam_node (int, int);
static PROBGAM *chk_prbgam_node (int, int);
static SIM_PTR alloc_sim_node (int);
static SIM_PTR chk_sim_node (int, int);
static SIMTRAIT_PTR alloc_sim_tr_node (void);
static SIMTRAIT_PTR chk_sim_tr_node (void);
static FIXEDEFF_PTR alloc_fixed_eff_node (TRAITEFF_PTR, int);
static TRAITEFF_PTR alloc_tr_eff_node (int);
static TRAITEFF_PTR chk_tr_eff_node (int);
static TRAITTLOC_PTR alloc_tr_tl_node (int);
static TRAITTLOC_PTR chk_tr_tl_node (int);
static TRTPED *alloc_trtped_node (int);
static TRTPED *chk_trtped_node (int, int);
/* static UJNTMAP_PTR alloc_ujnt (void); */
static UMAP_PTR add_map_unl_node (void);
static UMAP_PTR chk_map_unl_node (void);
static VARI *alloc_vari_node (int, int);
static VARI *chk_vari_node (int, int);
static VMAP *add_vmap_node (int, int);
static VMAP *chk_vmap_node (int, int);
static WNDPAT *alloc_wndpat_node (int, int);
static WNDPAT *chk_wndpat_node (int, int);
static XMAP_PTR alloc_vext (void);
static char *alloc_copy (char *);
static double *alloc_reco_frq (void);
static int   chk_covar (void);
static int   chk_covar_i (void);
static int   chk_ff (int);
static int   chk_inco_pene (void);
static int   chk_geno_means (void);
static int   chk_joint_geno_means (void);
static int   chk_indnam (int);
static int   chk_jmap (int, int);
/* static int   chk_jnt (JNTMAP_PTR, int); */
static int   chk_kinshp (int, int);
static int   chk_levels_ff (void);
static int   chk_locus (int, int);
static int   chk_locus_norm (void);
static int   chk_map (int, int, int);
static int   chk_map_unlnkd (void);
static int   chk_mrkdat (int);
static int   chk_mrknam (int);
static int   chk_prbgam (int, int);
static int   chk_pseudo (void);
static int   chk_reco_frq (void);
static int   chk_repeat (void);
static int   chk_set_inh_tl (void);
static int   chk_set_tr_eff (void);
static int   chk_set_tr_tl (int);
static int   chk_sim (int, int);
static int   chk_sim_tr (int);
static int   chk_stat (void);
static int   chk_trait (void);
static int   chk_trtped (int);
/* static int   chk_ujnt (UJNTMAP_PTR, int); */
static int   chk_var (void);
static int   chk_vari (void);
static int   chk_vext (XMAP_PTR, int);
static int   chk_vmap (int, int);
static int   chk_wind (int);
static int   chk_wndpat (int, int);
static int   compar_mkid_d (const void *, const void *);
/* static int   get_jord (int); */
static int   unnorm_frq (double);
static int   unnorm_frq_mx (int, double);
static void  alloc_dist (MAP_PTR, int);
static void  alloc_indnam_list (INDNAM_PTR);
static void  alloc_inh_tl_array (INHERTLOC_PTR);
/* static void  alloc_jntpos (JNTMAP_PTR); */
static void  alloc_mark_array (SIM_PTR);
static void  alloc_mrkdat_list (MARKDAT_PTR);
static void  alloc_mrknam_list (MARKNAM_PTR);
static void  alloc_nams_list (KINS *);
static void  alloc_prbgam_list (PROBGAM_PTR);
static void  alloc_trait_array (SIMTRAIT_PTR);
static void  alloc_fixed_eff_array (FIXEDEFF_PTR);
static void  alloc_tr_eff_array (TRAITEFF_PTR);
static void  alloc_tr_tl_array (TRAITTLOC_PTR, int);
static void  alloc_inco_pene_array (INCO_PENE_PTR);
static void  alloc_geno_means_array (GENO_MEANS_PTR);
/* static void  alloc_ujntpos (UJNTMAP_PTR); */
static void  alloc_unl_dist (UMAP_PTR);
static void  alloc_vdist (XMAP_PTR);
static void  alloc_vprop (IMAP_PTR);
static void  alloc_vmarknums (VMAP_PTR);
static void  alloc_wndpat_set (WNDPAT_PTR);
static void  chk_fil (char *);
static void  chk_seeds (void);
static void  free_dist (MAP_PTR);
static void  free_freq (LOCUS_PTR);
static void  free_indnam_list (INDNAM_PTR);
static void  free_inh_tl_array (INHERTLOC_PTR);
/* static void  free_jntpos (JNTMAP_PTR); */
static void  free_mark_array (SIM_PTR);
static void  free_mrkdat_list (MARKDAT_PTR);
static void  free_mrknam_list (MARKNAM_PTR);
static void  free_nams_list (KINS *);
static void  free_prbgam_list (PROBGAM_PTR);
static void  free_reco_frq (double **);
static void  free_trait_array (SIMTRAIT_PTR);
static void  free_fixed_eff_array (FIXEDEFF_PTR);
static void  free_tr_eff_array (TRAITEFF_PTR);
static void  free_tr_tl_array (TRAITTLOC_PTR);
static void  free_inco_pene_array (INCO_PENE_PTR);
static void  free_geno_means_array (GENO_MEANS_PTR);
static void  free_udist (UMAP_PTR);
/*static void  free_ujntpos (UJNTMAP_PTR); */
static void  free_vdist (XMAP_PTR);
static void  free_vprop (IMAP_PTR);
static void  free_wndpat_set (WNDPAT_PTR);
static void  reset_fit_cov (int);
static void  reset_fit_var (int);
static void  reset_set_eigen (void);
static void  reset_set_pseudo (void);
static void  reset_start_cov (int);
static void  reset_start_var (int);
static void  set_fit_cov (int);
static void  set_rewrt (char *);
static void  set_set_eigen (void);
static void  set_set_pseudo (void);
static void  set_start_cor (int);
static void  set_start_cov (int);
static void  set_start_var (int);
static void  start_trait_eff (void);
static void  start_trait_mean (void);
static void  stor_dflt_frq (FREQ_PTR);
static void  stor_dflt_locus_data (int, int, LOCUS_PTR, FREQ_PTR);
static void  stor_dflt_geno_mean_dat (int, int, GENO_MEANS_PTR);
static void  stor_dflt_vari_data (int, VARI_PTR);
static void  stor_frq (FREQ_PTR);
static void  stor_indnam (INDNAM_PTR, int);
static void  stor_inh_tl_array (INHERTLOC_PTR, int);
/* static void  stor_jntpos (JNTMAP_PTR, int, int ()); */
static void  stor_kinshp (KINS *, int, int);
static void  stor_locus_data (int, LOCUS_PTR, FREQ_PTR, int, int);
static void  stor_map (MAP_PTR, int, int, int);
static void  stor_map_unl (UMAP_PTR);
static void  stor_mark_array (SIM_PTR, int);
static void  stor_mrkdat (MARKDAT_PTR, int);
static void  stor_mrknam (MARKNAM_PTR, int);
static void  stor_prbgam (PROBGAM_PTR, int, int);
static void  stor_reco_frq (int *, double *);
static void  stor_sim_data (SIM_PTR, int, int);
static void  stor_sim_tr_data (SIMTRAIT_PTR, int);
static void  stor_trait_array (SIMTRAIT_PTR);
static void  stor_trtped_pos (int, int, int, int *, TRTPED_PTR);
static void  stor_fixed_eff_array (FIXEDEFF_PTR);
static void  stor_tr_eff_array (TRAITEFF_PTR);
static void  stor_tr_tl_array (TRAITTLOC_PTR, int, int);
static void  stor_inco_pene_array (INCO_PENE_PTR);
static void  stor_geno_means_array (GENO_MEANS_PTR);
/* static void  stor_ujntpos (UJNTMAP_PTR, int, int ()); */
static void  stor_vari_data (int, VARI_PTR);
static void  stor_vdist (XMAP_PTR, int, int ());
static void  stor_vprop (IMAP_PTR);
static void  stor_vmarknums (VMAP_PTR, int, int);
static void  stor_wind (LOCWND_PTR, int);
static void  stor_wndpat (WNDPAT_PTR, int, int);

/* *** function prototypes for proc_end.c *** */

int   proc_end (void);

/** @} */

#endif /*PARSEPROC_H*/
