/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

#include "markers.h"

/* void    alloc_xxgg  (int nl, int nped)     xx0, xx1, gg0, gg1 in sample.h
 * void    free_xxgg   (void)
 *
 * int  ** alloc_genes (int nl, int nped)       called by alloc_xxgg, alloc_y
 * void    free_genes  (int *** a)            called by free_xxgg, free_y
 *
 * arrays are allocated [nloci + 1] x [pedsize + 1]
 *                fixed to [nloci + 1] x [pedsize]??  EAT 12/2004
 *
 *  moved initialization to -1 from list_switches: EAT 11/2004
 */

void alloc_xxgg (int nl, int nped)
{
   int   i, j;
   xx0 = alloc_genes (nl, nped);
   xx1 = alloc_genes (nl, nped);
   gg0 = alloc_genes (nl, nped);
   gg1 = alloc_genes (nl, nped);

/* set-up corrected for only-kid, EAT Oct 2006;
   streamlined -1 setup will not work unless first sampling is MCMC setup */

   for (i = 0; i < nped; ++i)
   {
      if (nghd[i].pa->index == INVALID_INDEX)
         for (j = 0; j <= nl; ++j)
         {
            xx0[j][i] = -1;
            xx1[j][i] = -1;
         }
      else
         for (j = 0; j <= nl; ++j)
         {
            xx0[j][i] = 0;
            xx1[j][i] = 0;
         }
   }
   return;
}

void free_xxgg (void)
{
   free_genes (&xx0);
   free_genes (&xx1);
   free_genes (&gg0);
   free_genes (&gg1);
}

int **alloc_genes (int nl, int nped)
{
   int   j;
   int **a;

   a = malloc ((nl + 1) * sizeof (int *));
   alloc_chk (WHERE, a);
   a[0] = calloc (nped * (nl + 1), sizeof (int));
   alloc_chk (WHERE, a[0]);
   for (j = 1; j <= nl; ++j)
      a[j] = a[j - 1] + nped;

   return a;
}

void free_genes (int ***a)
{
   if (*a)
   {
      free (**a);
      **a = NULL;
      free (*a);
      *a = NULL;
   }
   return;
}
