/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
#include "quant.h"

double lik_eigen (double **gmat, double *evec, int nn, double *eigen2)
/*
 * This function takes a general nn x nn dimension matrix gmat as input,
 * and calls eispack function rg_() to solve the eigensystem.
 * It returns the largest eigenvalue and put the corresponding
 * eigenvector in evec.               - March 15. 2001 N.L.
 */
{
   int   i, j, ierr;
   int  *iv1;
   double *dummy, *rvec, *ivec, *zz, *fv1;

   /* Request calculating eigenvectors */
   int   calc_evec = 1;
   double max_evalue = -9999.9;
   int   max_evalue_index = -1;

   dummy = malloc (nn * nn * sizeof (double));
   alloc_chk (WHERE, dummy);
   rvec = malloc (nn * sizeof (double));
   alloc_chk (WHERE, rvec);
   ivec = malloc (nn * sizeof (double));
   alloc_chk (WHERE, ivec);
   zz = malloc (nn * nn * sizeof (double));
   alloc_chk (WHERE, zz);
   fv1 = malloc (nn * sizeof (double));
   alloc_chk (WHERE, fv1);
   iv1 = malloc (nn * sizeof (int));
   alloc_chk (WHERE, iv1);

   /* Transpose the matrix into an array */
   for (j = 0; j < nn; ++j)
      for (i = 0; i < nn; ++i)
         dummy[i + nn * j] = gmat[i][j];

   if (ChkEigC)
      printf (" about to call rg_()\n");

   rg_ (&nn, &nn, dummy, rvec, ivec, &calc_evec, zz, iv1, fv1, &ierr);

   if (ChkEigC)
      printf (" back out of rg_()\n");

   if (ierr)
      give_warn_where (WHERE, ERQ_rge, ierr);
   else if (ChkEigC)
      printf (" normal return from rg_() with ierr = %d\n\n", ierr);

   /* The largest real eigenvalue and its index and print all eigenvalues */
   /* the second largest eigenvalue */
   *eigen2 = 0.0;
#ifdef ALLEIGEN
   puts ("\n All eigenvalues :\n");
#endif /* ALLEIGEN */
   for (i = 0; i < nn; ++i)
   {
#ifdef ALLEIGEN
      printf ("%3d %8.4lf + %8.4lf i\n", i, rvec[i], ivec[i]);
#endif /* ALLEIGEN */
      if (abs (ivec[i]) < 1e-16)
      {
         if (rvec[i] > max_evalue)
         {
            /* save the second largest eigenvalue too */
            *eigen2 = max_evalue;
            max_evalue = rvec[i];
            max_evalue_index = i;
         }
         else if (rvec[i] > *eigen2)
         {
            *eigen2 = rvec[i];
         }
      }
   }
#ifdef ALLEIGEN
   putchar ('\n');
#endif /* ALLEIGEN */

   /* The max_eigenvalue_index th row of zz is the corresponding
      eigenvector */
   for (i = 0; i < nn; ++i)
      evec[i] = zz[i + max_evalue_index * nn];

   free ((void *) dummy);
   dummy = NULL;
   free ((void *) rvec);
   rvec = NULL;
   free ((void *) ivec);
   ivec = NULL;
   free ((void *) zz);
   zz = NULL;
   free ((void *) iv1);
   iv1 = NULL;
   free ((void *) fv1);
   fv1 = NULL;

   return max_evalue;
}
