/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

/*
 * proc_fils() checks for input and output pedigree files and output estimate
 * or scores files named.  It sets global variables for the programs 
 * processing these files.
 */

int proc_fils (void)
{
   if (!test_pedfil ())                 /* input pedigree file                */
      if (proc_err ())
         return FALSE;

   if (!test_opedfil ())                /* output pedigree file               */
      if (proc_err ())
         return FALSE;

   prnt_opedfil ();

   if (!test_orbfil ())                 /* ouput rao-blackwell estimate file  */
      if (proc_err ())
         return FALSE;

   if (!test_oscorfil ())               /* output score file                  */
      if (proc_err ())
         return FALSE;

   return proc_end ();
}

/* ------- test fncs:  test results of proccessing file statements --------- */

int test_pedfil (void)
/* checks if pedigree file named checks pedigree size against limits
 * checks record options (order, gender, observed)
 * and sets processing parameters
 */
{
   if (!(get_filent (PED)->fname))
      err_typ = ERF_EPF;

   if (PedSize <= 0)                    /* checks PedSize                    */
      err_typ = ERF_PE0;
   if (!allw_pedsiz)
      allw_pedsiz = pmx->max_ped;
   if (PedSize > allw_pedsiz)
   {
      if (UNIVAR)
         err_typ = ERF_PE1;
      else
         err_typ = ERF_PE2;
   }
   /*                                    * sets max component size,          */
   /*                                    * which is useful only for polyEM   */
   /*                                    * programs when they are computing  */
   /*                                    * eigenvalues                       */
   if (!allw_cmpsiz)
      allw_cmpsiz = pmx->max_cmp;
   if (allw_cmpsiz > PedSize)
      allw_cmpsiz = PedSize;

   if (!allw_obsiz)
      allw_obsiz = pmx->max_obs;
   if (allw_obsiz > PedSize)
      allw_obsiz = PedSize;

   MaxNCmps = pmx->max_n_cmp;           /* sets MaxNCmps                     */

   if (SexIn == UNDEF)                  /* checks SexIn and sets SexOp       */
      SexIn = Y;

   if (SexIn == Y)
   {
      if (SexOp == UNDEF)
         SexOp = CHECK_;
   }
   else if (SexIn == N)
   {
      if (SexOp != IGNORE_M)
         SexOp = ASSIGN;
   }
   /*                                    * sets SexOut                       */
   if (SexIn == Y || SexOp == ASSIGN)
      SexOut = Y;
   else
      SexOut = N;

   /*                                    * checks ObsIn and sets ObsOp       */
   if (ObsIn == UNDEF)
      ObsIn = N;

   if (ObsIn == Y)
   {
      if (ObsOp == UNDEF)
         ObsOp = USE;
   }
   else if (ObsIn == N)
   {
      if (ObsOp == UNDEF)
      {
         if (GENEDROP)
            ObsOp = ALL;
         else
            ObsOp = IGNORE_M;
      }
   }
   /*                                    * sets ObsOut                       */
   if (GENEDROP)
   {
      if (ObsOut == UNDEF)
         ObsOut = N;
   }
   else
   {
      if (ObsOp == ALL)
         ObsOut = Y;
      else
         ObsOut = ObsIn;
   }
   /*                                    * sets DiscardReals                 */
   if (KIN || GENEDROP || IBDDROP || MARKERDROP || LM_MAP || LM_PVAL
       || LM_IBDTESTS || LM_AUTO || LM_LODS || LM_BAYES)
      /* FIXME:  should probably test on TraitTyp             */
      DiscardReals = Y;

   else
      DiscardReals = N;

   StdIn[0] = NAM;                      /* sets NStdIn, StdIn[]              */

   if (parentOrdIn == UNDEF)
      parentOrdIn = PA1ST;

   if (parentOrdIn == PA1ST)
   {
      StdIn[1] = PA;
      StdIn[2] = MA;
   }
   else
   {
      StdIn[1] = MA;
      StdIn[2] = PA;
   }
   NStdIn = NNams;

   if (SexIn)
   {
      StdIn[NStdIn] = SEX;
      ++NStdIn;
   }
   if (ObsIn)
   {
      StdIn[NStdIn] = OBS;
      ++NStdIn;
   }
   /*                                    * checks NInts, NReals              */
   /*                                    * sets NNInts                       */
   if (!NNams)
      err_typ = ERF_IPR;

   else if (NInts < NStdIn - NNams)
   {
      err_typ = ERF_NIN;
      sprintf (errnumb, "%d", NStdIn - NNams);
      err_txt = errnumb;
   }
   else if (NInts < NFix)
   {
      err_typ = ERF_NIN;
      sprintf (errnumb, "%d", NFix);
      err_txt = errnumb;
   }
   else if (PEDCHECK || UNIVAR || UNIBIG || BIVAR || MULTIVAR)
      if (NReals < NTraits)
      {
         err_typ = ERF_NRL;
         sprintf (errnumb, "%d", NTraits);
         err_txt = errnumb;
      }
   NNInts = NInts + NNams - NStdIn;

   if (!err_typ)
      return TRUE;
   else
      return FALSE;
}

int test_opedfil (void)
{
   if (GENEDROP)
      if (!(get_filent (OPED)->fname))
         err_typ = ERF_OPF;

   StdOut[0] = NAM;                     /* sets NStdOut, StdOut[]            */

   if (parentOrdOut == UNDEF)
      parentOrdOut = parentOrdIn;

   if (parentOrdOut == PA1ST)
   {
      StdOut[1] = PA;
      StdOut[2] = MA;
   }
   else
   {
      StdOut[1] = MA;
      StdOut[2] = PA;
   }
   NStdOut = NNams;

   if (SexOut)
   {
      StdOut[NStdOut] = SEX;
      ++NStdOut;
   }

   if (ObsOut)
   {
      StdOut[NStdOut] = OBS;
      ++NStdOut;
   }

   if (LabelsOut == UNDEF)
      LabelsOut = N;

   if (LatentOut == UNDEF)
      LatentOut = N;

   if (!err_typ)
      return TRUE;
   else
      return FALSE;
}

void prnt_opedfil (void)
{
   int   i;

   if (!GENEDROP)
      return;

   fprintf (stdout, "\n Each record of %s \"%s\" will contain these items:\n\n",
            (get_filent (OPED))->fdesc, ((get_filent (OPED))->fname));

   fprintf (stdout, "  %4d -%4d:", 1, NNams + NInts);
   if (parentOrdOut == PA1ST)
      fprintf (stdout, "  name, father, mother");
   else if (parentOrdOut == MA1ST)
      fprintf (stdout, "  name, mother, father");
   if (SexOut)
      fprintf (stdout, ", gender");
   if (ObsOut)
      fprintf (stdout, ", observed indicator");
   if (NNInts)
      fprintf (stdout, ", %d additional integer", NNInts);
   if (NNInts > 1)
      fprintf (stdout, "s");

   fprintf (stdout, "\n");

   i = NNams + NInts;

   if (LabelsOut)
   {
      fprintf (stdout, "  %4d -%4d:", i + 1, i + 2 * NMarks);
      if (NMarks == 1)
         fprintf (stdout, "  labels for the 1 marker:\n");
      else if (NMarks > 1)
         fprintf (stdout, "  labels for all %d markers:\n", NMarks);
      if (parentOrdOut == MA1ST)
         fprintf (stdout, "               labels in maternal, paternal order");
      else if (parentOrdOut == PA1ST)
         fprintf (stdout, "               labels in paternal, maternal order");
      if (NMarks > 1)
         fprintf (stdout, "; markers in chromosome order");
      fprintf (stdout, "\n");
      i += 2 * NMarks;
   }

   fprintf (stdout, "  %4d -%4d:", i + 1, i + 2 * NMarks);
   if (NMarks == 1)
      fprintf (stdout, "  genotypes for the 1 marker:\n");
   else if (NMarks > 1)
      fprintf (stdout, "  genotypes for all %d markers:\n", NMarks);
   fprintf (stdout, "               alleles of ea genotype ordered");
   if (NMarks > 1)
      fprintf (stdout, "; markers in chromosome order");
   fprintf (stdout, "\n");
   i += 2 * NMarks;

   if (NTLocs && LabelsOut)
   {
      fprintf (stdout, "  %4d -%4d:", i + 1, i + 2 * NTLocs);
      if (NTLocs == 1)
         fprintf (stdout, "  labels for the 1 tloc:\n");
      else if (NTLocs > 1)
         fprintf (stdout, "  labels for all %d tlocs:\n", NTLocs);
      if (parentOrdOut == MA1ST)
         fprintf (stdout, "               labels in maternal, paternal order");
      else if (parentOrdOut == PA1ST)
         fprintf (stdout, "               labels in paternal, maternal order");
      if (NTLocs > 1)
         fprintf (stdout, "; tlocs in order of tloc no");
      fprintf (stdout, "\n");
      i += 2 * NTLocs;
   }

   if (NTLocs && LatentOut)
   {
      fprintf (stdout, "  %4d -%4d:", i + 1, i + 2 * NTLocs);
      if (NTLocs == 1)
         fprintf (stdout, "  genotypes for the 1 tloc:\n");
      else if (NTLocs > 1)
         fprintf (stdout, "  genotypes for all %d tlocs:\n", NTLocs);
      fprintf (stdout, "               alleles of ea genotype ordered");
      if (NTLocs > 1)
         fprintf (stdout, "; tlocs in order of tloc no");
      fprintf (stdout, "\n");
      i += 2 * NTLocs;
   }

   if (LatentOut && NTraits)
   {
      fprintf (stdout, "  %4d -%4d:", i + 1, i + 3 * NTraits);
      fprintf (stdout, "  additive, residual, and quantitative values");
      if (NTraits == 1)
         fprintf (stdout, " for the 1 trait:\n");
      else if (NTraits > 1)
         fprintf (stdout, " for all %d traits:\n", NTraits);
      fprintf (stdout, "               traits in order of trait no");
      fprintf (stdout, "\n");
      i += 3 * NTraits;
   }
   else if (NTraits)
   {
      fprintf (stdout, "  %4d -%4d:", i + 1, i + NTraits);
      fprintf (stdout, "  quantitative values");
      if (NTraits == 1)
         fprintf (stdout, " for the 1 trait\n");
      else if (NTraits > 1)
      {
         fprintf (stdout, " for all %d traits:\n", NTraits);
         fprintf (stdout, "               traits in order of trait no");
      }
      fprintf (stdout, "\n");
      i += NTraits;
   }

   fprintf (stdout, "\n");
   fflush (stdout);

   return;
}

int test_oscorfil (void)
{
   if (LM_AUTO || LM_MARKERS)
   {
      if (!(get_filent (OSCORE)->fname))
         FilIt = 0;
      else
         fprintf (stdout, " Interim scores file to be written every"
                  " %d iterations\n\n", FilIt);
   }
   if (!err_typ)
      return TRUE;
   else
      return FALSE;
}

int test_orbfil (void)
{
   if (LM_BAYES)
   {
      if (get_filent (ORBEST)->fname)
      {
         fprintf (stdout, " An output file containing the rao-blackwellized"
                  " estimates\n    will be written at the completion of the MC"
                  " iterations\n\n");
         OutRB = Y;
      }
      else
         OutRB = N;
   }
   if (!err_typ)
      return TRUE;
   else
      return FALSE;
}

void alloc_sel (void)
{
   int   n, i;

   n = NInts + (SexIn != N ? 0 : 1);
   SelInt = malloc (n * sizeof (int));
   alloc_chk (WHERE, SelInt);
   for (i = 0; i < n; ++i)
      SelInt[i] = 0;

   if (NReals)
   {
      SelReal = malloc (NReals * sizeof (int));
      alloc_chk (WHERE, SelReal);
      for (i = 0; i < NReals; ++i)
         SelReal[i] = 0;
   }

   return;
}
