/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

/* these functions are called from proc_lods().
 */

/**
 * checks select marker and trait statements for errors.
 */
int test_lods_sel (void)
{
   if (!first_sim_chrom_node)           /*select ... markers .. required      */
   {
      err_typ = ERA_SEL;
      return FALSE;
   }

   else if (first_sim_chrom_node->p_next) /*checks for only one chromosome    */
   {
      err_typ = ERA_OOS;
      return FALSE;
   }

   if (!sim_traits)                     /* checks for select trait            */
   {
      err_typ = ERA_MST;
      return FALSE;
   }

   return TRUE;
}

/**
 * checks scalar parameters and sets defaults
 */
int test_lods_pars (void)
{
   if (NIt == UNDEF)
   {
      err_typ = ERA_NIT;
      return FALSE;
   }

   if (SmplSetup == UNDEF)
      SmplSetup = SEQIMP;

   if (NSeqImp == UNDEF)
   {
      NSeqImp = RSEQIMP * NIt;
      if (NSeqImp < 2)
         NSeqImp = 2;
   }

   if (Scan == UNDEF)
      Scan = Y;

   if (ScorIt == UNDEF)
   {
      if (LM_TWOQTL)
         ScorIt = 10;
      else
         ScorIt = 1;
   }

   if (SeqImpIt == UNDEF)
      SeqImpIt = 0;

   if (GlobalMcmc == UNDEF)
      GlobalMcmc = 0;

   if (LM_BAYES)
      if (NPseudoPrIt == UNDEF)
         NPseudoPrIt = RPPIT * NIt;

   if (NBrnIt == UNDEF)
   {
      if (LM_LODS)
         NBrnIt = RBRNIT_LODS * NIt;
      else if (LM_BAYES)
         NBrnIt = RBRNIT_BAYES * NPseudoPrIt;
      else if (LM_MARKERS)
         NBrnIt = RBRNIT_MRKS * NIt;
      else if (LM_TWOQTL)
         NBrnIt = RBRNIT_TWOQTL * NIt;
   }

   if (LikRatStats == 1)                /* checks lhood-ratio stat options    */
   {
      IBDStats = 0;
      if (SpairsOn != UNDEF || RobdomOn != UNDEF /* first measures         */
          || SAffUnaffOn != UNDEF || LambdaOn != UNDEF || MaxFGLOn != UNDEF)
         err_typ = ERA_WLM;
      SpairsOn = 0;
      RobdomOn = 0;
      SAffUnaffOn = 0;
      LambdaOn = 0;
      MaxFGLOn = 0;

      if (DeltaOn == UNDEF || Lambda_pOn == UNDEF)
      {
         DeltaOn = 1;                   /* assigns meas defaults              */
         Lambda_pOn = 0;
      }
      if (NormalEnable != UNDEF || PermutEnable != UNDEF) /* checks tests  */
         err_typ = ERA_WLT;
      NormalEnable = 0;                 /* assigns tests defaults             */
      PermutEnable = 0;

      if (ChiSqEnable == UNDEF)
         ChiSqEnable = 1;

      if (Lambda_pOn)                   /* assigns grid spacing               */
      {
         IntervalSegL = .5 / (NumOfSegL - 1.);
         IntervalSegP = .8 / (NumOfSegP - 1.);
      }
   }

/* Fill in default settings for a few new global variables used by lm_twoqtl. */
   if (UseMCSummation == UNDEF)
      UseMCSummation = FALSE;           /* Default to exact summation.        */

/* When using MC summation for trait, the number of iterations to use for MC  */
/* summation must be defined.                                                 */
   if (UseMCSummation == TRUE && MCRealizations == UNDEF)
      err_typ = ERA_UMR;

   if (ChkIt == UNDEF)
      ChkIt = NIt / 5;

   if (FilIt == UNDEF)
      FilIt = NIt;

   if (err_typ)
      return FALSE;

   return TRUE;
}

/* Checks pseudo-priors.                                                      */

int test_ps_pr (void)
{
   if (InPseudo)
   {
      if (n_pseudo != NTPos + 1)
      {
         err_typ = ERA_NPP;
         sprintf (errnumb, "%d", NTPos + 1);
         err_txt = errnumb;
         return FALSE;
      }
      else if (PseudoPrior[0] == 0.0)
      {
         err_typ = ERA_ZUT;
         return FALSE;
      }
   }

   return TRUE;
}

/**
 * For selected chromosome and for each gender, adds a list of the varying
 * test position tlocs required.
 */
int updt_sim_vtloc (SIM_PTR p_s)
{
   int   sx, entry, nt, ix;
   VMAP_PTR p_vmap;
   SIM_PTR p_sim;

   for (sx = FEM; sx <= AVG; ++sx)      /* Find varying position test tlocs.  */
   {
      nt = 0;
      p_vmap = first_tloc_vmap_node[sx]; /* Look through each tloc_vmap list. */
/* Count the number of varying test position tlocs to simulate.               */
      while (p_vmap)
      {
         if (p_vmap->selectd)
            ++nt;
         p_vmap = p_vmap->p_next;
      }

      if (nt)                           /* If tlocs required,                 */
      {
         entry = FALSE;
         while ((p_sim = nxt_sim_chrom (&entry))
                && p_sim->chrom_no < p_s->chrom_no)
            ;                           /* Finds sim_node for the chromosome. */
         if (p_sim && p_sim->chrom_no == p_s->chrom_no)
         {
            if (!p_sim->no_tloc_sel)    /* If tloc list not already set,      */
            {
               p_sim->no_tloc_sel = nt; /* Set number of tlocs selected.      */
               p_sim->p_tloc = malloc (nt * sizeof (int));
               alloc_chk (WHERE, p_sim->p_tloc);

/* Copy varying position tloc numbers to sim list.                            */
               ix = 0;
               p_vmap = first_tloc_vmap_node[sx];
               while (p_vmap)
               {
                  if (p_vmap->selectd)  /* Adds tloc numbers to sim list.     */
                  {
                     *(p_sim->p_tloc + ix) = p_vmap->tlno;
                     ++ix;
                  }
                  p_vmap = p_vmap->p_next;
               }
            }
         }
         else
         {
/* Did not find varying test positions for chromosome.                        */
            err_typ = ERD_MSM;
            on_chrom (p_vmap->chrom_no);
         }
      }
   }

   if (err_typ)
      return FALSE;
   else
      return TRUE;
}

/* Use any "map test tlocs ... at markers ..." statement information to       */
/* construct the implied interval map nodes for the specified vmap_node.      */

void updt_at_markers_vmap (SIM_PTR p_s, VMAP_PTR p_vmap)
{
   int   marker_num, x1;
   IMAP_PTR p_vint;

/* If a "map test tlocs ... at markers ..." statement is in effect, each      */
/* interval proportion needs to be set to 0.0.                                */

   if (p_vmap->no_marker_nums > 0)
   {

/* Since a "map test tlocs ... at markers ..." statement is in effect, any    */
/* other starting, ending or interval positions specified by the user need to */
/* be ignored.  This can be accomplished by freeing up any storage already    */
/* allocated for distance arrays in the beginning positions map, the ending   */
/* positions map and in any internal interval map.                            */
      if (p_vmap->p_beg || p_vmap->p_end || p_vmap->p_intv)
      {
         free_vdist (p_vmap->p_beg);
         free_vdist (p_vmap->p_end);
         p_vint = p_vmap->p_intv;
         while (p_vint)
         {
            free_vprop (p_vint);
            p_vint->no_prop = 0;
            p_vint = p_vint->p_next;
         }
         err_typ = ERA_OMT;
      }

/* Since a "map test tlocs ... at markers ..." statement is in effect, don't  */
/* use any default beginning or ending positions or interval proportions.     */
      vmap_opt.default_xt = FALSE;
      vmap_opt.default_nt = FALSE;

/* There will be only one positioning (either interior or ending) proportion  */
/* for each marker, so set nx_par to 1.  The alloc_vprop and alloc_vdist      */
/* subroutines refer to nx_par to calculate how much storage to allocate.     */
/* The stor_vprop and stor_vdist subroutines refer to nx_par and d_par[] to   */
/* store proportions, distances and recombination fractions (where needed).   */
      nx_par = 1;
      d_par[0] = 0.0;

/* For each at marker number, check for an existing array of proportions.     */
      for (x1 = 0; x1 < p_vmap->no_marker_nums; ++x1)
      {
         marker_num = p_vmap->p_marker_nums[x1];

         if (marker_num < p_s->no_mark_map)
         {
/* If the at marker number is less than the total number of markers mapped,   */
/* then an interior interval proportion can be used to choose evaluation at   */
/* the marker.  If there is not already an Nt_map node for this at marker     */
/* number, allocate one.                                                      */

            p_vint = p_vmap->p_intv;
            while (p_vint && p_vint->intv_no < marker_num)
               p_vint = p_vint->p_next;
            if (!p_vint || p_vint->intv_no != marker_num)
               p_vint = add_vint (p_vmap, marker_num);
            p_vint->no_prop = nx_par;

/* Allocate storage for a single double precision number to be used as the    */
/* interval proportion past the at marker number and set it equal to 0.0.     */
            alloc_vprop (p_vint);
            stor_vprop (p_vint);
         }

         else
         {
/* Otherwise, if the at marker number equals the total number of markers      */
/* mapped, then we have to use an ending positions map to choose evaluation   */
/* at the marker.  If there is not already an ending positions map, allocate  */
/* one.                                                                       */
            if (!p_vmap->p_end)
               p_vmap->p_end = alloc_vext ();

/* Allocate storage for the distance arrays within the ending positions map.  */
/* Then store the requisite double precision numbers to reflect a single      */
/* chosen evaluation point at the last marker.                                */
            alloc_vdist (p_vmap->p_end);
            stor_vdist (p_vmap->p_end, EDM, qfunc_doubdec);
         }
      }

/* Update the global variable which indicates whether or not lodscores are,   */
/* by default, to be calculated at all markers or not.  Since a "map test     */
/* tlocs ... at markers ..." statement is in effect, we need to suppress the  */
/* historically default behavior of calculating lodscores at unselected       */
/* markers.                                                                   */
      Lodscores_at_markers = N;
   }

   return;
}

/* Checks the map for test positions for the tloc.                            */
/* Tests for tloc map present for this chromosome and tloc number.            */
/* Checks sex-specific maps for consistency.                                  */

int test_tloc_vmap (SIM_PTR p_sel, int tl_no, int *ss_map)
{
   int   sx, x1;
   VMAP_PTR p_vmap;
   VMAP_PTR p_m[3] = { NULL, NULL, NULL };

   for (sx = FEM; sx <= AVG; ++sx)      /* Check for tloc map present.        */
      if ((p_vmap = first_tloc_vmap_node[sx]))
      {
         while (p_vmap->p_next && p_vmap->chrom_no < p_sel->chrom_no)
            p_vmap = p_vmap->p_next;
         if (p_vmap->chrom_no == p_sel->chrom_no)
         {
            while (p_vmap->p_next && p_vmap->tlno < tl_no)
               p_vmap = p_vmap->p_next;
            if (p_vmap->chrom_no == p_sel->chrom_no && p_vmap->tlno == tl_no)
            {
/* Update any necessary vmap_node interior or ending positions structures.    */
               updt_at_markers_vmap (p_sel, p_vmap);
               p_m[sx] = p_vmap;        /* If found, check for sex-spec maps. */
               if (sx != AVG)
                  *ss_map = TRUE;
            }
         }
      }

/* At least one test tloc vmap_node should have been found.  If none were     */
/* found for this chromosome and test tloc, terminate with an informative     */
/* error message.                                                             */

   if (!p_m[AVG] && !p_m[MAL] && !p_m[FEM]) /* Check for at least 1 external  */
   {                                    /*     or interval proportions map.   */
      err_typ = ERA_MXM;
      on_chrom (p_sel->chrom_no);
      sprintf (err_format, "%s%s", " %d ", errnumb);
      sprintf (errnumb, err_format, tl_no);
      err_txt = errnumb;
      return FALSE;
   }

/* We have to have at least one gender-unspecified test tloc vmap_node for    */
/* this chromosome and tloc number.  Make sure we have found at least one.    */

   if (!p_m[AVG])
   {
      err_typ = ERA_MBM;
      on_chrom (p_sel->chrom_no);
      sprintf (err_format, "%s%s", " %d ", errnumb);
      sprintf (errnumb, err_format, tl_no);
      err_txt = errnumb;
      return FALSE;
   }

/* Verify that the tloc number is always the same for all the selected        */
/* test tloc vmap_nodes.                                                      */

   for (sx = FEM; sx <= AVG; ++sx)
   {
      x1 = (sx + 1) % (AVG + 1);
      if (p_m[sx] && p_m[x1])           /* Only check when both exist.        */
      {
         if (p_m[sx]->tlno != p_m[x1]->tlno)
            err_typ = ERA_ITI;
      }
   }

   if (err_typ)
   {
      for_chrom (p_sel->chrom_no);
      return FALSE;
   }

   return TRUE;
}

/* Checks the map for beginning and ending test positions for the tloc.       */

int test_tloc_vext (SIM_PTR p_sel, int tl_no)
{
   int   sx, beg = FALSE, end = FALSE;
   VMAP_PTR p_vmap;
   VMAP_PTR p_m[3] = { NULL, NULL, NULL };

   for (sx = FEM; sx <= AVG; ++sx)      /* Check for tloc map present.        */
      if ((p_vmap = first_tloc_vmap_node[sx]))
      {
         while (p_vmap->p_next && p_vmap->chrom_no < p_sel->chrom_no)
            p_vmap = p_vmap->p_next;
         if (p_vmap->chrom_no == p_sel->chrom_no)
         {
            while (p_vmap->p_next && p_vmap->tlno < tl_no)
               p_vmap = p_vmap->p_next;
            if (p_vmap->chrom_no == p_sel->chrom_no && p_vmap->tlno == tl_no)
            {
               p_m[sx] = p_vmap;        /* If found, check for sex-spec maps  */
               if (p_m[sx]->p_beg)      /* for beginning or ending positions. */
                  beg = TRUE;
               if (p_m[sx]->p_end)
                  end = TRUE;
            }
         }
      }

/* Warn the user if no beginning nor ending test positions were specified.    */
   if (Lodscores_at_markers && (!beg || !end))
   {
      if (vmap_opt.default_xt)          /* Positions not specified.           */
         err_typ = ERA_DBE;
      else                              /* Check for no defaults.             */
         err_typ = ERA_NBE;
   }

   if (err_typ)
   {
      on_chrom (p_sel->chrom_no);
      sprintf (err_format, "%s%s", " %d ", errnumb);
      sprintf (errnumb, err_format, tl_no);
      err_txt = errnumb;
      return FALSE;
   }

   return TRUE;
}

/* Checks the maps for interval proportions.  Then checks that interval       */
/* numbers are within range (ie, less than the number of markers mapped).     */

int test_tloc_vint (SIM_PTR p_sel, int tl_no)
{
   int   sx, intv = FALSE;
   VMAP_PTR p_vmap;
   VMAP_PTR p_m[3] = { NULL, NULL, NULL };
   IMAP_PTR p_n;

   for (sx = FEM; sx <= AVG; ++sx)      /* Check for tloc map present.        */
      if ((p_vmap = first_tloc_vmap_node[sx]))
      {
         while (p_vmap->p_next && p_vmap->chrom_no < p_sel->chrom_no)
            p_vmap = p_vmap->p_next;
         if (p_vmap->chrom_no == p_sel->chrom_no)
         {
            while (p_vmap->p_next && p_vmap->tlno < tl_no)
               p_vmap = p_vmap->p_next;
            if (p_vmap->chrom_no == p_sel->chrom_no && p_vmap->tlno == tl_no)
            {
               p_m[sx] = p_vmap;
               if (p_m[sx]->p_intv)
                  intv = TRUE;
            }
         }
      }

   if (p_sel->no_mark_map == 1)         /* If only 1 marker,                  */
   {
      if (intv)                         /* Test for extraneous intervl props. */
         err_typ = ERA_XEI;
   }
   else if (p_sel->no_mark_map > 1)     /* If more than 1 marker,             */
      if (!intv)                        /* test for no interval proportions.  */
      {
         if (vmap_opt.default_nt)       /* Check for no defaults specified.   */
            err_typ = ERA_DEI;
         else
            err_typ = ERA_NDI;
      }

   if (intv && p_sel->no_mark_map > 1)  /* Test for valid interval numbers.   */
      for (sx = MAL; sx <= AVG; ++sx)
         if (p_m[sx] && (p_n = p_m[sx]->p_intv))
         {
            while (p_n && p_n->p_next)
               p_n = p_n->p_next;
            if (p_n->intv_no > p_sel->no_mark_map - 1)
               err_typ = ERA_MIO;
         }

   if (err_typ)
   {
      on_chrom (p_sel->chrom_no);
      sprintf (err_format, "%s%s", " %d ", errnumb);
      sprintf (errnumb, err_format, tl_no);
      err_txt = errnumb;
      return FALSE;
   }

   return TRUE;
}

/* Makes vprntmap node.                                                       */
/* Note: Autozyg programs use only one chromosome.                            */

PRNTVMAP_PTR add_vprntmap (VMAP_PTR p_vm, int sex)
{
   PRNTVMAP_PTR p_p;

   p_p = malloc (sizeof (PRNTVMAP));
   alloc_chk (WHERE, p_p);

   p_p->chrom_no = p_vm->chrom_no;
   p_p->no = p_vm->tlno;
   p_p->p_beg = NULL;
   p_p->p_end = NULL;
   p_p->p_intv = NULL;
   p_p->p_sintv = NULL;

   return p_p;
}

/**
 * makes a list of external trial tloc positions.
 * makes the vprntmap list for positions at the start or end of the chrom
 */
void add_tloc_vext (PRNTVMAP_PTR p_pv, int sex)
{
   XMAP_PTR p_x;
   PXMAP_PTR p_px;

   int   k;
   int   i;

   for (k = BDM; k <= EDM; ++k)         /* sets beginning pos, then ending    */
   {
      p_px = malloc (sizeof (PXMAP));
      alloc_chk (WHERE, p_px);
      p_px->no_pos = 0;
      p_px->p_dist = NULL;
      p_px->p_rf = NULL;

      if (k == BDM)
         p_pv->p_beg = p_px;
      else
         p_pv->p_end = p_px;

      if ((p_x = nxt_tloc_vext (sex, p_pv, k))) /* looks for beg or end map   */
      {
         p_px->no_pos = p_x->no_pos;    /* uses map from selected gender      */
         p_px->p_dist = malloc (p_x->no_pos * sizeof (double));
         alloc_chk (WHERE, p_px->p_dist);
         p_px->p_rf = malloc (p_x->no_pos * sizeof (double));
         alloc_chk (WHERE, p_px->p_rf);
         for (i = 0; i < p_x->no_pos; ++i)
         {
            *(p_px->p_dist + i) = *(p_x->p_dist + i);
            *(p_px->p_rf + i) = *(p_x->p_rf + i);
         }
      }
      else if (!vmap_opt.default_xt)
         p_px->no_pos = 0;              /* does not use canned defaults       */
      else
      {
         p_px->no_pos = NXtrnPos;       /* uses canned defaults for map       */
         p_px->p_dist = malloc (NXtrnPos * sizeof (double));
         alloc_chk (WHERE, p_px->p_dist);
         p_px->p_rf = malloc (NXtrnPos * sizeof (double));
         alloc_chk (WHERE, p_px->p_rf);
         for (i = 0; i < NXtrnPos; ++i)
            if (k == BDM)
            {
               *(p_px->p_rf + i) = *(XtrnPos + NXtrnPos - i - 1);
               *(p_px->p_dist + i) = rf2dist (*(XtrnPos + NXtrnPos - i - 1));
            }
            else if (k == EDM)
            {
               *(p_px->p_rf + i) = *(XtrnPos + i);
               *(p_px->p_dist + i) = rf2dist (*(XtrnPos + i));
            }
      }
   }

   return;
}

/**
 * makes a list of trial tloc positions for ea marker interval.
 */
void add_tloc_vint (PRNTVMAP_PTR p_pv, PRNTMAP_PTR p_pm, int sex)
{
   IMAP_PTR p_n;
   PIMAP_PTR p_i, p_j;
   PRDIST_PTR p_l;
   int   m;
   int   i;

   p_l = p_pm->p_locus;

   for (m = 1; m < p_pm->no_mark_map; ++m)
   {
      p_j = malloc (sizeof (PIMAP));    /* allocs and initializes the node    */
      alloc_chk (WHERE, p_j);

      p_j->intv_no = m;
      p_j->selectd = p_l->selectd;
      p_j->no_prop = 0;
      p_j->p_prop = NULL;
      p_j->p_next = NULL;

      if (!(p_n = nxt_tloc_vint (sex, m, p_pv))) /* looks for intrvl props    */
         p_n = nxt_tloc_vint (sex, 0, p_pv); /* for this interval or "all"    */
      if (p_n)                          /* intervals, using default genders   */
      {
         p_j->no_prop = p_n->no_prop;
         p_j->p_prop = malloc (p_n->no_prop * sizeof (double));
         alloc_chk (WHERE, p_j->p_prop);
         for (i = 0; i < p_n->no_prop; ++i)
            *(p_j->p_prop + i) = *(p_n->p_prop + i);
      }
      else if (!vmap_opt.default_nt)
         p_j->no_prop = 0;              /* does not use canned defaults       */
      else
      {
         p_j->no_prop = NIntProp;       /* uses canned defaults for map       */
         p_j->p_prop = malloc (NIntProp * sizeof (double));
         alloc_chk (WHERE, p_j->p_prop);
         for (i = 0; i < NIntProp; ++i)
            *(p_j->p_prop + i) = *(IntProp + i);
      }
      p_l = p_l->p_next;

      if (m == 1)
         p_pv->p_intv = p_j;
      else
         p_i->p_next = p_j;
      p_i = p_j;
   }
   return;
}

/**
 * builds a list of tloc positions for ea selected marker interval,
 * and for the start and end of the chromosome.
 * all tloc positions and non-selected markers are preserved; the
 * map positions are determined relative to the selected markers
 */
void sel_tloc_vpos (PRNTVMAP_PTR p_pv, PRNTMAP_PTR p_pm)
{
   PIMAP_PTR p_i, pi_sav;
   PSMAP_PTR p_new;
   PRDIST_PTR p_l;
   double s;
   int   k;
   int   i;
   int   j;

   p_pv->p_sintv = alloc_sintv (0);     /* allocs a node for ea selected      */
   p_new = p_pv->p_sintv;               /* interval, and the beginning & end  */

   for (i = 1; i <= p_pm->no_mark_sel; ++i)
   {
      p_new->p_next = alloc_sintv (i);
      p_new = p_new->p_next;
   }
   p_new->p_next = NULL;

   p_i = p_pv->p_intv;                  /* BEGINNING:  counts no. positions   */
   p_new = p_pv->p_sintv;               /* and allocates list                 */
   p_new->no_pos = p_pv->p_beg->no_pos;
   while (p_i && !p_i->selectd)
   {
      p_new->no_pos += p_i->no_prop;
      if (Lodscores_at_markers)
         ++p_new->no_pos;
      p_i = p_i->p_next;
   }
   p_i = p_pv->p_intv;
   p_l = p_pm->p_locus;

   alloc_sintl (p_new);

   s = p_l->sel_dist;                   /* sets positions before marker 1     */

   for (k = 0; k < p_pv->p_beg->no_pos; ++k)
   {
      *(p_new->p_sdist + k) = *(p_pv->p_beg->p_dist + k) + s;
      *(p_new->p_rf + k) = dist2rf (*(p_new->p_sdist + k));
      *(p_new->p_star + k) = FALSE;
      if (k < p_pv->p_beg->no_pos - 1)
         *(p_new->p_locdist + k) = *(p_pv->p_beg->p_dist + k)
            - *(p_pv->p_beg->p_dist + k + 1);
      else
         *(p_new->p_locdist + k) = *(p_pv->p_beg->p_dist + k);
   }

   p_i = p_pv->p_intv;                  /* for ea proportion list before 1st  */
   while (p_i && !p_i->selectd)         /* sel marker, sets pos to marker pos */
   {                                    /* and to interval positions.         */
      if (Lodscores_at_markers)
      {
         *(p_new->p_sdist + k) = s;     /* Set unsel marker derived test pos. */
         *(p_new->p_rf + k) = dist2rf (*(p_new->p_sdist + k));
         *(p_new->p_star + k) = TRUE;
         ++k;
      }

      s = p_l->p_next->sel_dist;

      for (j = 0; j < p_i->no_prop; ++j, ++k) /* now set intv prop test pos   */
      {
         *(p_new->p_sdist + k) = s + (1. - *(p_i->p_prop + j))
            * p_l->p_next->dist;
         *(p_new->p_rf + k) = dist2rf (*(p_new->p_sdist + k));
         if (k > 0)
            *(p_new->p_locdist + k - 1) = *(p_new->p_sdist + k - 1)
               - *(p_new->p_sdist + k);
         *(p_new->p_star + k) = FALSE;
      }
      *(p_new->p_locdist + k - 1) = *(p_new->p_sdist + k - 1) - s;

      p_i = p_i->p_next;
      p_l = p_l->p_next;
   }
   p_new->sel_map_mark = p_l->no;

   for (i = 1; i < p_pm->no_mark_sel; ++i) /* INTERVALS: for ea interval      */
   {                                    /* between selected markers, counts   */
      pi_sav = p_i;                     /* no. positions and allocs list      */
      p_new = p_new->p_next;
      p_new->no_pos = p_i->no_prop;
      while ((p_i = p_i->p_next) && !p_i->selectd)
      {
         p_new->no_pos += p_i->no_prop;
         if (Lodscores_at_markers)
            ++p_new->no_pos;
      }
      p_new->sel_map_mark = p_l->no;

      alloc_sintl (p_new);

      p_i = pi_sav;
      s = 0.;
      k = 0;

      do                                /* for this prop list, sets test pos  */
      {
         for (j = 0; j < p_i->no_prop; ++j, ++k)
         {
            *(p_new->p_sdist + k) =
               s + *(p_i->p_prop + j) * (p_l->p_next->dist);
            *(p_new->p_rf + k) = dist2rf (*(p_new->p_sdist + k));
            if (Lodscores_at_markers || p_i->selectd)
            {
               if (j == 0)
                  *(p_new->p_locdist + k) =
                     *(p_i->p_prop) * (p_l->p_next->dist);
               else
                  *(p_new->p_locdist + k) = *(p_new->p_sdist + k)
                     - *(p_new->p_sdist + k - 1);
            }
            *(p_new->p_star + k) = FALSE;
         }
         p_l = p_l->p_next;
         p_i = p_i->p_next;
         s = p_l->sel_dist;
         if (p_new->no_pos)
         {
            if (k > 0)
               *(p_new->p_locdist + k) = s - *(p_new->p_sdist + k - 1);
            else
               *p_new->p_locdist = s;

/* Sets distances for unselected marker derived test positions when lodscores */
/* are to be calculated at these positions by default, rather than when a     */
/* "map test tlocs ... at markers ..." statement was used.                    */
            if (Lodscores_at_markers && p_i && !p_i->selectd)
            {
               *(p_new->p_sdist + k) = s;
               *(p_new->p_rf + k) = dist2rf (*(p_new->p_sdist + k));
               *(p_new->p_star + k) = TRUE;
               ++k;
            }
         }
      }
      while (p_i && !p_i->selectd);
   }

   p_new = p_new->p_next;               /* ENDING: counts no. positions       */
   pi_sav = p_i;                        /* and allocates list                 */
   p_new->no_pos = 0;

   if (p_l->no < p_pm->no_mark_map)
      while (p_i)
      {
         p_new->no_pos += p_i->no_prop;
         if (Lodscores_at_markers)
            ++p_new->no_pos;
         p_i = p_i->p_next;
      }
   p_new->no_pos += p_pv->p_end->no_pos;
   p_new->sel_map_mark = p_l->no;

   alloc_sintl (p_new);

   p_i = pi_sav;
   k = 0;
   s = 0.;
   while (p_i)                          /* for ea remaining prop list between */
   {                                    /* non-selectd markers,               */
      for (j = 0; j < p_i->no_prop; ++j, ++k) /* first sets test positions    */
      {
         if (j < p_i->no_prop)
         {
            *(p_new->p_sdist + k) = s + *(p_i->p_prop + j) * p_l->p_next->dist;
            *(p_new->p_rf + k) = dist2rf (*(p_new->p_sdist + k));
         }
         else
         {
            s = p_l->p_next->sel_dist;
            *(p_new->p_sdist + k) = s;
            *(p_new->p_rf + k) = dist2rf (*(p_new->p_sdist + k));
            *(p_new->p_star + k + 1) = TRUE;
         }
         if (k == 0)
            *(p_new->p_locdist + k) = *(p_new->p_sdist + k);
         else
            *(p_new->p_locdist + k) = *(p_new->p_sdist + k)
               - *(p_new->p_sdist + k - 1);
         *(p_new->p_star + k) = FALSE;
      }
      s = p_l->p_next->sel_dist;

      if (Lodscores_at_markers)
      {
         *(p_new->p_sdist + k) = s;     /* Then sets marker derived test pos. */
         *(p_new->p_rf + k) = dist2rf (*(p_new->p_sdist + k));
         *(p_new->p_locdist + k) = *(p_new->p_sdist + k)
            - *(p_new->p_sdist + k - 1);
         *(p_new->p_star + k) = TRUE;
         ++k;
      }

      p_l = p_l->p_next;
      p_i = p_i->p_next;
   }

   for (j = 0; j < p_pv->p_end->no_pos; ++j, ++k) /* sets pos after last mark */
   {
      *(p_new->p_sdist + k) = s + *(p_pv->p_end->p_dist + j);
      *(p_new->p_rf + k) = dist2rf (*(p_new->p_sdist + k));
      if (j == 0)
         *(p_new->p_locdist + k) = *(p_pv->p_end->p_dist);
      else
         *(p_new->p_locdist + k) = *(p_pv->p_end->p_dist + j)
            - *(p_pv->p_end->p_dist + j - 1);
      *(p_new->p_star + k) = FALSE;
   }
   return;
}

PSMAP_PTR alloc_sintv (int intv_no)
{
   PSMAP_PTR p;

   p = malloc (sizeof (PSMAP));
   alloc_chk (WHERE, p);
   p->intv_no = intv_no;
   p->p_sdist = NULL;
   p->p_rf = NULL;
   p->p_locdist = NULL;
   p->p_star = NULL;
   p->p_next = NULL;

   return p;
}

void alloc_sintl (PSMAP_PTR p_si)
{

   if (!p_si->no_pos)
      return;

   p_si->p_sdist = malloc (p_si->no_pos * sizeof (double));
   p_si->p_rf = malloc (p_si->no_pos * sizeof (double));
   p_si->p_locdist = malloc ((p_si->no_pos + 1) * sizeof (double)); /*  + 1   */
   p_si->p_star = malloc ((p_si->no_pos + 1) * sizeof (int)); /*        + 1   */

   return;
}

/* This subroutine checks that the number of test tloc positions for the tloc */
/* represented by the p_vmap argument are consistent between female and male  */
/* beginning and ending px_map (external) elements.                           */

int test_tloc_vext_sx (VMAP_PTR p_vmap, SIM_PTR p_sel)
{
   if (p_vmap->p_vprntmap[FEM]->p_beg->no_pos !=
       p_vmap->p_vprntmap[MAL]->p_beg->no_pos)
      err_typ = ERA_WDX;
   else if (p_vmap->p_vprntmap[FEM]->p_end->no_pos !=
            p_vmap->p_vprntmap[MAL]->p_end->no_pos)
      err_typ = ERA_WDX;

   if (err_typ)
   {
      for_chrom (p_sel->chrom_no);
      return FALSE;
   }

   return TRUE;
}

/* This subroutine checks intra-interval tloc positions in all the marker     */
/* intervals for consistency between female and male pi_map_list proportions. */

int test_tloc_vint_sx (VMAP_PTR p_vmap, SIM_PTR p_sel)
{
   PIMAP_PTR p_f, p_m;

   p_f = p_vmap->p_vprntmap[FEM]->p_intv;
   p_m = p_vmap->p_vprntmap[MAL]->p_intv;

   while (p_f)
   {
      if (p_f->no_prop != p_m->no_prop)
         err_typ = ERA_WIP;
      p_f = p_f->p_next;
      p_m = p_m->p_next;
   }

   if (err_typ)
   {
      for_chrom (p_sel->chrom_no);
      return FALSE;
   }

   return TRUE;
}

/* Checks the rules for the external and interior interval test positions for */
/* specific Lodscore programs.                                                */

int test_vpos_rules (VMAP_PTR p_vmap)
{

/* The 3 local integer variables used within test_vpos_rules are:             */
/*    it   is the number of intervals between selected markers.               */
/*    nx   is the number of beginning or ending test position arrays found.   */
/*    ni   is the number of interior intervals with test position arrays.     */

   PSMAP_PTR p_si;
   int   it, nx, ni;

   it = 0;                              /* Set the 3 integer variables to 0.  */
   nx = 0;
   ni = 0;

/* By the time test_vpos_rules is called proc_lods has already made sure that */
/* all 3 gender linked lists representing the varying positions print maps,   */
/* p_vmap->p_vprntmap[], have identically the same number of elements in the  */
/* list and that the values set for the no_pos integer variables in           */
/* corresponding list elements are all equal.  As a result it does not make   */
/* any difference which of the 3 gender linked lists is used in the following */
/* statement.                                                                 */

   p_si = p_vmap->p_vprntmap[AVG]->p_sintv;

   if (p_si->no_pos > 0)                /* Checks for beginning positions.    */
      ++nx;

   while (p_si->p_next->p_next)         /* Checks the interior intervals.     */
   {
      p_si = p_si->p_next;
      ++it;
      if (p_si->no_pos > 0)
         ++ni;
   }

   if (p_si->p_next->no_pos > 0)        /* Checks for ending positions.       */
      ++nx;

/* By the time test_vpos_rules is called most Lodscore programs require that  */
/* there should have been at least one beginning, ending or interior interval */
/* with a test position list.  So (nx + ni) should no longer be equal to 0.   */
/* For lm_lods and lm_schnell, terminate with a fatal error message.          */

   if (LM_LODS || LM_SCHNELL)           /* Tests for no test positions.       */
      if ((nx + ni) == 0)
         err_typ = ERA_TPM;

/* Test for whether or not all interior intervals have proportions given.     */
/* For these 3 programs it is an error if not all interior intervals do not   */
/* have proportions given.                                                    */

   if (LM_BAYES || LM_LODS || LM_SCHNELL)
      if (ni < it)
         err_typ = ERA_IPM;

/* By the time test_vpos_rules is called most Lodscore programs require that  */
/* there should have been at least one beginning, ending or interior interval */
/* with a test position list.  So (nx + ni) should no longer be equal to 0.   */
/* For lm_markers, as well as lm_multiple and any other program running as    */
/* lm_markers, if the TraitTyp is genotypic, terminate with a fatal error.    */
/* But when these programs are using any other type of trait, just produce a  */
/* warning and continue running.                                              */

   if (LM_MARKERS)                      /* Tests for no test positions.       */
      if ((nx + ni) == 0)
      {
         if (TraitTyp == GENOTYPIC)
            err_typ = ERA_GMP;
         else
            err_typ = ERA_NMP;
      }

   if (err_typ)
   {
      err_txt = ProgNam;
      return FALSE;
   }

   return TRUE;
}

/**
 * allocates global arrays for trial tloc positions.
 *    IntVal:   no. tloc pos + 2
 *    RecVal:   2 sexes, no. tloc pos + 2,  recomb fracts
 *    NTPosInt: no. intervals (no. markers + 1, for beginning,
 *                             between markers and ending)
 */
void alloc_intval (void)
{
   int   sex;

   RecVal = malloc (2 * sizeof (double *));
   alloc_chk (WHERE, RecVal);

   for (sex = FEM; sex <= MAL; ++sex)
   {
      RecVal[sex] = malloc ((NTPos + 2) * sizeof (double));
      alloc_chk (WHERE, RecVal[sex]);
   }
   IntVal = malloc ((NTPos + 2) * sizeof (int));
   alloc_chk (WHERE, IntVal);

   NTPosInt = malloc ((NMarks + 1) * sizeof (int));
   alloc_chk (WHERE, NTPosInt);

   return;
}

/* Stores marker and varying test tloc positions.                             */
/*    NLocPos - total number of marker and test tloc positions.               */
/*    LocPos  - data structure containing locus type, interval number,        */
/*              marker or tloc number, recombination fraction, and            */
/*              distance from marker 1.                                       */

void mak_locpos (VMAP_PTR p_vmap)
{
   PSMAP_PTR p_tr[2];
   PRDIST_PTR p_mk[2];
   LOCPOS_PTR p_l;

   double dm[2];                        /* Distance to current selectd marker */
   int   sex;
   int   tl_no;
   int   entry;
   int   x1;

   NLocPos = NTPos + NMarks;

   p_l = malloc (NLocPos * sizeof (LOCPOS));
   alloc_chk (WHERE, p_l);
   LocPos = p_l;

   for (sex = FEM; sex <= MAL; ++sex)
   {
      p_tr[sex] = p_vmap->p_vprntmap[sex]->p_sintv;
      entry = FALSE;
      p_mk[sex] = nxt_prntmap (&entry, sex)->p_locus;
      dm[sex] = p_mk[sex]->m1_dist;
   }
   tl_no = p_vmap->p_vprntmap[FEM]->no;

   while (p_tr[FEM])
   {

/* For each TEST position, either interval or exterior, do the following:     */
      for (x1 = 0; x1 < p_tr[FEM]->no_pos; ++x1)
      {
         p_l->intv_no = p_tr[FEM]->intv_no; /*For this intv, ea test pos, sex */
         p_l->no = tl_no;               /* Stores recomb fracts and positions */

         if (*(p_tr[FEM]->p_star + x1)) /* Checks for tloc or unselctd marker */
            p_l->locus_typ = UNSELMRK_LOC;
         else
            p_l->locus_typ = TLOC_LOC;

         for (sex = FEM; sex <= MAL; ++sex)
         {
            if (p_l->intv_no == 0)
            {
               p_l->rf[sex] = .5;
               p_l->rfnxt[sex] = *(p_tr[sex]->p_rf + x1);
               p_l->pos[sex] = dm[sex] - *(p_tr[sex]->p_sdist + x1);
            }
            else
            {
               p_l->rf[sex] = *(p_tr[sex]->p_rf + x1);
               if (p_mk[FEM])
                  p_l->rfnxt[sex] =
                     ((dist2rf (p_mk[sex]->sel_dist)) - p_l->rf[sex])
                     / (1.0 - 2. * p_l->rf[sex]);
               else
                  p_l->rfnxt[sex] = .5;
               p_l->pos[sex] = dm[sex] + *(p_tr[sex]->p_sdist + x1);
            }
         }
         ++p_l;
      }
      for (sex = FEM; sex <= MAL; ++sex)
         p_tr[sex] = p_tr[sex]->p_next;

      if (p_mk[FEM])
      {
         p_l->locus_typ = MARK_LOC;     /* Selected MARKER postion.           */
         p_l->intv_no = p_mk[FEM]->selectd;
         p_l->no = p_mk[FEM]->no;       /* Stores recomb fract & pos, ea sex. */
         for (sex = FEM; sex <= MAL; ++sex)

         {
            if (p_mk[sex]->selectd == 1) /* Tests for first selected marker.  */
            {
               p_l->rf[sex] = .5;
               p_l->pos[sex] = dm[sex];
            }
            else
            {
               p_l->rf[sex] = dist2rf (p_mk[sex]->sel_dist);
               dm[sex] += p_mk[sex]->sel_dist;
               p_l->pos[sex] = dm[sex];
            }
            if (p_mk[sex]->p_next)      /*Tests for not last selected marker. */
               p_l->rfnxt[sex] = dist2rf (p_mk[sex]->p_next->sel_dist);
            else
               p_l->rfnxt[sex] = .5;
            p_mk[sex] = p_mk[sex]->p_next;
         }
         ++p_l;
      }
   }

   return;
}

/* Stores trial tloc positions.                                               */
/*    NTPosInt[] - list of number of tloc positions in each marker interval.  */
/*    IntVal[]   - the interval number corresponding to each tloc position.   */

void mak_intval (VMAP_PTR p_vmap)
{
   PSMAP_PTR p_s;
   int   ix;
   int   i1;
   int   i2;

   p_s = p_vmap->p_vprntmap[FEM]->p_sintv;
   IntVal[0] = 0;
   ix = 1;

   for (i1 = 0; i1 <= NMarks && p_s; ++i1)
   {
      NTPosInt[i1] = p_s->no_pos;       /* Sets no. positions in ea interval. */
      for (i2 = 0; i2 < p_s->no_pos; ++i2, ++ix)
         IntVal[ix] = i1;               /*Sets interval for ea tloc position. */
      p_s = p_s->p_next;
   }
   IntVal[NTPos + 1] = 0;

   return;
}

/* Stores trial tloc recombinaton fractions.                                  */
/* Makes RecVal[][] - sex-specific recombination fracts corresp to IntVal[].  */

void mak_recval (VMAP_PTR p_vmap)
{
   PSMAP_PTR p_s;
   int   sex;
   int   ix;
   int   i1;
   int   i2;

   for (sex = FEM; sex <= MAL; ++sex)
   {
      p_s = p_vmap->p_vprntmap[sex]->p_sintv;
      RecVal[sex][0] = .5;              /* Sets rec fracts for start of chrom */
      ix = 1;

      for (i1 = 0; i1 <= NMarks && p_s; ++i1) /*and for each marker interval, */
      {
         for (i2 = 0; i2 < p_s->no_pos; ++i2, ++ix)
            RecVal[sex][ix] = *(p_s->p_rf + i2);
         p_s = p_s->p_next;
      }
      RecVal[sex][NTPos + 1] = .5;      /* and for end of chromosome.         */
   }

   return;
}

/* This subroutine copies the 9 joint genotype means for the 2 interacting    */
/* selected tlocs from among genotype means collected by the parser to the    */
/* global variable, GxGMeans, used by Lodscore programs, such as lm_twoqtl.   */
/* The argument to this subroutine is the trait_tloc_node selected using a    */
/* "set trait K1 multiple tlocs L1 L2" statement.                             */

void copy_quant_means (TRAITTLOC_PTR p_tt)
{
   GENO_MEANS_PTR p_means;
   int   i1, i2;
   int   x1, x2;

/* Allocate storage for a 3x3 matrix of double precision mean values for 2    */
/* interacting tlocs.                                                         */

   GxGMeans = alloc_dbl2d (3, 3);

   p_means = first_geno_means_node;

   while (p_means)
   {
      if (p_means->trait_no == p_tt->trait_no)
      {
         for (i1 = 0; i1 < p_tt->no_loci; ++i1)
         {
            if (p_means->tloc_no == p_tt->p_tloc[i1])
            {
               if (p_means->joint && p_means->num_means == 9)
               {
                  i2 = 0;
                  for (x1 = 0; x1 < 3; ++x1)
                  {
                     for (x2 = 0; x2 < 3; ++x2)
                     {
                        GxGMeans[x1][x2] = p_means->means[i2];
                        ++i2;
                     }
                  }
               }
            }
         }
      }
      p_means = p_means->p_next;
   }

   return;
}

void prnt_lods_pars (void)
{

   if (SmplSetup == INDEP_LOC)
      fprintf (stdout, " Locus-by-locus sampling to be used for setup\n\n");
   else if (SmplSetup == SEQIMP)
      fprintf (stdout, " Sequential imputation to be used for setup,"
               " using %d samples\n\n", NSeqImp);
   if (PeelCmpnt)
      fprintf (stdout, " Peeling is to be by pedigree component\n\n");

   if (GlobalMcmc)
      fprintf (stdout,
               " Lod scores for all components are estimated globally\n\n");

   if (Scan)
   {
      fprintf (stdout, " Sampling is to be by scan");
      if (ScorIt == UNDEF)
         fprintf (stdout, "\n");
      else if (ScorIt == 1)
         fprintf (stdout, ", with scores estimated every scan\n");
      else
         fprintf (stdout, ", with scores estimated every %d scans\n", ScorIt);
   }
   else
   {
      fprintf (stdout, " Sampling is to be by single step");
      if (ScorIt == UNDEF)
         fprintf (stdout, "\n");
      if (ScorIt == 1)
         fprintf (stdout, ", with scores estimated every step\n");
      else
         fprintf (stdout, ", with scores estimated every %d steps\n", ScorIt);
   }

   if (LM_BAYES && SeqImpIt)
      fprintf (stdout, "         and a restart configuration proposed"
               " by sequential imputation\n"
               "         every %d iterations\n\n", SeqImpIt);
   else
      fprintf (stdout, "\n");

   fprintf (stdout, " L-sampler probability set to %.3f\n\n", LSamplProb);

   if (NBrnIt != UNDEF)
   {
      fprintf (stdout, " The number of MC burn-in iterations is %d\n", NBrnIt);
      if (LM_BAYES)
      {
         if (!InPseudo)
            fprintf (stdout,
                     " and the number of pseudo-prior iterations is"
                     " %d\n", NPseudoPrIt);
         else
            fprintf (stdout, "\n Preliminary MCMC iterations to compute"
                     " pseudo-priors are skipped,\n"
                     " as pseudo-priors are input\n");
      }
      fprintf (stdout, "\n");
   }

   if (NIt != UNDEF)
   {
      fprintf (stdout, " Number of MC iterations requested is %d\n", NIt);
      if (ChkIt != UNDEF)
         fprintf (stdout, " with progress checked every %d iterations\n",
                  ChkIt);
   }

   if (LM_MARKERS)
      fprintf (stdout,
               "\n Number of batches for MC variance estimation is %d\n",
               NBtch);

   fprintf (stdout, "\n");

   return;
}

/* Prints a map showing the varying test tloc positions for the tloc          */
/* indicated by the vmap_node passed as the first argument.  This subroutine  */
/* produces 5 (6, if sex-specific maps) print lines for each chromosome       */
/* segment (where a segment consists of as many loci as will fit on the line) */
/* showing the tloc * positions at the start of the chromosome, in each       */
/* interval between selected markers, and at the end of the chromosome.       */

void prnt_vmap (VMAP_PTR p_vmap, int nmarks, int sxsp_mp)
{
   PRNTVMAP_PTR p_vpmap;
   PSMAP_PTR p_si[3];
   double *p_d[3];
   int  *p_f;
   int   sex;
   int   pos_ctr;
   int   no_loci;
   int   i1;

   for (sex = FEM; sex <= AVG; ++sex)
   {
      if ((p_vpmap = p_vmap->p_vprntmap[sex]))
      {
         p_si[sex] = p_vpmap->p_sintv;
         p_d[sex] = p_si[sex]->p_locdist;
      }
      else
         return;                        /* Unexpected.                        */
   }

   /*                                    * First prints the heading.          */
   if (PrintLevel >= 4)
      fprintf (stdout, "\n T%d test positions (cM):\n\n", p_vmap->tlno);

   p_f = p_si[AVG]->p_star;

   /*                                    * Prints out the BEGINNING positions */
   if (p_si[AVG]->no_pos)               /* (before first selected marker).    */
   {
      prntvmap_tl0 (*p_f);              /* Prints first T.                    */
      pos_ctr = 0;

      while (pos_ctr < p_si[AVG]->no_pos)
      {
         no_loci = 1;

         while (pos_ctr < p_si[AVG]->no_pos && no_loci < max_line_loc)
         {                              /* Adds next locus for this line.     */
            if (pos_ctr < p_si[AVG]->no_pos - 1)
               prntvmap_tl (no_loci, *(p_f + 1)); /* Next T,                  */
            else
               prntmap_mrk (no_loci, p_si[AVG]->sel_map_mark); /* or next M.  */
            for (sex = FEM; sex <= AVG; ++sex)
            {
               if ((sxsp_mp && (sex == FEM || sex == MAL))
                   || (!sxsp_mp && sex == AVG))
                  prntmap_dist (sex, no_loci, *p_d[sex]); /* dist- AVG F or M */
               ++p_d[sex];
            }
            ++p_f;
            ++pos_ctr;
            ++no_loci;
         }
         prntmap_line (no_loci, sxsp_mp); /* Print line for this map segment. */
         prntmap_tl0b ();               /* Initializes next segment.          */
      }
   }

   for (i1 = 1; i1 < nmarks; ++i1)      /* Now prints INTERVAL test positions */
   {                                    /* (between selected markers).        */
      pos_ctr = 0;
      for (sex = FEM; sex <= AVG; ++sex)
      {
         p_si[sex] = p_si[sex]->p_next;
         p_d[sex] = p_si[sex]->p_locdist;
      }
      p_f = p_si[AVG]->p_star;
      prntmap_mrk0 (p_si[AVG]->sel_map_mark); /* next M                       */

      while (pos_ctr < p_si[AVG]->no_pos)
      {
         no_loci = 1;

         while (pos_ctr < p_si[AVG]->no_pos && no_loci < max_line_loc)
         {                              /* Adds next locus for this line.     */
            if (pos_ctr < p_si[AVG]->no_pos)
               prntvmap_tl (no_loci, *p_f); /* Next T.                        */
            for (sex = FEM; sex <= AVG; ++sex)
            {
               if ((sxsp_mp && (sex == FEM || sex == MAL))
                   || (!sxsp_mp && sex == AVG))
                  prntmap_dist (sex, no_loci, *p_d[sex]); /* dist- AVG F or M */
               ++p_d[sex];
            }
            ++p_f;
            ++pos_ctr;
            ++no_loci;
         }
         if (pos_ctr == p_si[AVG]->no_pos)
         {
            prntmap_mrk (no_loci, p_si[AVG]->p_next->sel_map_mark); /* Next M */
            for (sex = FEM; sex <= AVG; ++sex)
               if ((sxsp_mp && (sex == FEM || sex == MAL))
                   || (!sxsp_mp && sex == AVG))
                  prntmap_dist (sex, no_loci, *p_d[sex]); /* dist- AVG F or M */
            ++no_loci;
         }
         prntmap_line (no_loci, sxsp_mp); /* Print line for this map segment. */
         prntmap_tl0b ();               /* Initializes next segment.          */
      }
   }

   for (sex = FEM; sex <= AVG; ++sex)
   {
      p_si[sex] = p_si[sex]->p_next;
      p_d[sex] = p_si[sex]->p_locdist;
   }
   p_f = p_si[AVG]->p_star;

   if (p_si[AVG]->no_pos)               /* Now print out the ENDING positions */
   {                                    /* (after last selected marker).      */
      pos_ctr = 0;
      prntmap_mrk0 (p_si[AVG]->sel_map_mark); /* Last M.                      */

      while (pos_ctr < p_si[AVG]->no_pos)
      {
         no_loci = 1;

         while (pos_ctr < p_si[AVG]->no_pos && no_loci < max_line_loc)
         {                              /* Adds next locus for this line.     */
            prntvmap_tl (no_loci, *p_f); /* Next T.                           */
            for (sex = FEM; sex <= AVG; ++sex)
            {
               if ((sxsp_mp && (sex == FEM || sex == MAL))
                   || (!sxsp_mp && sex == AVG))
                  prntmap_dist (sex, no_loci, *p_d[sex]); /* dist- AVG F or M */
               ++p_d[sex];
            }
            ++p_f;
            ++pos_ctr;
            ++no_loci;
         }
         prntmap_line (no_loci, sxsp_mp); /* Print line for this map segment. */
         prntmap_tl0b ();               /* Initializes next segment.          */
      }
   }

   return;
}

/**
 * Looks for chromosome maps, or unlinked spec, for the requested tlocs
 * (tlocs in the trait/tloc list which are marked as selected).
 * Marks these tlocs in the test position maps as selected.
 * Checks for just one location for each tloc:
 *   if a tloc does not have a map, it is an error.
 *   if a tloc is mapped on more than one chromosome, it is an error.
 *
 *      for ea trait
          for ea sel trait
            for ea tloc for the trait
              for ea gender
                for ea chrom tloc test pos map
                  for ea mapd tloc in list
                    tloc nos match?
                      set mapd tloc as selctd
                      is found mapd tloc flag set?
                        n - set found mapd tloc flag
                            set chrom no. found to map chrom no.
                        y - is chrom no. found = map chrom no?
                            n - error         (mapd on more than 1 chrom)
            if !found (map not present)
              error
*/
int find_tr_tloc_vmap (void)
{
   TRAITTLOC_PTR p_tt;
   int   entry;
   int   tl_no;
   int   x1;

   entry = FALSE;                       /* For each trait/tloc in the list,   */
   while ((p_tt = nxt_tr_tl (&entry)) && !err_typ)
   {
      if (p_tt->selectd)                /* Checks trait for selectd.          */
      {
         for (x1 = 0; x1 < p_tt->no_loci; ++x1)
         {
            tl_no = *(p_tt->p_tloc + x1); /* For each tloc for the trait,     */
            find_nxt_tl_vmap (tl_no);   /* Look for a vmap for the tloc.      */
         }
      }
   }

   if (err_typ)
      return FALSE;
   else
      return TRUE;
}

/* Check that the specified test tloc is not mapped on more than one          */
/* chromosome.                                                                */

void find_nxt_tl_vmap (int tl_no)
{
   VMAP_PTR p_vmap;
   int   chrom_no;
   int   found;
   int   sex;

   chrom_no = 0;
   found = FALSE;

   for (sex = FEM; sex <= AVG; ++sex)   /* For each possible gender list,     */
   {
      p_vmap = first_tloc_vmap_node[sex];

      while (p_vmap && !err_typ)        /* For each tloc map,                 */
      {

/* Does this vmap include a reference to the specified tloc number?           */
         if (p_vmap->tlno == tl_no)
         {
/* If this is the specified tloc number, set the vmap flag indicating that    */
/* this vmap is selected.  If a previous vmap for this tloc has not already   */
/* been found, save the chromosome number and set a local variable reflecting */
/* the fact that a vmap for the specified tloc number was found.              */
            p_vmap->selectd = TRUE;
            if (!found)
            {
               chrom_no = p_vmap->chrom_no;
               found = TRUE;
            }
            else if (chrom_no != p_vmap->chrom_no)
/* If this tloc number is mapped on more than one chromosome, cause an error  */
/* message to be given to the user, before terminating the program.           */
            {
               err_typ = ERD_MCT;
               sprintf (errnumb, "%d", tl_no);
               err_txt = errnumb;
            }
         }
         p_vmap = p_vmap->p_next;
      }
   }

   return;
}

/**
 * Looks for a test joint position map for the specified tloc.
 */
#if 0
int find_nxt_tl_jmap (int tlno)
{
   int   chrom_no, found, sex, i1;
   JMAP_PTR p_jmap;

   chrom_no = 0;
   found = FALSE;

   for (sex = FEM; sex <= AVG; ++sex)   /* For each gender,                   */
   {
      p_jmap = first_tloc_jmap_node[sex];

      while (p_jmap && !err_typ)        /* For each tloc map,                 */
      {
         for (i1 = 0; i1 < 2; ++i1)     /* Check both tloc numbers.           */
         {
            if (p_jmap->no[i1] == tlno) /* Found tloc number?                 */
            {
               p_jmap->selectd[i1] = TRUE;
               if (!found)
               {
                  found = TRUE;
                  chrom_no = p_jmap->chrom_no;
               }
               else if (chrom_no != p_jmap->chrom_no)
               {                        /* Checks for tloc on > 1 chromosome. */
                  err_typ = ERD_MCT;
                  sprintf (errnumb, "%d", tlno);
                  err_txt = errnumb;
               }
            }
         }
         p_jmap = p_jmap->p_next;
      }
   }

   return found;
}
#endif

/* For this chromosome and gender, puts test tloc positions into "print map". */
/* Maps not present default to sex-averaged map, then F map.                  */

int mak_tlocvprmap (SIM_PTR p_s, PRNTMAP_PTR p_m, int sex)
{
   VMAP_PTR p_vmap_cur, p_vmap_tmp;
   PRNTVMAP_PTR p_vpmap;
   int   entry1, entry2;

   entry1 = FALSE;                      /* Uses default rules for gender.     */
   while ((p_vmap_cur = nxt_tl_vmap (&entry1, sex, p_s)))
      if (p_vmap_cur->selectd)
      {
         p_vpmap = add_vprntmap (p_vmap_cur, sex); /* Insert tloc positions,  */
         add_tloc_vext (p_vpmap, sex);  /* for start and end of chromosome.   */
         add_tloc_vint (p_vpmap, p_m, sex); /* The same for marker intervals. */
         sel_tloc_vpos (p_vpmap, p_m);  /* Now condense this tloc map.        */
         p_vmap_cur->p_vprntmap[sex] = p_vpmap; /* Save pointer to the map.   */

/* The remaining tests performed in this subroutine are only executed the     */
/* final time this subroutine is called for each selected tloc vmap_node.     */
         if (sex == AVG)
         {
            entry2 = FALSE;             /* Uses default rules for gender.     */
            while ((p_vmap_tmp = nxt_tl_vmap (&entry2, FEM, p_s))
                   && p_vmap_tmp->tlno < p_vmap_cur->tlno);
            if (p_vmap_tmp && p_vmap_tmp->tlno == p_vmap_cur->tlno)
               p_vmap_cur->p_vprntmap[FEM] = p_vmap_tmp->p_vprntmap[FEM];

            entry2 = FALSE;             /* Uses default rules for gender.     */
            while ((p_vmap_tmp = nxt_tl_vmap (&entry2, MAL, p_s))
                   && p_vmap_tmp->tlno < p_vmap_cur->tlno);
            if (p_vmap_tmp && p_vmap_tmp->tlno == p_vmap_cur->tlno)
               p_vmap_cur->p_vprntmap[MAL] = p_vmap_tmp->p_vprntmap[MAL];

/* Tests for the same number of female and male positions at the start and    */
/* the end of the chromosome.  Then do the same testing within each interval. */
/* Lastly, call test_vpos_rules to make sure that the requisite number of     */
/* intervals and external map locations have test positions.                  */
            if (!test_tloc_vext_sx (p_vmap_cur, p_s))
               if (proc_err ())
                  return FALSE;
            if (!test_tloc_vint_sx (p_vmap_cur, p_s))
               if (proc_err ())
                  return FALSE;
            if (!test_vpos_rules (p_vmap_cur))
               if (proc_err ())
                  return FALSE;
         }
      }

   return TRUE;
}

/* Prints allele frequencies for the test tloc positions for the specified    */
/* chromosome.                                                                */

void prnt_vtloc_freqs (SIM_PTR p_s)
{
   char  prnt_lin[81] = "";
   int   tl_no;
   int   x1;

   if (p_s && PrintLevel >= 2)
   {
      fprintf (stdout, " Tloc allele frequencies:\n\n");

      for (x1 = 0; x1 < p_s->no_tloc_sel; ++x1)
      {
         /* Get next tloc number from sim_chrom_node array of selected tlocs. */
         tl_no = p_s->p_tloc[x1];
         /* Formats lods test tloc allele frequencies.                        */
         fmt_tloc_freqs (prnt_lin, x1, tl_no);

         if (x1 + 1 == p_s->no_tloc_sel || !((x1 + 1) % 4))
            fprintf (stdout, "%s\n", prnt_lin); /* Prints the line.           */
         else
            sprintf (prnt_lin + (x1 % 4) * 20 + 15, "     ");
      }

      fprintf (stdout, "\n\n");
      fflush (stdout);
   }

   return;
}
