/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

#include "parsie.h"
#include "parsedefs.h"
#include "parsepars.h"
#include "parseproc.h"
#include "parseprog_opts.h"
#include "errspars.h"
#include "parsedrop.h"
#include "parseauto.h"
#include "parselods.h"
#include "parsekin.h"
#include "parsepoly.h"

#include "parsefils.h"
#include "parseseeds.h"
#if 0
#include "parsedisp.h"
#endif

/*
 * proc_pars() parses the parameter statements and stores the results in
 * intermediate tables.  Then proc_drop(), for the genedrop program,
 * proc_drop() and proc_auto() for the Autozyg programs, or proc_poly(),
 * for the polyEM programs, take this information and prepare global
 * variables for the specific MORGAN program.
 *                                                 MAJ Jul 1999, Dec 2001
 */

int proc_pars (int argc, char **argv, char *ppn)
{
   char  err_lev;

   /* looks up calling program's name in tables specifying parameter limits   */
   for (psx = ProgSimSel; psx->pnam[0] != NUL && strcmp (ppn, psx->pnam); ++psx)
      ;
   for (pmx = ProgDat; pmx->pcod != NUL && pmx->pcod != psx->pcod; ++pmx)
      ;
   for (pox1 = ProgOpt1; pox1->pcod != NUL && pox1->pcod != psx->pcod; ++pox1)
      ;
   for (pox2 = ProgOpt2; pox2->pcod != NUL && pox2->pcod != psx->pcod; ++pox2)
      ;
   for (pox3 = ProgOpt3; pox3->pcod != NUL && pox3->pcod != psx->pcod; ++pox3)
      ;
   for (pox4 = ProgOpt4; pox4->pcod != NUL && pox4->pcod != psx->pcod; ++pox4)
      ;
   for (pox5 = ProgOpt5; pox5->pcod != NUL && pox5->pcod != psx->pcod; ++pox5)
      ;
   set_defaults ();

   if (!get_cmd_fnams (argc, argv))     /* gets command line file names       */
      if (proc_err ())
         return FALSE;

   pfloop = Filtab;                     /* process one file each loop         */

   while (pfloop->type[0] != NUL)
   {
      if (pfloop->fname && pfloop->parstat && !pfloop->done)
      {
         if (!open_par_fil ())          /* opens input files                  */
            if (proc_err ())
               return FALSE;

         while (!pfloop->done)          /* processes 1 statement each loop    */
         {

            if (!get_tokens ())         /* gets list of tokens                */
            {
               if (proc_err ())
                  return FALSE;
            }
            else if (!get_aggr_key ())  /* gets the statement type            */
            {
               if (proc_err ())
                  return FALSE;
            }
            else
            {
               err_lev = get_data_list (); /* gets the parameter data         */
               if (err_lev && proc_err ())
                  return FALSE;

               if (err_lev != 'E')
                  if ((ppp->func) ())   /* calls statement-specific fncs      */
                     if (proc_err ())
                        return FALSE;
            }
         }
         if (!close_par_fil ())         /* closes input files                 */
            if (proc_err ())
               return FALSE;

         fflush (stdout);
      }
      ++pfloop;                         /* goes on to next file type          */
   }
   return TRUE;
}

/* ----------- set_defaults:  sets default values to parameters ------------- */

void set_defaults (void)
{
   err_typ = 0;

   add_cor_flag = 0;
   res_cor_flag = 0;
   reset_fit_var (0);
   reset_fit_var (1);
   reset_fit_cov (0);
   reset_fit_cov (1);

   n_eigen = 0;
   eigen_set = N;
   eigen_compute = N;

   max_npar = MAXNTOKEN;
   token = NULL;
   tok_typ = NULL;
   j_par = NULL;
   u_par = NULL;
   d_par = NULL;
   m_par = NULL;
   realloc_arrs ();

   max_tchar = MAXNTOKEN * NOMTOKLEN;
   token[0] = NULL;
   realloc_toks (0);

   return;
}

/* ------ get_cmd_fnams:  gets the file names from the command line --------- */
/*                        checks the file types specified                     */
/*                        stores the names in Filtab[]                        */

int get_cmd_fnams (int argc, char **argv)
{
   int   i;

   if (argc < 2)                        /* tests for parameter file as        */
   {                                    /* first argument in command line     */
      err_typ = ERR_EIF;
      return FALSE;
   }
   else
   {
      pft = Filtab;
      pft->fname = alloc_copy (argv[1]);
   }
   for (i = 2; i < argc - 1; i += 2)    /* identifies other files named       */
   {                                    /* in command line                    */
      pft = get_filent (argv[i]);
      if (pft)
      {
         if (strchr (pft->pcods, psx->pcod))
         {
            if (pft->reclo)             /* checks for rewritable file type    */
            {
               --pft;
               pft->rewrit = Y;
            }
            pft->command = Y;
            pft->fname = alloc_copy (argv[i + 1]);
         }
         else
         {                              /* checks for file used by program    */
            err_typ = ERR_PCL;
            err_txt = pft->type;
         }
      }
      else
      {
         err_typ = ERR_FNU;             /* reports bad file type              */
         err_txt = argv[i];
      }
   }
   if ((argc & 1) != 0)                 /* tests for odd no. of arguments     */
   {
      i = 0;
      err_typ = ERR_FCL;
      sprintf (errnumb, COM_LIN_FORM);
      while ((pft = get_filtyp (psx->pcod)))
      {
         if (i && !(i % 3))
            sprintf (&errnumb[strlen (errnumb)], "\n                      ");
         sprintf (&errnumb[strlen (errnumb)], " [%s <%sfile>]",
                  pft->type, pft->type);
         ++i;
      }
      err_txt = errnumb;
   }
   if (err_typ)
      return FALSE;
   else
      return TRUE;
}

/* ---------------- open_par_fil:  opens the input file --------------------  */
/*                     gets the first line of input                           */

int open_par_fil (void)
{
   char *ptok, *pstr;

   /*                                    * opens the file                     */
   errno = 0;
   if (!(pfile = fopen (pfloop->fname, pfloop->fop)))
   {
      err_typ = ERR_OPF;
      sprintf (errnumb, "%s \"%s\":  %s", pfloop->fdesc,
               pfloop->fname, strerror (errno));
      err_txt = errnumb;
      pfloop->done = Y;
      return FALSE;
   }

   pstr = line_in;                      /* tests for file not empty           */
   while (get_line () && !(ptok = strtok (pstr, WHITESPACE)))
      ;
   if (err_typ)
      return FALSE;

   if (pfloop->done)
   {
      err_typ = ERR_SIF;
      err_txt = pfloop->fname;
      return FALSE;
   }
   else
      *strchr (ptok, NUL) = ' ';        /* gets first token                   */

   return TRUE;
}

/* ---------------- close_par_fil:  closes the input file ------------------- */

int close_par_fil (void)
{
   errno = 0;
   if (pfile)
      if (fclose (pfile) == EOF)
      {
         err_typ = ERR_CLF;
         sprintf (errnumb, "%s \"%s\":  %s", pfloop->fdesc,
                  pfloop->fname, strerror (errno));
         err_txt = errnumb;
      }
   pfile = NULL;

   if (err_typ)
      return FALSE;
   return TRUE;
}

/* ------ chk_stor:  checks for enough storage allocated for arrays -------   */

void chk_stor (int n_tok, int m_char)
{                                       /* n_tok:  no. tokens stored          */
   if (n_tok + 2 > max_npar)
   {
      max_npar *= 2;
      realloc_arrs ();
   }                                    /* m_char:  string length             */
   if (token[n_tok] - token[0] + m_char + 1 >= max_tchar)
   {
      max_tchar *= 2;
      realloc_toks (n_tok);
   }
   return;
}

/* ---------- realloc_arrs:  allocates or reallocates storage -------------   */
/*                            for arrays for input data                       */

void realloc_arrs (void)
{
   int   i;
   char **p;

   token = realloc (token, max_npar * sizeof (char *));
   alloc_chk (WHERE, token);
   tok_typ = realloc (tok_typ, max_npar * sizeof (int));
   alloc_chk (WHERE, tok_typ);
   j_par = realloc (j_par, max_npar * sizeof (int));
   alloc_chk (WHERE, j_par);
   u_par = realloc (u_par, max_npar * sizeof (uint32));
   alloc_chk (WHERE, u_par);
   d_par = realloc (d_par, max_npar * sizeof (double));
   alloc_chk (WHERE, d_par);

   p = m_par;
   m_par = realloc (m_par, max_npar * sizeof (char *));
   alloc_chk (WHERE, m_par);
   if (!p)
      m_par[0] = NULL;
   m_par[0] = realloc (m_par[0], max_npar * (MAXNAMLEN + 1) * sizeof (char));
   alloc_chk (WHERE, m_par[0]);
   for (i = 1; i < max_npar; ++i)
      m_par[i] = m_par[i - 1] + MAXNAMLEN + 1;

   return;
}

/* ----------- realloc_toks:  allocates or reallocates storage -------------  */
/*                             for array of tokens                            */

void realloc_toks (int n_tok)
{
   int   i;
   char *ptmp;

   ptmp = token[0];
   token[0] = (char *) realloc (token[0], max_tchar);
   alloc_chk (WHERE, token[0]);
   if (n_tok && token[0] != ptmp)
      for (i = 1; i <= n_tok; ++i)
         token[i] = token[0] + (token[i] - ptmp);

   return;
}

/* --------- free_arrs:  frees all of proc_pars' malloc'd arrays ------------ */

void free_arrs (void)
{
   free ((void *) token[0]);
   token[0] = NULL;
   free ((void *) token);
   token = NULL;
   free ((void *) tok_typ);
   tok_typ = NULL;
   free ((void *) j_par);
   j_par = NULL;
   free ((void *) u_par);
   u_par = NULL;
   free ((void *) d_par);
   d_par = NULL;
   free ((void *) m_par[0]);
   m_par[0] = NULL;
   free ((void *) m_par);
   m_par = NULL;

   return;
}

/* ------------ get_line:  gets the next line of input data  ---------------- */
/*                         replaces first "#" with "\0" (for comment)         */
/*                         returns pointer to line,                           */
/*                         unless eof or line contains "****"                 */
/*                         as first keyword, then                             */
/*                         sets file pointer to NULL and returns NULL         */

char *get_line (void)
{
   char *pt, *ps;

   if (!fgets (line_in, MAXLINLEN + 1, pfile))
   {
      pfloop->done = Y;
      return NULL;
   }
   else
   {
      ps = line_in;
      if (strlen (line_in) == MAXLINLEN + 1 && *(line_in + MAXLINLEN) != '\n')
      {
         err_typ = ERR_LTL;
         err_txt = line_in;
         pfloop->done = Y;
         return NULL;
      }
      if ((pt = strchr (ps, COMMENT)))
         *pt = NUL;
      if ((pt = strtok (ps, WHITESPACE)))
      {
         if (!strncmp (ENDPAR, pt, 4))
         {
            pfloop->done = Y;
            return NULL;
         }
         else
            *strchr (pt, NUL) = ' ';
      }
      return line_in;
   }
}

/* ---------------- get_tokens:  stores a list of tokens -------------------- */
/*                               returns the number found                     */

int get_tokens (void)
{
   char *ptok, *pstr;
   char  chkverb[5];
   int   end_statement;
   int   n, i, m;

   pstr = line_in;                      /* gets tokens for current line       */
   for (n = 0; (ptok = strtok (pstr, WHITESPACE)) != NULL; ++n)
   {
      m = strlen (ptok) + 1;
      chk_stor (n, m);

      strncpy (token[n], ptok, m);
      token[n + 1] = token[n] + m;
      pstr = NULL;
   }
   token[n][0] = NUL;

   end_statement = FALSE;               /* gets the next line                 */
   while (!pfloop->done && !end_statement)
   {
      if (get_line ())
      {
         pstr = line_in;                /* tests for end of statement         */
         if ((ptok = strtok (pstr, WHITESPACE)) != NULL)
         {
            if (isalpha ((int) *ptok))
            {
               for (i = 0; i < 4 && *(ptok + i) != NUL; ++i)
                  chkverb[i] = tolower ((int) *(ptok + i));
               chkverb[i] = NUL;
               for (i = 0; strcmp (verb[i], ""); ++i)
                  if (!strncmp (verb[i], chkverb, 4))
                     end_statement = TRUE;
            }
            *strchr (ptok, NUL) = ' ';
            if (!end_statement)         /* gets tokens from the next line     */
            {
               pstr = line_in;
               for (; (ptok = strtok (pstr, WHITESPACE)) != NULL; ++n)
               {
                  m = strlen (ptok) + 1;
                  chk_stor (n, m);
                  strncpy (token[n], ptok, m);
                  token[n + 1] = token[n] + m;
                  pstr = NULL;
               }
               token[n][0] = NUL;
            }
         }
      }
   }
   if (err_typ)
      return 0;

   return n;
}

/* ---- get_aggr_key:  sets a pointer to the input statement prototype ------ */
/*                     returns key # from list                                */

int get_aggr_key (void)
{
   int   i, j, n;
   char **os;

   aggre_key[0] = NUL;
   for (i = 0, n = 0; token[i][0] != NUL; ++i)
   {
      if (isalpha ((int) token[i][0]))
      {
         tok_typ[i] = KEYWD;
         if (++n <= MAXNKEY)            /* gets aggregate key                 */
            strncat (aggre_key, token[i], 4);
      }
      else
         tok_typ[i] = PDAT;
   }
   for (i = 0; aggre_key[i] != NUL; ++i) /* converts to lower case            */
      aggre_key[i] = tolower ((int) aggre_key[i]);

   /*                                    * searches for key in table          */
   for (ppp = proto, j = 0; ppp->key[0] != NUL
        && strcmp (aggre_key, ppp->key); ++ppp, ++j)
      ;
   if (ppp->key[0] != NUL)              /* found key;  return ptr             */
      return key_ix = j + 1;

   /*                                    * another attempt to find key;
    *                                      statement may contain names        */
   for (ppp = proto, j = 0; ppp->key[0] != NUL
        && (!strpbrk (ppp->par, nam_cod)
            || strncmp (aggre_key, ppp->key, strlen (ppp->key))); ++ppp, ++j)
      ;                                 /* continue                           */

   if (ppp->key[0] != NUL)              /* found key; reassign token types    */
   {
      aggre_key[0] = NUL;
      for (i = 0; token[i][0] != NUL; ++i)
         if (isalpha ((int) token[i][0])
             && strlen (aggre_key) < strlen (ppp->key))
         {
            tok_typ[i] = KEYWD;
            strncat (aggre_key, token[i], 4);
         }
         else
            tok_typ[i] = PDAT;

      return key_ix = j + 1;
   }
   /*                                    * another attempt to find key;
    *                                      statement may have been replaced   */
   for (os = old_stat; os[0] != NUL
        && strncmp (aggre_key, os[0], strlen (os[0])); ++os)
      ;
   if (os[0][0] != NUL)                 /* found obsolete statement           */
      err_typ = ERR_OLD;
   else
      err_typ = ERR_KEY;
   return key_ix = 0;
}

/* ------- get_data_list:  converts tokens into Morgan internal data types
 *                         (integers, unsigned integers, doubles, or character
 *                         arrays)
 *                         converts all data from the current parameter
 *                         statement
 *                         returns the error level ('F', 'E', 'W') if any
 *                         errors detected, else returns the null character
 */

char get_data_list (void)
{
   int   i, j, k, n, s;
   char  c, *endp;

   prntd_flg = N;

   ni_par = nk_par = nh_par = nr_par = nx_par = nm_par = ns_par = 0;

   j = 0;

   for (i = 0; (c = ppp->par[i]) != NUL; ++i)
   {                                    /* processes all data (non-keywords)
                                           keywords) for the input statement  */
      for (; token[j][0] != NUL && tok_typ[j] == KEYWD; ++j)
         ;
      if (token[j][0] == NUL)
      {
         err_typ = ERR_DAT;
         return (proc_errors + (int) err_typ)->level;
      }
      switch (c)
      {
          case 'i':                    /* expects single integer value        */
             j_par[ni_par + nk_par] = atoi (token[j]);
             ++ni_par;
             break;
          case 'k':                    /* expects list of int values          */
             for (; token[j][0] != NUL && tok_typ[j] == PDAT; ++j)
             {
                j_par[ni_par + nk_par] = atoi (token[j]);
                ++nk_par;
             }
             --j;
             break;
          case 'h':                    /* expects single hexadec value        */
             u_par[nh_par] = (uint32) strtoul (token[j], &endp, 0);
             if (*endp != NUL)
                err_typ = ERR_HEX;
             ++nh_par;
             break;
          case 'r':                    /* expects single real value           */
             d_par[nr_par + nx_par] = atof (token[j]);
             ++nr_par;
             break;
          case 'x':                    /* expects list of real values         */
             for (; token[j][0] != NUL && tok_typ[j] == PDAT; ++j)
             {
                d_par[nr_par + nx_par] = atof (token[j]);
                ++nx_par;
             }
             --j;
             break;
          case 's':                    /* expects a string                    */
             s = strlen (token[j]) - 2;
             if (s < 1)
                err_typ = ERR_STR;
             if (token[j][0] != '\"' && token[j][0] != '\'')
                err_typ = ERR_STR;
             if (token[j][s + 1] != '\"' && token[j][s + 1] != '\'')
                err_typ = ERR_STR;
             if (!err_typ)
             {
                strncpy (s_par, &token[j][1], s);
                s_par[s] = NUL;
             }
             ++ns_par;
             break;
          case 'n':                    /* expects a list of names             */
             for (; token[j][0] != NUL && tok_typ[j] == PDAT; ++j)
             {
                s = strlen (token[j]);
                if (s > MAXNAMLEN && !err_typ)
                {
                   err_typ = ERR_NML;
                   sprintf (errnumb, "%d", MAXNAMLEN);
                   err_txt = errnumb;
                }
                strncpy (m_par[nm_par], token[j], MAXNAMLEN);
                m_par[nm_par][MAXNAMLEN] = NUL;
                ++nm_par;
             }
             --j;
             break;
          case 'g':                    /* expects a list of pairs of names
                                           and integers                       */
             for (k = 0; token[j][0] != NUL && tok_typ[j] == PDAT; ++j, k = !k)
             {
                if (!k)                 /* expects name                       */
                {
                   s = strlen (token[j]);
                   if (s > MAXNAMLEN && !err_typ)
                   {
                      err_typ = ERR_NML;
                      sprintf (errnumb, "%d", MAXNAMLEN);
                      err_txt = errnumb;
                   }
                   strncpy (m_par[nm_par], token[j], MAXNAMLEN);
                   m_par[nm_par][MAXNAMLEN] = NUL;
                   ++nm_par;
                }
                else
                {                       /* expects an integer                 */
                   if (!tok2int (token[j], &j_par[ni_par + nk_par]))
                   {
                      err_typ = ERR_BGM;
                      sprintf (errnumb, "%s", token[j]);
                      err_txt = errnumb;
                   }
                   ++nk_par;
                }
             }
             if (k)
                err_typ = ERR_PGP;
             --j;
             break;
          case 'a':                    /* expects an integer (the number of
                                           integer pairs in each set), then
                                           sets of a name plus a list of
                                           integers                           */
             n = j_par[ni_par + nk_par - 1];
             if (n < 1)
             {
                err_typ = ERR_BNN;
                sprintf (errnumb, "%d", j_par[ni_par + nk_par - 1]);
                err_txt = errnumb;
             }
             else
             {
                for (k = 0; token[j][0] != NUL && tok_typ[j] == PDAT; ++k, ++j)
                {
                   if (!(k % (2 * n + 1)))
                   {
                      s = strlen (token[j]); /* expects a name                */
                      if (s > MAXNAMLEN && !err_typ)
                      {
                         err_typ = ERR_NML;
                         sprintf (errnumb, "%d", MAXNAMLEN);
                         err_txt = errnumb;
                      }
                      strncpy (m_par[nm_par], token[j], MAXNAMLEN);
                      m_par[nm_par][MAXNAMLEN] = NUL;
                      ++nm_par;
                   }
                   else
                   {                    /* expects a list of integers         */
                      j_par[ni_par + nk_par] = atoi (token[j]);
                      ++nk_par;
                   }
                }
                if (k % (2 * n + 1))
                {
                   err_typ = ERR_AIL;
                   sprintf (errnumb, "%d", n);
                   err_txt = errnumb;
                }
                --j;
             }
             break;
      }
      ++j;
   }
   for (; token[j][0] != NUL; ++j)
      if (tok_typ[j] == PDAT)
         err_typ = ERR_XTR;

   if (err_typ)
      return (proc_errors + (int) err_typ)->level;
   else
      return NUL;
}

/* -------- there is one function for each parameter statement type --------  */
/*            they are named in a dispatch table in parsepars.h               */
/*          these functins are in the following included source code          */

/* allow, assign, assume                                                      */
#include "proc_pars_a.c"
/* check, compute                                                             */
#include "proc_pars_c.c"
/* display                                                                    */
#include "proc_pars_d.c"
/* fit                                                                        */
#include "proc_pars_f.c"
/* ignore, input                                                              */
#include "proc_pars_i.c"
/* limit                                                                      */
#include "proc_pars_l.c"
/* map                                                                        */
#include "proc_pars_m.c"
/* output                                                                     */
#include "proc_pars_o.c"
/* set, start                                                                 */
#include "proc_pars_s.c"
/* select, simulate                                                           */
#include "proc_pars_ss.c"
/* use                                                                        */
#include "proc_pars_u.c"
/* -------------- end of functions for parameter statements ----------------  */

/* functions called by proc_pars_*.c parameter statement functions            */
#include "proc_parf.c"
#include "proc_parf_ss.c"
/* error function                                                             */
#include "proc_err.c"
/* functions for setup for PedComp/kin program                                */
#include "proc_kin.c"

/* functions for setup for Genedrop, Autozyg and Lodscore programs            */
#include "proc_drop.c"
#include "proc_dropf.c"
#include "proc_next.c"

/* functions for setup for Autozyg and Lodscore programs                      */
#include "proc_auto.c"
#include "proc_autof.c"

/* functions for setup for Lodscore programs                                  */
#include "proc_vnext.c"
#include "proc_lods.c"
#include "proc_lodsf.c"

/* functions for setup for PolyEM programs                                    */
#include "proc_poly.c"

/* functions for setup for program input and output files                     */
#include "proc_fils.c"
/* last function called after processing parameter statements                 */
#include "proc_end.c"
/* functions for seed setup                                                   */
#include "proc_seeds.c"
#if 0
/* functions for setup for interactive graphic display                        */
#include "proc_disp.c"
#endif
/* function for checking no. of PolyEM fixed effects levels                   */
#include "check_pars.c"
/* function to read the eigenvalue file                                       */
#include "input_eigen.c"
/* function to free data structures used by the Pars/ functions               */
#include "proc_free.c"
