/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
#include "lsample.h"

void allocate_loci (int pen_flag)
{
   struct Locus *locus;
   TLOCPARS_PTR p_tlocp;

   int  *tp1, n_all, n_gen;
   int   i1, i2;

   /* Allocate locus structures.                                              */
   loci = calloc (NLoci, sizeof (struct Locus));
   alloc_chk (WHERE, loci);
   tp1 = calloc (NLoci * (2 * pedsize + n_compnts), sizeof (int));
   alloc_chk (WHERE, tp1);

   /* Initialize a pointer to the first tloc_pars_node so that the incomplete */
   /* penetrances for a discrete trait will be accessible.                    */
   if (P_trait_pars != NULL)
      p_tlocp = P_trait_pars->p_tloc_pars;

   for (i1 = 0; i1 < NLoci; ++i1)
   {
      locus = loci + i1;
      /* Set basic stuff.                                                     */
      locus->g_mu = 0;                  /* Will be reset if necessary.        */
      locus->index = i1;
      locus->data_index = i1;
      locus->n_all = NAll[i1];
      locus->chromosome = 1;
      locus->malepos = locus->fempos = 0.0;
      locus->freq = Qq[i1] + 1;

      if ((i1 < NTLocs) && (TraitTyp == QUANTITATIVE)) /* Quantitative data.  */
      {
         locus->locus_type = TRAIT_LOCUS;

         /* The following section was moved from schnell.c, EAT 6/2003.       */
         if (pen_flag == 4)
         {
            locus->penetrance = qtlpen;
            locus->data_index = -1;     /* Ignore the trait y[][0][].         */

            /* The following may not be necessary -- per Na Li, 2002.         */
            locus->sparse_flag = 0;
            locus->allele_trans = 0;
            locus->all_set = 0;
            locus->peel = 0;
            locus->max_alternates = 0;
            locus->genes1[0] = locus->genes1[1] = 0;
            locus->req_set[0] = locus->req_set[1] = locus->req_set[2] = 0;
            locus->r_func = 0;
            locus->n_rfuncs = 0;
         }
      }

      else if ((i1 < NTLocs) && (TraitTyp == DISCRETE)) /* Phenotypic data.   */
      {
         locus->locus_type = TRAIT_LOCUS;
         /* Several choices of penetrance model are available now as          */
         /* indicated by the flag pen_flag, but pen_flag must be              */
         /* set "by hand" in the appropriate main programs.                   */
         /* An improvement on the previous hardwiring: EAT 12/2003.           */

         /* "RECESSIVE" -- equivalent to genotypic data -- for checking.      */
         if (pen_flag == 0)
            locus->penetrance = recessive;

         /* INCOMPLETE (Added by AWG) is now the default for phenotypic data. */
         /* Values are read from the ped file, NOT hardwired: EAT 6/2003.     */
         if (pen_flag == 1)
            locus->penetrance = incomplete;

         /* ORDINAL.  Used in GAW13 Analysis.                                 */
         /* Trait is an ordinal trait with 5 levels.                          */
         if (pen_flag == 2)
            locus->penetrance = ordinal;

         /* LIABILITY.  Used in AD Analysis.                                  */
         if (pen_flag == 3)
            locus->penetrance = liability;

         /* Logistic Penetrance function.                                     */
         if (pen_flag == 5)
            locus->penetrance = logisticpen;
      }

      else if ((i1 == 0 && TraitTyp == GENOTYPIC) || (i1 != 0)) /* Genotypic. */
      {
         /* Locus is genotypic trait or marker locus.                         */
         locus->locus_type = MARKER_LOCUS;
         locus->penetrance = 0;
      }

      else                              /* This trait locus will not be used. */
      {
         /* This locus will not be used by the subroutines in lsample_peel.c. */
         locus->locus_type = ABSENT;
      }

      if ((locus->penetrance == incomplete) || (locus->penetrance == qtlpen))
      {
         n_all = locus->n_all;          /* Number of alleles at the locus.    */
         n_gen = n_all * (n_all + 1) / 2; /* Number of (unordered) genotypes. */
         /* Allocate memory for the penetrances for each genotype.            */
         locus->g_mu = calloc (n_gen, sizeof (double));
         alloc_chk (WHERE, locus->g_mu);

         if (locus->penetrance == incomplete)
         {
            for (i2 = 0; i2 < n_gen; ++i2)
               loci[i1].g_mu[i2] = p_tlocp->penetrances[i2];
            ++p_tlocp;
         }
      }

      locus->n_alls = tp1;              /* Number of alleles per component.   */
      tp1 += n_compnts;
      locus->gt[0] = tp1;
      locus->gt[1] = tp1 + pedsize;
      tp1 += 2 * pedsize;

      locus->seg[0] = xx0[i1];          /* Set up seg stuff.                  */
      locus->seg[1] = xx1[i1];
      locus->genes[0] = gg0[i1];
      locus->genes[1] = gg1[i1];
   }

   fflush (stdout);
   return;
}

void free_allocate_loci (void)
{
   if (loci)
   {
      if (loci[0].g_mu)
         free (loci[0].g_mu);
      if (loci[0].n_alls)
         free (loci[0].n_alls);
      free (loci);
      loci = NULL;
   }

   return;
}
