#include "lsample.h"
#include "msample.h"
#include "mulmsample.h"
/* cmp = 0: consider all components simutaneously;
   cmp > 0: indicator for each component.
   nburn: number of burn-in iterations
   nmcmc: number of mcmc iterations
   mul = 0: M-sampler; otherwise, MM-sampler
 */
void mcmc_lods (int cmp, int nburn, int nmcmc, int mul, int ntp,
                double **rec_left, double **rec_rght, int *loc_left,
                int *loc_rght, double **prob, double *mcmc_lod, double *mcmc_sd)
{

   int   c, k, j, t, jj, nctemp;
   int   tot_count;                     /* total number of iterations to be used
                                           for lod score estimation, for example,
                                           if 1000 iterations are scored every
                                           20, then tot_count = 50.           */
   int   bsz, bc;                       /* btch-size = nmcmc/NBtch,
                                           bc = current batch                 */
   /* These base_lik and rb_ arrays contain total (element 0) and by compnt.  */
   double *rb_prob, **rb_prbm;          /* Cumulative rb_likelihood
                                           and batch totals by position.      */
   double *rb_sd;                       /* Monte Carlo SDevs                  */
   double base_lik;                     /* Base-point log P(Y_T).             */
   double *cmpt_lik;                    /* Log-lik by position and component. */
   /*   ; position added EAT  5/2010     */
   double *ctot;                        /* CDLL by component                  */
   int ***recounts;
   double **tot_log_prob_seg;
   double **tot_log_prob_pen;
   double **temp_log_prob_seg;
   static FILE *oScorF = NULL;

   nctemp = n_compnts;

   if (cmp)
      n_compnts = 1;

   ctot = malloc (n_compnts * sizeof (double));
   alloc_chk (WHERE, ctot);
   alloc_loglik (&recounts, &tot_log_prob_seg, &temp_log_prob_seg,
                 &tot_log_prob_pen);

   /* moved down initial peeling after alloc_rb etc : EAT July 2004 */

   alloc_rb (ntp, NBtch, &cmpt_lik, &rb_prob, &rb_prbm, &rb_sd);

   base_lik = 0.0;
   for (k = 0; k < ntp; ++k)
   {
      cmpt_lik[k] = 0.0;
      rb_prob[k] = 0.0;
      rb_sd[k] = 0.0;
      for (jj = 0; jj < NBtch; ++jj)
         rb_prbm[jj][k] = 0.0;
   }

   /* Added EAT -- P(Y_T) */
   /* set_transmission (loci, 0, NMarks);   */

   set_simple_trans (loci, -1, -1, 0.5, 0.5, 0.5, 0.5);

   if (TraitTyp != QUANTITATIVE)
      base_lik = locus_mjrpeel_cmp (cmp, 1.0, 1.0, 0, loci);
   else
      base_lik = locus_mjrpeel_cmp (cmp, loci[0].ve, 1.0, 0, loci);

   /* Checks for zero probability in peeling operation.                       */
   if (base_lik == -DBL_MAX)
      give_up_where (WHERE, ERS_zpp);

   printf (" Base log-likelihood log_10 P(Y_T) = %e\n\n",
           base_lik / log (10.0));
   fflush (stdout);

   if ((nmcmc / ScorIt) * ScorIt != nmcmc)
      give_warn (ERR_err,
                 "Scoring interval does not divide realizations count");
   tot_count = nmcmc / ScorIt;

   bsz = tot_count / NBtch;
   /* checks for no. batches dividing scored iterations */
   if (bsz * NBtch != tot_count)
      give_warn (ERL_bt1, NBtch);
   /* checks for batch size > 0 */
   if (!bsz)
      give_up (ERL_bt2, NBtch);

   printf (" %d realizations will be scored, at scoring interval %d\n",
           tot_count, ScorIt);
   printf (" Scored realizations will be batched in %d sets of size %d",
           NBtch, bsz);
   printf ("\n\n");
   fflush (stdout);

   for (c = 0; c < n_compnts; ++c)
   {
      for (j = 0; j <= NMarks; ++j)
      {
         tot_log_prob_seg[c][j] = 0.0;
         tot_log_prob_pen[c][j] = 0.0;
      }
      for (j = 0; j <= NMarks; ++j)
      {
         for (jj = 0; jj < 6; ++jj)
            recounts[c][j][jj] = 0;
      }
   }
   tot_count = 0;                       /* reset tot_count to count when
                                           scored */
   bc = 0;                              /* current batch indicator */

   fflush (stdout);

   alloc_logprob ();

   /* open output scores file, if requested */
   if (FilIt && !oScorF)
      if (!(oScorF = open_fil (OSCORE)))
         give_up (ERR_qit);

   prntStatusBMC ();                    /* print status for beginning MCMC */

   /* setup variables for LM-multiple sampling: LPT Apr 2006 */
   setup_sampling_cmp (cmp, mul, NMarks);

   /* now do burn-in iterations */

   printf (" Now starting %d iterations of burn-in\n", nburn);
   fflush (stdout);

   for (t = 1; t <= nburn; t++)
   {
      if (!(t % ChkIt))
      {
         printf (" Starting burn-in iteration %d\n", t);
         fflush (stdout);
      }
      sampling_cmp (t, cmp, 1, mul, NMarks, nFGL, prob);
   }

   /* now do MCMC sampling, and score results */

   printf (" Now starting %d iterations of MCMC\n", nmcmc);
   fflush (stdout);
   for (t = 0; t < nmcmc; ++t)
   {
      if (DebugMain)
         dbgXXGG ();

      /* MCMC sampling step or scan */

      sampling_cmp (t + 1, cmp, 1, mul, NMarks, nFGL, prob);

      if (!((t + 1) % ScorIt))
      {
         /* score tot log-probs and recombs */
         if (!cmp)
         {
            for (c = 0; c < n_compnts; ++c)
            {
               ctot[c] = 0.0;
               full_seg_prob_cmp (NMarks, c, recounts[c], temp_log_prob_seg[c]);
               for (j = 1; j < NMarks; ++j)
               {
                  tot_log_prob_pen[c][j] += log (prob[c][j]);
                  tot_log_prob_seg[c][j] += temp_log_prob_seg[c][j];
                  ctot[c] += (log (prob[c][j]) + temp_log_prob_seg[c][j]);
               }
               tot_log_prob_pen[c][NMarks] += log (prob[c][NMarks]);
               ctot[c] += log (prob[c][NMarks]);
            }
         }
         else
         {
            c = 0;
            ctot[c] = 0.0;
            full_seg_prob_cmp (NMarks, cmp - 1, recounts[c],
                               temp_log_prob_seg[c]);
            for (j = 1; j < NMarks; ++j)
            {
               tot_log_prob_pen[c][j] += log (prob[cmp - 1][j]);
               tot_log_prob_seg[c][j] += temp_log_prob_seg[c][j];
               ctot[c] += (log (prob[c][j]) + temp_log_prob_seg[c][j]);
            }
            tot_log_prob_pen[c][NMarks] += log (prob[cmp - 1][NMarks]);
            ctot[c] += log (prob[cmp - 1][NMarks]);
         }

         /* score rb_likelihoods  */

         /* start new batch if current btch is complete */
         /* if( t == 0 ) peel_trace = 1; *//* liping's own command */

         if (bc < NBtch)
            do_rblik_cmp (cmp, ntp, bc, loc_left, loc_rght,
                          rec_left[1], rec_left[0], rec_rght[1], rec_rght[0],
                          rb_prob, rb_prbm, base_lik, cmpt_lik);
         ++tot_count;

         if (!((tot_count) % bsz))
            ++bc;
      }
      if (FilIt && !((t + 1) % FilIt))  /* tests for output to scores fil */
         /* modified 5/2010 to output log-likelihoods of scored iterations */
         /*  need to be scored, by component and by position; need new array */
      {
#if 0
         if (cmp == 0)
            for (c = 0; c < n_compnts; ++c)
            {
               fprintf (oScorF, "\n log-likelihoods: iter %d, cmpt %d\n",
                        t + 1, c + 1);
               for (k = 1; k < ntp; ++k)
                  fprintf (oScorF, " %f", cmpt_lik[k][c]);
               fprintf (oScorF, "\n");
            }
         else
#endif
         {
            fprintf (oScorF, "\n log-likelihoods: iter %d, cmpt %d\n",
                     t + 1, cmp);
            for (k = 1; k < ntp; ++k)
               fprintf (oScorF, " %f", cmpt_lik[k]);
            fprintf (oScorF, "\n");

            for (c=0; c<PedSize; ++c)
             { if (nghd[c].obs)
                  { fprintf (oScorF, " %s", nghd[c].self);
                    for (k = 1; k < ntp; ++k)
                        { j = loc_left[k];
                          fprintf (oScorF, " %d", j);
                        }
                    for (k = 1; k < ntp; ++k)
                        { j = loc_left[k];
                          fprintf (oScorF, " %d", gg0[j][c]);
                        }
                    for (k = 1; k < ntp; ++k)
                        { j = loc_left[k];
                          fprintf (oScorF, " %d", gg1[j][c]);
                        }
                    fprintf(oScorF,"\n");
                     }
              }
          exit(99);

         }
      }
   }
   printf ("\n");

   prntStatusFMC ();                    /* print status for finished with mcmc */

   /* compute likelihood estimates  */

   for (k = 1; k < ntp; ++k)
      /* if (loc_left[k] != loc_rght[k])      */
   {
      rb_prob[k] = rb_prob[k] / (NBtch * bsz);
      for (bc = 0; bc < NBtch; ++bc)
         rb_prbm[bc][k] = rb_prbm[bc][k] / bsz;
   }

   /* Batch-means standard devation estimates.                                */
   /* Careful with component indexing in arrays!                              */

   for (k = 1; k < ntp; ++k)
   {
      /*   if (loc_left[k] != loc_rght[k])                   */
      {
         for (bc = 0; bc < NBtch; ++bc)
            rb_sd[k] += (rb_prbm[bc][k] - rb_prob[k])
               * (rb_prbm[bc][k] - rb_prob[k]);
         rb_sd[k] = sqrt (rb_sd[k] / ((NBtch - 1) * NBtch));
      }
   }

   /* Output results for this simulation/estimation set.                      */
   /* Moved down so liks can also be by compnt: EAT July 2004  */

   if (!cmp)
   {
      for (c = 0; c < n_compnts; ++c)
      {
         pr_cmpnthead (c);

         pr_log_prob_seg (1, NMarks, -cmplist[c + 1].nmeioses,
                          tot_log_prob_seg[c], tot_count);

         pr_log_prob_pen (1, NMarks, tot_log_prob_pen[c], tot_count);

         pr_reco_counts (1, NMarks, recounts[c]);

      }
   }
   else
   {
      c = 0;
      pr_cmpnthead (cmp - 1);

      pr_log_prob_seg (1, NMarks, -cmplist[cmp].nmeioses,
                       tot_log_prob_seg[c], tot_count);

      pr_log_prob_pen (1, NMarks, tot_log_prob_pen[c], tot_count);

      pr_reco_counts (1, NMarks, recounts[c]);

   }

   for (k = 1; k < ntp; ++k)
      if ((loc_left[k] != loc_rght[k]) || (TraitTyp != GENOTYPIC))
      {
         mcmc_lod[k] = log10 (rb_prob[k]);
         mcmc_sd[k] = rb_sd[k] / (rb_prob[k] * log (10.0));
      }

   free_logprob ();
   free_1d ((void *) &ctot);
   free_loglik (&recounts, &tot_log_prob_seg, &temp_log_prob_seg,
                &tot_log_prob_pen);
   free_sampling_cmp ();

   n_compnts = nctemp;

   if (FilIt && (!cmp || (cmp == n_compnts)))
      fclose (oScorF);

   return;
}
