/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
/*****************************************************************************
  routine name: sampling2.c

  Take out the MCMC sampling portion of lm_auto program, 
  to separate from scoring of results, and accommodate burn-in etc.
                                                                   EAT Sep 2002

  set up permutation arrays for loci and switches moved from main  YJS Jan 2005

  allow for individual, sibs and generations update in M-sampler   LPT Feb 2005
 
  Macro constants PMEIO etc and MAXNMupdate changed to globals     EAT Apr 2007

******************************************************************************/

#include "lsample.h"
#include "msample.h"
#include "mulmsample.h"

static int **perm_meio = NULL;          /* permutation matrix of meioses 
                                           for each component */
static int **perm_ind = NULL;           /* permutation matrix of individuals 
                                           for each component */
static int **perm_marr = NULL;          /* permutation martix of marriages
                                           for each component */
static int *perm_loc = NULL;            /* permutation array of loci */
static int *perm_sw = NULL;             /* and switches */
static int ncmp, cmp0, nswt;

/* allocate memory and assign initial values to perm_meio,
   perm_ind and perm_loc, which are used in sampling_mul () 
   cmp = 0: sample all components; otherwise, sample cmp only
   mul = 0: use M-sampler; otherwise, use MM-sampler
*/

void setup_sampling_cmp (int cmp, int mul, int nloc)
{
   int   nms, nid, nmr, i, c, k, cc;
   Indlist *pcurr;

   /* ncmp components in total, index start from cmp0 */
   if (!cmp)
   {
      cmp0 = 1;
      ncmp = n_compnts;
   }
   else
   {
      cmp0 = cmp;
      ncmp = 1;
   }

   nswt = 0;
   for (c = cmp0, cc = 0; cc < ncmp; c++, cc++)
      nswt += cmplist[c].nmeioses;

   if (!nswt)
      return;

   if (!mul)                            /* perm_sw is used in M-sampler */
   {
      al_int_vector (&perm_sw, nswt);
      i = 0;
      for (c = cmp0, cc = 0; cc < ncmp; c++, cc++)
      {
         nms = cmplist[c].nmeioses;
         for (k = 0; k < nms; k++)
         {
            perm_sw[i] = cmplist[c].meioses[k];
            i++;
         }
      }
   }
   else                                 /* perm_meio is used in MM-sampler */
   {
      perm_meio = malloc (ncmp * sizeof (int *));
      alloc_chk (WHERE, perm_meio);
      for (c = cmp0, cc = 0; cc < ncmp; c++, cc++)
      {
         nms = cmplist[c].nmeioses;
         perm_meio[cc] = malloc (nms * sizeof (int));
         alloc_chk (WHERE, perm_meio[cc]);
         for (k = 0; k < nms; k++)
            perm_meio[cc][k] = cmplist[c].meioses[k];
      }
   }

   if (Scan)
   {
      al_int_vector (&perm_loc, nloc + 1);
      for (i = 0; i <= nloc; ++i)
         perm_loc[i] = i;

      if (mul)
      {
         perm_ind = malloc (ncmp * sizeof (int *));
         alloc_chk (WHERE, perm_ind);
         perm_marr = malloc (ncmp * sizeof (int *));
         alloc_chk (WHERE, perm_marr);

         for (c = cmp0, cc = 0; cc < ncmp; c++, cc++)
         {
            nid = cmplist[c].size;
            perm_ind[cc] = malloc (nid * sizeof (int));
            alloc_chk (WHERE, perm_ind[cc]);
            pcurr = cmplist[c].people->next;
            k = 0;
            while (pcurr && (k < nid))
            {
               perm_ind[cc][k] = pcurr->current->index;
               pcurr = pcurr->next;
               k++;
            }

            nmr = cmpmarr[c - 1].nmarr;
            perm_marr[cc] = malloc (nmr * sizeof (int));
            alloc_chk (WHERE, perm_marr[cc]);
            for (k = 0; k < nmr; k++)
               perm_marr[cc][k] = cmpmarr[c - 1].marr[k];
         }
      }
   }
   return;
}

void free_sampling_cmp (void)
{
   free_2d ((void *) &perm_meio);
   free_2d ((void *) &perm_ind);
   free_2d ((void *) &perm_marr);
   free_1d ((void *) &perm_loc);
   free_1d ((void *) &perm_sw);
   return;
}

/* t = 1, 2, ... indices of mcmc iterations */

void sampling_cmp (int t, int cmp, int mm, int mul, int nloc, int ngenes,
                   double **prob)
/* mm=0 if L-sample and M-sample sample all loci, mm=1 if markers only  */
{
   int   k, l, i, jj, c, cc, nms, nid, nmr;
   double ptemp;
   int   nupdate;
   int  *sx, *ind, *ptms, *ptid, *ptmr;
   int   ssx;

   if (rans () < LSamplProb)
   {                                    /* ***** Do L-sample *****  */
      if (!((t) % ChkIt))
      {
         printf (" Starting L-sampler iteration %d\n", t);
         fflush (stdout);
      }
      if (Scan)                         /* ****   scan   **** */
      {
         permute (perm_loc, nloc + 1);  /* permute loci for L-sampler */
         for (k = 0; k <= nloc; ++k)
         {                              /* do l-sampling for each locus */
            jj = perm_loc[k];
            if (jj >= mm)               /* if mm=1, don't do jj=0  */
               (void) do_lsmpl_cmp (cmp, nloc, jj, prob, ngenes);
         }
      }
      else
      {                                 /* ****   step   **** */

         jj = intran (mm, nloc);        /* choose a random locus
                                         * do l-sampling for this locus */
         (void) do_lsmpl_cmp (cmp, nloc, jj, prob, ngenes);

      }
   }
   else
   {                                    /* ***** Do M-sample *****  */
      if (!((t) % ChkIt))
      {
         printf (" Starting M-sampler iteration %d\n", t);
         fflush (stdout);
      }
      if (Scan)                         /* ****   scan   **** */
      {

         if (!mul)                      /* M-sampler */
         {
            permute (perm_sw, nswt);    /* permute switches for M-sampler */
            for (k = 0; k < nswt; ++k)
            {                           /* for ea. switch, get meiosis */
               l = perm_sw[k];          /* and component */
               i = (meiosis[l].ind).index;
               ssx = meiosis[l].sex;
               c = nghd[i].component - 1; /* do m-sampling for the meiosis */
               do_msmpl (nloc, mm, c, i, ssx, prob, ngenes);
            }
         }
         else                           /* MM-sampler */
         {
            for (c = cmp0, cc = 0; cc < ncmp; cc++, c++)
            {
               ptemp = rans ();

               ptms = perm_meio[cc];
               nms = cmplist[c].nmeioses;
               ptid = perm_ind[cc];
               nid = cmplist[c].size;
               ptmr = perm_marr[cc];
               nmr = cmpmarr[c - 1].nmarr;

               if (ptemp < P_meio)
                  update_meio (c, ptid, nid, ptms, nms, ptmr, nmr, nloc, mm,
                               prob, ngenes);
               else if (ptemp < P_meio + P_ind)
                  update_ind (c, ptid, nid, ptms, nms, ptmr, nmr, nloc, mm,
                              prob, ngenes);
               else if (ptemp < P_meio + P_ind + P_sib)
                  update_sib (c, ptid, nid, ptms, nms, ptmr, nmr, nloc, mm,
                              prob, ngenes);
               else
                  update_3gen (c, ptid, nid, ptms, nms, ptmr, nmr, nloc, mm,
                               prob, ngenes);
            }
         }
      }
      else                              /* ****   step   **** */
      {
         if (!mul)                      /* M-sampler */
         {
            k = intran (1, nswt) - 1;   /* choose a random switch */
            l = perm_sw[k];
            i = (meiosis[l].ind).index;
            ssx = meiosis[l].sex;       /* get meiosis and component */
            c = nghd[i].component - 1;  /* do m-sampling for this meiosis */
            do_msmpl (nloc, mm, c, i, ssx, prob, ngenes);
         }
         else
         {
            /* choose a component */
            if (cmp)                    /* if only one component is involved */
            {
               c = cmp;
               cc = 0;
            }
            else                        /* if all components are considered */
            {
               c = intran (cmp0, ncmp);
               cc = c - 1;
            }
            ptms = perm_meio[cc];
            nms = cmplist[c].nmeioses;  /* nms is the number of meioses in c */
            if (nms > 0)
            {
               permute (ptms, nms);

               /* number of meioses to update */
               i = intran (1, Max_update);
               nupdate = min (i, nms);

               al_int_vector (&(ind), nupdate);
               al_int_vector (&(sx), nupdate);
               for (i = 0; i < nupdate; i++)
               {
                  l = ptms[i];
                  ind[i] = (meiosis[l].ind).index;
                  sx[i] = meiosis[l].sex;
               }
               do_msmpl2 (c, nloc, mm, ind, sx, nupdate, prob, ngenes);
               free_1d ((void *) &ind);
               free_1d ((void *) &sx);
            }
         }
      }
   }
   return;
}

double do_lsmpl_cmp (int cmp, int nl, int jj, double **prob, int ng)
{
   double z;
   int   err, c, cc;
   set_transmission (loci, jj, nl + 1);
   z = locus_mjrpeel_cmp (cmp, loci[0].ve, 1.0, 1, loci + jj);

   /* checks for Zero probability in peeling operation                        */
   if (z == -DBL_MAX)
      give_up_where (WHERE, ERS_zpp);

   reset_FGL (jj, cmp);

   for (c = cmp0 - 1, cc = 0; cc < ncmp; cc++, c++)
   {
      prob[c][jj] = seg_pen2 (c, loci + jj, ng, &err);
      if (prob[c][jj] <= 0.0)
         give_up_where (WHERE, ERS_zpl, jj);
   }

   return z;
}

void do_rblik_cmp (int cmp, int ntp, int batch,
                   int *lleft, int *lrght, double *rml, double *rfl,
                   double *rmr, double *rfr, double *rb, double **rbb,
                   double bl, double *cl)
/* ntp = # linked trait positions, at intervals loc and recombination rec,
   loc_left, etc, are recombinations, intervals etc, for trait positions.
   rb_prob is cumulative likelihood estimate
   base_lik = base (unlinked) likelihood, computed once only   */
{
   double z;
   int   k;
   int  *genes[2], err;

   for (k = 1; k < ntp; ++k)
   {
      if ((lleft[k] != lrght[k]) || (TraitTyp != GENOTYPIC))
/* This check is reinstated, EAT Oct 2005, as it is NECESSARY for some
   computations involving some complex peeling operations if trait-marker
   incompatibility is possible */
/* NOTE: only GENOTYPIC trait case is excluded here: if user is doing
    discrete trait with some penetraces 0, same could happen: warning
    on input needed for those cases ??  */
      {
/*
 *  printf ("\n starting trait position %d", k); fflush(stdout);
 *  printf ("\n locus left %d recom %f %f,  right %d recom %f %f",
 *  lleft[k], rml[k], rfl[k], lrght[k], rmr[k], rfr[k]);
 *  fflush(stdout);
 */

         if (lleft[k] != lrght[k])
            z = do_mjrlik_cmp (cmp, lleft[k], lrght[k], rml[k], rfl[k],
                               rmr[k], rfr[k]);
         else
         {

            genes[0] = gg0[lleft[k]];
            genes[1] = gg1[lleft[k]];
            z = log (gen_pen (cmp - 1, genes, loci + 0, nFGL, &err));

         }

         cl[k] = z;
         rb[k] += exp (z - bl);
         if (rbb)
            rbb[batch][k] += exp (z - bl);
/*
 *    printf ("\n end of loop for k=%d", k); fflush(stdout);
 *    printf ("   current rb_prob = %e", rb_prob[k]); fflush(stdout);
 */
      }
   }
   return;
}

double do_mjrlik_cmp (int cmp, int ll, int lr, double rml, double rfl,
                      double rmr, double rfr)
{
   double z;
/*
   transmissions MUST be computed by set_simple_trans ONLY
   lsample_flag  MUST be 0 -- S's are NOT resampled.
 */
   set_simple_trans (loci, ll, lr, rml, rfl, rmr, rfr);

   /* trait locus peel; sample_flag=0   */
   if (TraitTyp == QUANTITATIVE)
/* FIXME -- see note above  */
      z = locus_mjrpeel_cmp (cmp, loci[0].ve, 1.0, 0, loci);
   else
/* FIXME -- see note above  */
      z = locus_mjrpeel_cmp (cmp, 1.0, 1.0, 0, loci);
   /*    printf ("\n completed peel; like = %e",z); fflush(stdout);  */

   /* checks for Zero probability in peeling operation            */
   if (z == -DBL_MAX)
      give_up_where (WHERE, ERS_zpp);

   return (z);
}
