/*******************************************************************************
  program name: gl_lods.c (Estimate or calculate LodScore by component. 
                      Using result of gl_auto )

  EAT April 2010

  Note that this program is modified  from lm_multiple.c in Lodscore.
  Note it will not do MCMC, but will take gl_auto output to compute
     lodscore contributions.

*******************************************************************************/

/**
 * @defgroup lm_markers lm_markers
 * @ingroup lodscore
 * @{
 */

#define PNAM  "lm_markers"
#define MARKERS

#include "lodscore.h"

static char **DGL_names;
static int *DGL_index;

#include "gl_lods_rtnes.c"

int main (int argc, char *argv[])
{
   int   i, j, t;
   int   nsw, n_dgl, c_dgl;
   int  *genes[2], err;
   double z;
   FILE *fp2;
   FILTAB_PTR ptb;

   /* .................................................. */

   nsw = setupAutozyg (argc, argv, PNAM, 0);
   setSampling ();

   /* At this point we have:
    *    PedSize which shd be number of individuals in gl_auto output file
    *    pedsize (or nInd) which is (perhaps fewer) number for whom the
    *       data (possibly missing) has been stored in nghds structure.
    *    NMarks which is the number of markers at which lod scores are to
    *       be evaluated
    *    MarSel which is a list of these markers by index in gl_auto output:
    *      Note: gl_auto is modified do that both programs can use
    *       orinal input marker indexing.
    *      Note: if a marker is called for that in not in the gl_auto output
    *       reconstruction of FGL labels will result in the evaluation at
    *       the proximal preceding marker that is in the gl_auto output
    *    NIt the number of scored FGL graphs to be processed from the
    *       gl_auto output
    *    Curently it is assumed that there is a SINGLE component pedigree
    *       is the gl_auto output, or that GLOBAL MCMC was used
    *         (the latter is generally NOT recommended).
    */

/*  check the setup */
/*  check the requested markers */
   printf ("\n %d Selected markers:", NMarks);
   for (i = 0; i < NMarks; ++i)
      printf ("  %d", MarSel[i]);
/*  check the data file input 
 *  for (i = 0; i < pedsize; ++i)
 *  {
 *     printf ("\n  %s  %d  ", nghd[i].self, nghd[i].obs);
 *     if (!nghd[i].obs)
 *        printf ("  NA ");
 *     else
 *     {
 *        if (TraitTyp == QUANTITATIVE)
 *           printf (" %f", nghd[i].q_dat[0]);
 *        if (TraitTyp == DISCRETE)
 *           printf (" %d ", nghd[i].d_dat[0]);
 *     }
 *     fflush (stdout);
 *  }
 */

/*  First we will reduce the DGL graphs to observed individuals       
 *  They are input in "extra file" and will output to "output scores"
 *  We will then close output scores, and re-read for computations    
 */

/* First we will reduce the DGL graph to the set of observed individuals,
 *   and output it by nghd index, not by name. (Note index is 0-origin.)
 *  Names in DGL but not in nghd will be dropped
 *  Names in DGL and in nghd but with nghd.obs ==0 will be dropped
 *  Names NOT in DGL but in nghd, with ngdh.obs !=0, will be WARNED
 */

   printf ("\n nFGL from dummy pedigree input = %d \n", nFGL);

   dgl_index ();
   c_dgl = dgl_strip (&n_dgl);
   /* here is where we now reset nFGL */

   if (n_dgl < NIt)
   {
      printf (" \n %d Graphs were requested, but only %d were given (W)",
              NIt, n_dgl);
      NIt = n_dgl;
   }
   printf (" \n Number of individuals in each DGL graph = %d", c_dgl);
   printf (" \n Number of individuals in nghd structure = %d", pedsize);
   printf (" \n Reset number of FGL = %d", nFGL);

   fflush (stdout);

/*  DGL file is now in format that IBD-graph software could be run;
 *    need to run for each marker at which lod is requested.
 *    Then we would have smaller set to run, + weights  
 *  FOR NOW:  just run them all
 */

/*      if (!(fp2 = fopen("fgl.reduce", "r")))  */

   ptb = get_filent (OSCORE);
   if (!(fp2 = fopen (ptb->fname, "r")))
      /* This is processed DGL graphs  */
      give_up (ERR_qit);
   printf ("\n Reopened reduced DGL graph file %s \n", ptb->fname);

   if (NLoci != NMarks + 1)
      printf ("\n   why???  NLoci=%d, NMarks=%d", NLoci, NMarks);
   alloc_gen_pen (nFGL, NLoci);

   for (t = 0; t < NIt; ++t)            /* Big loop over FGL graphs  */
   {
      for (i = 0; i < pedsize; ++i)     /* number in nghd structure  */
         for (j = 0; j < NMarks; ++j)   /* number of selected markers */
         {
            gg0[j][i] = -1;             /* should be unnecessary ??   */
            gg1[j][i] = -1;
         }

      fgl_set (fp2, c_dgl);
      printf ("\n done fgl-setting for grph %d \n", t + 1);
      fflush (stdout);

      for (j = 0; j < NMarks; ++j)      /* number of selected markers */
      {
         genes[0] = gg0[j];
         genes[1] = gg1[j];
         printf ("\n now to call gen_pen at marker %d orig %d",
                 j + 1, MarSel[j]);
         z = log (gen_pen (0, genes, loci + 0, nFGL, &err));
         printf ("\n lod at marker %d orig %d is %f", j + 1, MarSel[j], z);
      }
      exit (99);
   }

   fclose (fp2);
   exit (EXIT_SUCCESS);
}
