/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
/*******************************************************************************
  program name: lm_markers.c  modified to produce quantiles of
      the lod-score contributions. Renamed lm_fuzlod (at least for now).

  Program to do lod scores (replaces homozyg.c from old Autozyg)
  The state (X) consists of segregation indicators -- so can generalize to
        cases of more than one observed/pedigree, and multiple loci.

  The data, Y, is the multilocus genotypes of observed individuals
     in the pedigree.

Since the data are locus-by-locus, the likelihood ratio contribution is
just ratio of P_theta(S)/P_sim_th(S).

Output scores file added for CDLL output                    EAT Apr 2004

*******************************************************************************/

/**
 * @defgroup lm_markers lm_markers
 * @ingroup lodscore
 * @{
 */

#define PNAM  "lm_markers"
#define MARKERS

#include "lodscore.h"

void do_fzlik (int ntp, int tt, int *lleft, int *lrght, double *rml,
               double *rfl, double *rmr, double *rfr, double *rb,
               double bl, double *cl, double **fuzlod)
/* ntp = # linked trait positions, at intervals loc and recombination rec,
   loc_left, etc, are recombinations, intervals etc, for trait positions.
   rb_prob is cumulative likelihood estimate
   base_lik = base (unlinked) likelihood, computed once only   */
{
   double z;
   int   k;

   for (k = 1; k < ntp; ++k)
   {
/*      if (lleft[k] != lrght[k])                             */
/* Only necessary if trait-marker incompatibility is possible */
      {
/*
 *  printf ("\n starting trait position %d", k); fflush(stdout);
 *  printf ("\n locus left %d recom %f %f,  right %d recom %f %f",
 *  lleft[k], rml[k], rfl[k], lrght[k], rmr[k], rfr[k]);
 *  fflush(stdout);
 */
         z = do_mjrlik (lleft[k], lrght[k], rml[k], rfl[k], rmr[k], rfr[k]);

         fuzlod[k][tt] = (z - bl);
         rb[k] += exp (z - bl);
/*
 *    printf ("\n end of loop for k=%d", k); fflush(stdout);
 *    printf ("   current rb_prob = %e", rb_prob[k]); fflush(stdout);
 */
      }
   }

   return;
}

int main (int argc, char *argv[])
{
   int   j, k, jj;
   int   t, c, h;
   int   ntp;
   int   nloc;
   int   nsw;

   /* These base_lik and rb_ arrays contain total (element 0) and by compnt.  */
   double *rb_prob;                     /* Cumulative rb_likelihood
                                           and batch totals by position.      */
   double base_lik;                     /* Base-point log P(Y_T).             */
   double *cmpt_lik;                    /* Log-likelihood by component.       */

   double *ctot;                        /* CDLL by component.                 */
   int ***recounts;
   double **prob;
   /* likelihoods etc.        */
   double **tot_log_prob_seg;
   double **tot_log_prob_pen;
   double **temp_log_prob_seg;
   int   tot_count;

   int   nfuz;
   double **fuzlod;                     /* [npos+1][NIt/ScorIt]             */
   double **qfzld;                      /* [npos+1][101]                    */
   /* summed over cmpt [c]           */

   /* Map position (cM) of current test position */
#if 0
   double markpos[2];                   /* Map position (cM) of current marker
                                         ** NOTE unused */
   int   lastmarker = 0;                /* Last test position
                                         ** NOTE unused */
#endif
   double **rec_left;                   /* fem =0, male=1; by test position   */
   double **rec_rght;                   /* fem =0, male=1; by test position   */
   int  *loc_left;                      /* locus left of test position        */
   int  *loc_rght;                      /* locus rght of test position        */

   FILE *oScorF, *oSeedF;

   /* .................................................. */

   nsw = setupAutozyg (argc, argv, PNAM, 0);
   setupSampling ();
   prob = setupInherit (1);

   /* .................................................. */

   ctot = malloc ((nCompnt + 1) * sizeof (double));
   alloc_chk (WHERE, ctot);

   alloc_loglik (&recounts, &tot_log_prob_seg, &temp_log_prob_seg,
                 &tot_log_prob_pen);

   /* now set evaluation points for lange_lik */
   /* ntp is number of trait positions plus number of markers (NLocPos + 1) */

   alloc_lf_rt (NLocPos + 1, &rec_left, &rec_rght, &loc_left, &loc_rght);

   ntp = set_eval_points (NMarks, loc_left, loc_rght,
                          rec_left[0], rec_rght[0], rec_left[1], rec_rght[1]);

   if (ntp != NLocPos + 1)
      give_up_where (WHERE, ERA_mrk, NLocPos, ntp);

   nfuz = NIt / ScorIt;
   fuzlod = alloc_dbl2d (ntp, nfuz + 1);

   /* moved down initial peeling after alloc_rb etc : EAT July 2004 */

   alloc_rb (ntp, 20, &cmpt_lik, &rb_prob, NULL, NULL);

   base_lik = 0.0;
   for (k = 0; k < ntp; ++k)
   {
      cmpt_lik[k] = 0.0;
      rb_prob[k] = 0.0;
   }

   /* Added EAT -- P(Y_T) */
   /* set_transmission (loci, 0, NMarks);   */

   set_simple_trans (loci, -1, -1, 0.5, 0.5, 0.5, 0.5);

   if (TraitTyp != QUANTITATIVE)
      base_lik = locus_mjrpeel (1.0, 1.0, 0, loci);
   else
      base_lik = locus_mjrpeel (loci[0].ve, 1.0, 0, loci);

   /*   checks for Zero probability in peeling operation                      */
   if (base_lik == -DBL_MAX)
      give_up_where (WHERE, ERS_zpp);

   printf (" Base log-likelihood P(Y_T) = %e\n\n", base_lik);
   fflush (stdout);

   if ((NIt / ScorIt) * ScorIt != NIt)
      give_warn (ERR_err,
                 "Scoring interval does not divide realizations count");
   tot_count = NIt / ScorIt;

   printf (" %d realizations will be scored, at scoring interval %d\n",
           tot_count, ScorIt);
   printf ("\n\n");
   fflush (stdout);

   for (c = 0; c < nCompnt; ++c)
   {
      for (h = 0; h <= NMarks; ++h)
      {
         tot_log_prob_seg[c][h] = 0.0;
         tot_log_prob_pen[c][h] = 0.0;
      }
      for (j = 0; j <= NMarks; ++j)
      {
         for (jj = 0; jj < 6; ++jj)
            recounts[c][j][jj] = 0;
      }
   }
   tot_count = 0;                       /* reset tot_count to count when
                                           scored */

   fflush (stdout);

   alloc_logprob ();

   /* open output scores file, if requested */

   if (FilIt)
      if (!(oScorF = open_fil (OSCORE)))
         give_up (ERR_qit);

   prntStatusBMC ();                    /* print status for beginning MCMC */

   /* now do burn-in iterations */

   printf (" Now starting %d iterations of burn-in\n", NBrnIt);
   fflush (stdout);

   for (t = 1; t <= NBrnIt; t++)
   {
      if (!(t % ChkIt))
      {
         printf (" Starting burn-in iteration %d\n", t);
         fflush (stdout);
      }
      sampling (t - 1, 1, NMarks, nsw, nFGL, prob);
   }

   /* now do MCMC sampling, and score results */

   nloc = NMarks;

   printf (" Now starting %d iterations of MCMC\n", NIt);
   fflush (stdout);
   for (t = 0; t < NIt; ++t)
   {
      if (DebugMain)
         dbgXXGG ();

      /* MCMC sampling step or scan */
      sampling (t, 1, NMarks, nsw, nFGL, prob);

      if (!((t + 1) % ScorIt))
      {
         /* score tot log-probs and recombs */
         for (c = 0; c < nCompnt; ++c)
         {
            ctot[c] = 0.0;
            full_seg_prob (NMarks, c, recounts, temp_log_prob_seg);
            for (j = 1; j < NMarks; ++j)
            {
               tot_log_prob_pen[c][j] += log (prob[c][j]);
               tot_log_prob_seg[c][j] += temp_log_prob_seg[c][j];
               ctot[c] += (log (prob[c][j]) + temp_log_prob_seg[c][j]);
            }
            tot_log_prob_pen[c][nloc] += log (prob[c][nloc]);
            ctot[c] += log (prob[c][nloc]);
         }

         /* score rb_likelihoods  */
#if 0
         /* zero out meiosis indicators for line-by-line check */
         for (i = 0; i < PedSize; ++i)
            if (nghd[i].pa != &nullnode)
               for (j = 0; j <= nloc; ++j)
               {
                  xx0[j][i] = 0;
                  xx1[j][i] = 0;
               }
#endif

         do_fzlik (ntp, tot_count, loc_left, loc_rght, rec_left[1],
                   rec_left[0], rec_rght[1], rec_rght[0], rb_prob,
                   base_lik, cmpt_lik, fuzlod);
         ++tot_count;

      }
      if (FilIt && !((t + 1) % FilIt))  /* tests for output to scores fil */
         for (c = 0; c < nCompnt; ++c)
         {
            fprintf (oScorF, "%d %d %f", t + 1, c + 1, ctot[c]);
            fprintf (oScorF, "\n");
         }
   }
   printf ("\n");

   prntStatusFMC ();                    /*Print status for finished with mcmc. */
   /* now make quantiles of the fuzlod */
   qfzld = alloc_dbl2d (ntp, 101);

   for (j = 0; j < ntp; ++j)
      make_quantiles (100, nfuz / 100, fuzlod[j], qfzld[j]);

   /* now print quantiles */
   pr_quantiles (100, nfuz, 1, ntp, qfzld);

   /* compute likelihood estimates  */

   for (k = 1; k < ntp; ++k)
      /* if (loc_left[k] != loc_rght[k])      */
      rb_prob[k] = rb_prob[k] / nfuz;

   /* output results for this simulation.estimation set */
   /* Moved down so liks can also be by compnt: EAT July 2004  */

   for (c = 0; c < nCompnt; ++c)
   {
      pr_cmpnthead (c);

      pr_log_prob_seg (1, nloc, -cmplist[c + 1].nmeioses, tot_log_prob_seg[c],
                       tot_count);

      pr_log_prob_pen (1, nloc, tot_log_prob_pen[c], tot_count);

      pr_reco_counts (1, nloc, recounts[c]);

   }
/* print cumulative lhood ests */

   prntLodSHead ();

   for (k = 1; k < ntp; ++k)
   {
      if ((!prntLodSPos (k - 1)) && (TraitTyp == GENOTYPIC))
         /* the current test position is a marker position */
         printf (" %9s\n", "NA");
      else
         printf (" %8.3f %8.3f %8.3f %8.3f %8.3f\n", log10 (rb_prob[k]),
                 qfzld[k][5] / log (10.0), qfzld[k][25] / log (10.0),
                 qfzld[k][75] / log (10.0), qfzld[k][95] / log (10.0));
   }
   printf ("\n");
   fflush (stdout);

   /* now we are done with MCMC */

   if (SamplSeedsOut)
   {
      getSeeds (SAMPL);
      if (!(oSeedF = open_fil (OSEED)))
         give_up (ERR_qit);
      writSeeds (oSeedF);
      fclose (oSeedF);
   }

   /* Free up allocated memory */
   free_xxgg ();
   free_allefrq ();
   free_chromo (0);
   free_allocate_loci ();
   free_perm ();
   locus_peel_free ();

   /* now we are done  */

   exit (EXIT_SUCCESS);
}
