/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
/*****************************************************************************
  program name: lm_fzconf.c
      (based on new lm_pval in Autozyg; formerly lm_fuzzp)
     and then developed from lm_fzplod.
                                                             EAT Oct 2005

  Program to do posterior IBD probs
  The state (X) consists of segregation indicators

  X determines the IBD pattern among the specified descendant gametes.

  The marker data, Y, is the (unphased) multilocus genotypes of observed 
  individuals in the pedigree.

  The realized probability distributions are scored using a lod score
   for the trait data given a trait model, location, and realized marker
   inheritances.  A multiple-testing corrected version is based on the 
   maximum over loci.

  Output are the fuzzy-pval dsns (per Geyer & Thompson) EAT Oct 2004
          First get null-null dsn quantiles accurate to .001 (0.1%)
          Then score marker loci, as generated, against this dsn.a

  This version is to do confidence intervals, so simulation under the
    null at each gamma, and MCMC conditional on trait data is required.

******************************************************************************/

/**
 * @defgroup lm_markers lm_markers
 * @ingroup lodscore
 * @{
 */

#define PNAM  "lm_markers"
#define MARKERS
#define CMPNT  0                        /* set =1 for peeling by component */

#include "lodscore.h"
/* #include "fzp_rtnes.c"  */

int main (int argc, char *argv[])
{
   int   j, k, err;
   int   nloc, nitnull, ntp, ntpos, ntp0, nt; /* needs cleaning   */
   int   t, c, h, nsw;
   int ***recounts;                     /* [nc][nl][6] counts of
                                         *             recombinations     */
   double **prob, zz;
   /* likelihoods etc.        */
   double **tot_log_prob_seg;
   double **tot_log_prob_pen;
   double **temp_log_prob_seg;

   int   tot_count = 0;

   /* Map position (cM) of test positions */
   double **rec_left;                   /* fem =0, male=1; by test position   */
   double **rec_rght;                   /* fem =0, male=1; by test position   */
   int  *loc_left;                      /* locus left of test position        */
   int  *loc_rght;                      /* locus rght of test position        */

   double **hdsn;                       /* [ntp+1][nitnull]
                                         * list of null imputation scores
                                         * at all evaluatio positions     */
   double **fzpval;                     /* [npos+1][nnull]             */
   /* summed over cmpt [c]           */
   double **fzcval;                     /* [npos+1][nnull]             */
   /* summed over cmpt [c]           */

   double **qndsn;                      /* quantiles of the null dsn 
                                           sampled  [npos+1][nnull]     */
   /*  F^{-1}((k+1)/nnull)   is element k     */
   int   nnull;                         /* number of quantiles for null  */
   int   id0;                           /* indicator for id null marginals */
   int   decr;                          /* number of realizations per nnull */

   FILE *oSeedF;

   /* .................................................. */

   nsw = setupAutozyg (argc, argv, PNAM, 0);
   setupSampling ();

   nloc = NMarks;
   nnull = 1000;

   /* now set evaluation points for lange_lik */
   /* ntp is number of trait positions plus number of markers (NLocPos + 1) */

   alloc_lf_rt (NLocPos + 1, &rec_left, &rec_rght, &loc_left, &loc_rght);

   ntp = set_eval_points (NMarks, loc_left, loc_rght,
                          rec_left[0], rec_rght[0], rec_left[1], rec_rght[1]);
   ntpos = ntp - 1;                     /* number of in-map quantile positions   */
/*  number of null realizations   -- want a multiple of 1000  */
/*   nitnull = ((NIt*ntp)/1000)*1000 ;        too many!! */
   nitnull = NIt;                       /* temporary ?? */
/* indicator of id null marginals -- will be 0 for lodscores*/
   id0 = 0;

/*  this will be needed to score the nulls */

   set_locpeel (nloc);

   /* First generate the null-null distribution quantiles,      EAT  Oct 2004 */
   /*    The "score"  is sup_gamma* P_gamma*(Z | S_M)/log P_gamma(Z | S_M)    */
   /*       (note we have to put it this way up: reject for large values      */
   /*   For the given gamma, S_M are simulated given Z and                    */
   /*     the "score" then computed for each gamma*  (K*K things).            */
   /* This is now getting fixed (Sept 2006) from Sept 2005  (EAT)             */

/* First all the alloc stuff must be moved to up here !!  */
/*  allocate here, and reinitialize below where previously alloced            */
   qndsn = alloc_dbl2d (ntpos + 1, nnull + 1);
   hdsn = alloc_dbl2d (ntpos + 1, nitnull);
/* allocate and zero the fuzzy-pvals   */
   fzpval = alloc_dbl2d (ntpos + 1, nnull + 1);
   fzcval = alloc_dbl2d (ntpos + 1, nnull + 1);
/* allocate log-seg and log-pen stuff */
   alloc_loglik (&recounts, &tot_log_prob_seg, &temp_log_prob_seg,
                 &tot_log_prob_pen);
   alloc_logprob ();

/* Can we set up inheritance here ? */
   prob = setupInherit (0);

/* Now we can start   */
/* we will do for all positions, excl unlinked and markers */
   for (ntp0 = 3; ntp0 <= ntpos; ++ntp0)
   {
      if (loc_left[ntp0] != loc_rght[ntp0])
      {
         Loc0 = loc_left[ntp0];
         Rec0[0] = rec_left[0][ntp0];
         Rec0[1] = rec_left[1][ntp0];

         for (k = 0; k <= 1; ++k)
            set_chromosomes (NMarks, Marek[k], 0, &Loc0, &Rec0[k], k);
         nt = revchro[0][0];            /* position of trait */

/* compute lod score given realized S_M and set hdsn from nitnull nulls */
         for (j = 0; j <= ntpos; ++j)
            for (k = 0; k < nitnull; ++k)
               hdsn[j][k] = 0.0;

         printf ("\n now do %d null realizations with ntpos= %d", nitnull,
                 ntpos);
         fflush (stdout);

         for (t = 0; t < nitnull; ++t)
/* first realize S_T, and then S_M given S_T  at each gamma**, and
   comput score at given gamma, for each gamma** sample                   */
         {
            set_transmission (0, 0, 0);
            zz = locus_mjrpeel (1.0, 1.0, 1, &loci[0]);

/* set other markers to the right  */
            if (nt < NMarks)
            {
               next_marker_segs (nsw, 0, nt + 1, nt);
               for (j = nt + 1; j < NMarks; ++j)
                  next_marker_segs (nsw, j, j + 1, j);
            }

/* set other markers to the left */
            if (nt > 0)
            {
               next_marker_segs (nsw, 0, nt, nt - 1);
               for (j = nt; j > 1; --j)
                  next_marker_segs (nsw, j, j - 1, j - 2);
            }

/* at each position, compute the lod score given S_M: find the max; 
    get the score for the fixed gamma                                */
/* score the lod score measure -- store the value  */
/* note  the 0 component already stores the max score !! ;-))        */

            do_fzplik2 (t, ntp, loc_left, loc_rght, rec_left[1], rec_left[0],
                        rec_rght[1], rec_rght[0], hdsn);
            for (k = 1; k <= ntpos; ++k)
               hdsn[k][t] = hdsn[0][t] - hdsn[k][t];
         }
/* convert dsn to 0.1 % quantiles */

         for (j = 0; j <= ntpos; ++j)
            for (k = 0; k <= nnull; ++k)
               qndsn[j][k] = 0.0;

         decr = nitnull / nnull;
         for (j = 0; j <= ntpos; ++j)
            make_quantiles (nnull, decr, hdsn[j], qndsn[j]);

/* now print quantiles for check -- at some large interval                   */
         printf (" Now print null quantiles\n");
         fflush (stdout);
         pr_quantiles (nnull, nitnull, nnull / 10, ntpos + 1, qndsn);
         fflush (stdout);

/* if marginals should be ident distributed (id0=1), then merge quantiles    */
         if (id0)
         {
            compress_marginals (1, nloc, nnull, qndsn);

/* now print quantiles again for check -- at some large interval             */
            printf (" Now print merged null quantiles\n");
            fflush (stdout);
            pr_quantiles (nnull, nitnull, nnull / 10, ntpos + 1, qndsn);
            fflush (stdout);
         }

/* initialize log-seg and log-pen stuff */
/* probabilities now computed separately by component  */

         for (c = 0; c < n_compnts; ++c)
         {
            for (h = 0; h <= NMarks; ++h)
            {
               tot_log_prob_seg[c][h] = 0.0;
               tot_log_prob_pen[c][h] = 0.0;
            }
            for (j = 0; j <= NMarks; ++j)
            {
               for (k = 0; k < 6; ++k)
                  recounts[c][j][k] = 0;
            }
         }

/* zero the fzpval   */

         for (j = 0; j <= ntpos; ++j)
            for (k = 0; k <= nnull; ++k)
               fzpval[j][k] = 0.0;

/* now intialise inheritance  but do not realloc !! */
/*   prob = setupInherit (0);  */
/* set xx0, xx1 */

         if (TraitTyp != QUANTITATIVE)
            do_seq_imp (NSeqImp, 0, PedSize, NMarks + 1, 1.0);
         else
            do_seq_imp (NSeqImp, 0, PedSize, NMarks + 1, loci[0].ve);

/* set FGL  */
         for (j = 0; j <= NMarks; ++j)
            set_FGL (j);

/* set the prob array */
         for (j = 0; j <= NMarks; ++j)
            for (c = 0; c < nCompnt; ++c)
               prob[c][j] = seg_pen2 (c, loci + j, nFGL, &err);

         prntStatusBMC ();
         /* now do burn-in iterations */

         for (t = 1; t <= NBrnIt; t++)
         {
            if (!(t % ChkIt))
            {
               printf (" Starting burnin iteration %d\n", t);
               fflush (stdout);
            }
            sampling (t, 0, nloc, nsw, nFGL, prob); /* include trait locus */
         }

         /* now do main MCMC iterations */

         printf (" Now starting %d main MCMC iterations  for posn %d \n", NIt,
                 ntp0);
         fflush (stdout);

         for (t = 0; t < NIt; ++t)
         {
            /* MCMC sampling step or scan */
            sampling (t, 0, nloc, nsw, nFGL, prob);

            if (!((t + 1) % ScorIt))
            {

/* now scoring will be as a fuzzy-p dsn for each location */
/* If t(S^{(j)}) is exceeded by exactly k of the t(S_0^{(i)}) and tied with w,
 * then contrib to the histogram is rectangle area 1/n, and base endpoints
 *    k/(m+1), (k+1+w)/(m+1)
 */

               scor_fzconf (fzpval, ntp, nnull,
                            loc_left, loc_rght, rec_left[1], rec_left[0],
                            rec_rght[1], rec_rght[0], qndsn);

/* scores log-segs and log-pens       */
/*
         for (c = 0; c < n_compnts; ++c)
         {
            full_seg_prob (nloc, c, recounts, temp_log_prob_seg);
            for (j = 0; j < nloc; ++j)
            {
               tot_log_prob_pen[c][j] += log (prob[c][j]);
               tot_log_prob_seg[c][j] += temp_log_prob_seg[c][j];
            }
            tot_log_prob_pen[c][nloc] += log (prob[c][nloc]);
         }
*/
               ++tot_count;
            }
         }

         norcum (NIt, nnull, fzpval[ntp0]);
         for (k = 0; k < nnull; ++k)
            fzcval[ntp0][k] = 1.0 - fzpval[ntp0][k];

/*
   printf (" chk interim ntp0 = %d \n", ntp0);
   for (k=0; k< nnull; k=k+10)
         printf (" %d %5.3f \n", k, fzcval[ntp0][k]);
 */
      }
   }

   printf ("\n    Pointwise distributions of fuzzy c-values, by location:");
   printf ("\npval maxim");

   fflush (stdout);
   pr_pvals (ntpos, fzcval, nnull, nnull / 10);

/*
   for (c = 0; c < n_compnts; ++c)
   {
      pr_cmpnthead (c);
      pr_log_prob_seg (0, NMarks, -cmplist[c + 1].nmeioses, tot_log_prob_seg[c],                       tot_count);

      pr_log_prob_pen (0, NMarks, tot_log_prob_pen[c], tot_count);

      pr_reco_counts (0, NMarks, recounts[c]);

   }
 */
/* free the full set of null dsn realizations: only the quantiles are needed */
   free_2d ((void *) &hdsn);

   printf ("done MCMC\n ");             /* ****  now we are done **** */

   /* Free up allocated memory */

   free_xxgg ();
   free_loglik (&recounts, &tot_log_prob_seg, &temp_log_prob_seg,
                &tot_log_prob_pen);
   locus_peel_free ();
   free_allocate_loci ();
   free_chromo (0);
   free_gen_pen ();
   free_seg_pen (nloc + 1, &prob);
   free_logprob ();
   free_msmpl ();
   free_traits ();
   free_cmplists ();

   if (SamplSeedsOut)
   {
      getSeeds (SAMPL);
      if (!(oSeedF = open_fil (OSEED)))
         give_up (ERR_qit);
      writSeeds (oSeedF);
      fclose (oSeedF);
   }
   if (Scan)
      free_perm ();

   exit (EXIT_SUCCESS);
}

/** @} */
