/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
/* $Id: alloc_array.c,v 1.3.2.4 2009-03-23 18:37:56 slewis Exp $ */

#include "stuff.h"

int **alloc_int2d (int dim1, int dim2)
{
   int   x1;
   int **array2d;

   /* Allocate memory for a two dimensional array of integers.                */
   /* The first dimension will consist of pointers to the second dimension.   */
   array2d = (int **) malloc (dim1 * sizeof (int *));
   alloc_chk (WHERE, array2d);

   /* The second dimension will consist of the actual integer vectors.        */
   array2d[0] = (int *) malloc (dim1 * dim2 * sizeof (int));
   alloc_chk (WHERE, array2d[0]);

   /* Fill in the pointers in the first dimension so that they point to the   */
   /* actual integer vectors which make up the second dimension.              */
   for (x1 = 1; x1 < dim1; ++x1)
      array2d[x1] = array2d[x1 - 1] + dim2;

   /* Initialize all of the actual integer data to zeroes.                    */
   memset (array2d[0], 0, dim1 * dim2 * sizeof (int));

   return array2d;
}

int ***alloc_int3d (int dim1, int dim2, int dim3)
{
   int   x1, x2, *next_addr;
   int ***array3d;

   /* Allocate memory for a three dimensional array of integers.              */
   /* The first dimension will consist of pointers to the second dimension.   */
   array3d = (int ***) malloc (dim1 * sizeof (int **));
   alloc_chk (WHERE, array3d);

   /* The second dimension will consist of pointers to the third dimension.   */
   array3d[0] = (int **) malloc (dim1 * dim2 * sizeof (int *));
   alloc_chk (WHERE, array3d[0]);

   /* Fill in the pointers in the first dimension so that they point to the   */
   /* pointers which make up the second dimension.                            */
   for (x1 = 1; x1 < dim1; ++x1)
      array3d[x1] = array3d[x1 - 1] + dim2;

   /* The third dimension will consist of the actual integer vectors.         */
   array3d[0][0] = (int *) malloc (dim1 * dim2 * dim3 * sizeof (int));
   alloc_chk (WHERE, array3d[0][0]);

   /* Fill in the pointers in the second dimension so that they point to the  */
   /* actual integer vectors which make up the third dimension.               */
   next_addr = array3d[0][0];
   for (x1 = 0; x1 < dim1; ++x1)
   {
      for (x2 = 0; x2 < dim2; ++x2)
      {
         array3d[x1][x2] = next_addr;
         next_addr += dim3;
      }
   }

   /* Initialize all of the actual integer data to zeroes.                    */
   memset (array3d[0][0], 0, dim1 * dim2 * dim3 * sizeof (int));

   return array3d;
}

double **alloc_dbl2d (int dim1, int dim2)
{
   int   x1;
   double **array2d;

   /* Allocate memory for a two dimensional array of double precision         */
   /* floating point numbers.  The first dimension will consist of pointers   */
   /* to the second dimension.                                                */
   array2d = (double **) malloc (dim1 * sizeof (double *));
   alloc_chk (WHERE, array2d);

   /* The second dimension will consist of the actual double precision        */
   /* floating point number vectors.                                          */
   array2d[0] = (double *) malloc (dim1 * dim2 * sizeof (double));
   alloc_chk (WHERE, array2d[0]);

   /* Fill in the pointers in the first dimension so that they point to the   */
   /* actual double precision floating point number vectors which make up the */
   /* second dimension.                                                       */
   for (x1 = 1; x1 < dim1; ++x1)
      array2d[x1] = array2d[x1 - 1] + dim2;

   /* Initialize all of the actual double precision floating point numbers to */
   /* zeroes.                                                                 */
   memset (array2d[0], 0.0, dim1 * dim2 * sizeof (double));

   return array2d;
}

double ***alloc_dbl3d (int dim1, int dim2, int dim3)
{
   int   x1, x2;
   double ***array3d, *next_addr;

   /* Allocate memory for a three dimensional array of double precision       */
   /* floating point numbers.  The first dimension will consist of pointers   */
   /* to the second dimension.                                                */
   array3d = (double ***) malloc (dim1 * sizeof (double **));
   alloc_chk (WHERE, array3d);

   /* The second dimension will consist of pointers to the third dimension.   */
   array3d[0] = (double **) malloc (dim1 * dim2 * sizeof (double *));
   alloc_chk (WHERE, array3d[0]);

   /* Fill in the pointers in the first dimension so that they point to the   */
   /* pointers which make up the second dimension.                            */
   for (x1 = 1; x1 < dim1; ++x1)
      array3d[x1] = array3d[x1 - 1] + dim2;

   /* The third dimension will consist of the actual double precision         */
   /* floating point number vectors.                                          */
   array3d[0][0] = (double *) malloc (dim1 * dim2 * dim3 * sizeof (double));
   alloc_chk (WHERE, array3d[0][0]);

   /* Fill in the pointers in the second dimension so that they point to the  */
   /* actual double precision floating point number vectors which make up the */
   /* third dimension.                                                        */
   next_addr = array3d[0][0];
   for (x1 = 0; x1 < dim1; ++x1)
      for (x2 = 0; x2 < dim2; ++x2)
      {
         array3d[x1][x2] = next_addr;
         next_addr += dim3;
      }

   /* Initialize all of the actual double precision floating point numbers to */
   /* zeroes.                                                                 */
   memset (array3d[0][0], 0.0, dim1 * dim2 * dim3 * sizeof (double));

   return array3d;
}

char **alloc_char2d (int dim1, int dim2)
{
   int   x1;
   char **array2d;

   /* Allocate memory for a two dimensional array of character strings.       */
   /* The first dimension will consist of pointers to the second dimension.   */
   array2d = (char **) malloc (dim1 * sizeof (char *));
   alloc_chk (WHERE, array2d);

   /* The second dimension will consist of the actual character string        */
   /* vectors.                                                                */
   array2d[0] = (char *) malloc (dim1 * dim2 * sizeof (char));
   alloc_chk (WHERE, array2d[0]);

   /* Fill in the pointers in the first dimension so that they point to the   */
   /* actual character string vectors which make up the second dimension.     */
   for (x1 = 1; x1 < dim1; ++x1)
      array2d[x1] = array2d[x1 - 1] + dim2;

   /* Initialize all of the actual character string data to zeroes.           */
   memset (array2d[0], 0, dim1 * dim2 * sizeof (char));

   return array2d;
}

char ***alloc_char3d (int dim1, int dim2, int dim3)
{
   int   x1, x2;
   char ***array3d, *next_addr;

   /* Allocate memory for a three dimensional array of character strings.     */
   /* The first dimension will consist of pointers to the second dimension.   */
   array3d = (char ***) malloc (dim1 * sizeof (char **));
   alloc_chk (WHERE, array3d);

   /* The second dimension will consist of pointers to the third dimension.   */
   array3d[0] = (char **) malloc (dim1 * dim2 * sizeof (char *));
   alloc_chk (WHERE, array3d[0]);

   /* Fill in the pointers in the first dimension so that they point to the   */
   /* pointers which make up the second dimension.                            */
   for (x1 = 1; x1 < dim1; ++x1)
      array3d[x1] = array3d[x1 - 1] + dim2;

   /* The third dimension will consist of the actual character string         */
   /* vectors.                                                                */
   array3d[0][0] = (char *) malloc (dim1 * dim2 * dim3 * sizeof (char));
   alloc_chk (WHERE, array3d[0][0]);

   /* Fill in the pointers in the second dimension so that they point to the  */
   /* actual character string vectors which make up the third dimension.      */
   next_addr = array3d[0][0];
   for (x1 = 0; x1 < dim1; ++x1)
      for (x2 = 0; x2 < dim2; ++x2)
      {
         array3d[x1][x2] = next_addr;
         next_addr += dim3;
      }

   /* Initialize all of the actual character string data to zeroes.           */
   memset (array3d[0][0], 0, dim1 * dim2 * dim3 * sizeof (char));

   return array3d;
}

void free_1d (void **array)
{
   if (*array)
      free (*array);

   *array = NULL;
   return;
}

void free_2d (void ***array)
{
   if (*array)
   {
      if (**array)
         free (**array);
      free (*array);
   }

   *array = NULL;
   return;
}

void free_3d (void ****array)
{
   if (*array)
   {
      if (**array)
      {
         if (***array)
            free (***array);
         free (**array);
      }
      free (*array);
   }

   *array = NULL;
   return;
}

/******************************************************************************/
/* Routines to allocate int and dbl vectors transferred here: EAT May 2006.   */
/* These are used by Liping's routines that were in MulMsample.               */
/******************************************************************************/

void al_int_vector (int **array, int dim1)
{
   *array = malloc (dim1 * sizeof (int));
   alloc_chk (WHERE, *array);

   return;
}

void al_double_vector (double **array, int dim1)
{
   *array = malloc (dim1 * sizeof (double));
   alloc_chk (WHERE, *array);

   return;
}
