/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

/*  functions are:
 *       intWid(min, max)         - returns no. of chars for writing 
 *                                  integers in range of min and max
 *       decWid(min, max, sigfig) - returns no. of decimal places to write 
 *                                  doubles in specified range and with
 *                                  a max. of "sigfig: significant figures
 *       fldWid(min, max, dec)    - returns no. of characters to write 
 *                                  doubles in specified range and with "dec"
 *                                  decimal places
 *       fmt_marknam(*nam, mno)   - converts marker no. to marker name
 *       fmt_traitnam(*nam, tno)  - converts trait no. to trait name
 *       get_colfmt(linw, headw, dataw, spac, *ncols, *colw)
 *                                - determines the column width needed,
 *                                  and the number of columns per line
 *        get_colfmt2 (linw, part1, part2, spac)
 *                                - determines the number of column pairs
 *                                  per line
 */

#include "stuff.h"

int intWid (int min, int max)
{
   int   n, m;

   n = min < 0 ? floor (log10 ((double) (-min))) + 2 : 1;
   m = max > 0 ? floor (log10 ((double) max)) + 1 : 1;

   return (m > n ? m : n);
}

int decWid (double dmin, double dmax, int sigfig)
{
   int   n, m, p;

   dmin = dmin < 0. ? -dmin : dmin;
   n = dmin == 0. ? 0 : floor (log10 (dmin)) + 1;
   dmax = dmax < 0. ? -dmax : dmax;
   m = dmax == 0. ? 0 : floor (log10 (dmax)) + 1;
   p = sigfig - (m > n ? m : n);

   return (p > 0 ? p : 0);
}
int fldWid (double dmin, double dmax, int dec)
{
   int   n, m;

   n = dmin < 0. ? floor (log10 (-dmin)) + 1 : 1;
   n = n < 0 ? 1 : n;
   n = dmin < 0 ? n + 1 : n;
   m = dmax > 0. ? floor (log10 (dmax)) + 1 : 1;

   return (m > n ? m + dec + 1 : n + dec + 1);
}

/**
 * converts marker number into marker name.
 * returns the number of characters used for the name.
 */
int fmt_marknam (char *nam, int mno)
{
   char *prefix = "marker-";

   strncpy (nam, prefix, 7);
   sprintf (nam + 7, "%d", mno);

   return strlen (nam);
}

/**
 * converts trait number into trait name.
 * returns the number of characters used for the name.
 */
int fmt_traitnam (char *nam, int tno)
{
   char *prefix = "trait-";

   strncpy (nam, prefix, 6);
   sprintf (nam + 6, "%d", tno);

   return strlen (nam);
}

/**
 * converts tloc number into tloc name.
 * returns the number of characters used for the name.
 */
int fmt_tlocnam (char *nam, int tno)
{
   char *prefix = "tloc-";

   strncpy (nam, prefix, 5);
   sprintf (nam + 5, "%d", tno);

   return strlen (nam);
}

/**
 * Determines the number of characters to use for data and heading columns
 * and the number of columns per line.
 * The givens are the number of characters to use in the line,
 * the number needed for the column heading and for the data,
 * and the spacing between columns
 */
void get_colfmt (int linw, int headw, int dataw, int spac,
                 int *ncols, int *colw)
{
   *colw = headw > dataw ? headw : dataw;

   *ncols = linw / (*colw + spac) > 1 ? linw / (*colw + spac) : 1;

   return;
}

/**
 * Determines the number of pairs of columns per line.
 * The givens are the number of characters to use in the line,
 * the number needed for each pair of data,
 * and the spacing between columns
 * It is assumed that there is a single space between the pairs of data.
 */
int get_colfmt2 (int linw, int part1, int part2, int spac)
{
   return linw / (part1 + part2 + 1 + spac);
}
